/**
 *
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 */

package com.ibm.sdo.tck.testcasebase;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.TimeZone;

import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.datatype.XMLGregorianCalendar;

import org.junit.After;
import org.junit.Before;

import com.ibm.sdo.tck.framework.TCKTestBase;
import com.ibm.sdo.tck.util.EqualityTest;
import com.ibm.sdo.tck.util.TCKUtil;

import commonj.sdo.Type;

public class ConversionTestBase extends TCKTestBase 
{
    protected static final int TYPE_INDEX_BOOLEAN   = 0;
    protected static final int TYPE_INDEX_BYTE      = 1;
    protected static final int TYPE_INDEX_CHARACTER = 2;
    protected static final int TYPE_INDEX_DOUBLE    = 3;
    protected static final int TYPE_INDEX_FLOAT     = 4;
    protected static final int TYPE_INDEX_INT       = 5;
    protected static final int TYPE_INDEX_LONG      = 6;
    protected static final int TYPE_INDEX_SHORT     = 7;
    protected static final int TYPE_INDEX_STRING    = 8;
    protected static final int TYPE_INDEX_BYTES     = 9;
    protected static final int TYPE_INDEX_DECIMAL   = 10;
    protected static final int TYPE_INDEX_INTEGER   = 11;
    protected static final int TYPE_INDEX_DATE      = 12;
    protected static final String [] TYPE_NAMES = { "Boolean", "Byte", "Character", "Double", "Float", "Int",
                                                  "Long", "Short", "String", "Bytes", "Decimal", "Integer", "Date" };
    protected static final char [] TYPE_CONV_TABLE = { 
         /*Boolean*/    'X', ' ', ' ', ' ', ' ', ' ', ' ', ' ', 'x', ' ', ' ', ' ', ' ',  
         /*Byte*/       ' ', 'X', ' ', 'x', 'x', 'x', 'x', 'x', 'x', ' ', ' ', ' ', ' ', 
         /*Character*/  ' ', ' ', 'X', ' ', ' ', ' ', ' ', ' ', 'x', ' ', ' ', ' ', ' ', 
         /*Double*/     ' ', 'x', ' ', 'X', 'x', 'x', 'x', 'x', 'x', ' ', 'x', 'x', ' ', 
         /*Float*/      ' ', 'x', ' ', 'x', 'X', 'x', 'x', 'x', 'x', ' ', 'x', 'x', ' ', 
         /*Int*/        ' ', 'x', ' ', 'x', 'x', 'X', 'x', 'x', 'x', ' ', 'x', 'x', ' ', 
         /*Long*/       ' ', 'x', ' ', 'x', 'x', 'x', 'X', 'x', 'x', ' ', 'x', 'x', 'x', 
         /*Short*/      ' ', 'x', ' ', 'x', 'x', 'x', 'x', 'X', 'x', ' ', ' ', ' ', ' ', 
         /*String*/     'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'X', 'x', 'x', 'x', 'x', 
         /*Bytes*/      ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', 'x', 'X', ' ', 'x', ' ', 
         /*Decimal*/    ' ', ' ', ' ', 'x', 'x', 'x', 'x', ' ', 'x', ' ', 'X', 'x', ' ', 
         /*Integer*/    ' ', ' ', ' ', 'x', 'x', 'x', 'x', ' ', 'x', 'x', 'x', 'X', ' ',
         /*Date*/       ' ', ' ', ' ', ' ', ' ', ' ', 'x', ' ', 'x', ' ', ' ', ' ', 'X' 
    };
    
    protected String NS       = "http://example.com/simple";
    protected String TypeName = "Simple";
    
    protected Type propContainerType;
    protected static EqualityTest.GeneralComparator COMPARE_ANY = new EqualityTest.GeneralComparator();

    @Before
    public void setUp () throws Exception {
      super.setUp();
      
      String elements = "";
      for( int i=0; i<TYPE_NAMES.length; i++ ) {
          elements += "<xs:element name=\""+TYPE_NAMES[i]+"\" type=\"sdo:"+TYPE_NAMES[i]+"\" /> "; 
      }
      String xsd = "<xs:schema xmlns:sdo=\"commonj.sdo\" " + 
                     "xmlns:sdoJava=\"commonj.sdo/java\" " +
                     "xmlns:helpconvert=\""+NS+"\" " +
                     "xmlns:xs=\"http://www.w3.org/2001/XMLSchema\" " +
                     "targetNamespace=\""+NS+"\"> " +
                   "<xs:import namespace=\"commonj.sdo\" schemaLocation=\"http://www.osoa.org/sdo/2.1/schemas/sdoModel.xsd\" /> "+                     
                   "<xs:complexType abstract=\"false\" name=\""+TypeName+"\"> " +
                   "<xs:sequence> " +
                   elements + 
                   "</xs:sequence> " +
                   "</xs:complexType>" +
                   "<xs:element name=\""+TypeName+"\" type=\"helpconvert:"+TypeName+"\"/> " +
                   "</xs:schema>";
      
      getScope().getXSDHelper().define(xsd);
      propContainerType = getScope().getTypeHelper().getType(NS,TypeName);
    }
    
    @After
    public void tearDown() throws Exception {
      super.tearDown();
    }
    

    protected static Object getTestValue(int fromType, int toType ) throws Exception
    {
        Object value;
        switch(fromType)
        {
            case TYPE_INDEX_BOOLEAN: 
                value = true;
                break;
            case TYPE_INDEX_BYTE:    
                value = (byte)50;
                break;
            case TYPE_INDEX_CHARACTER:  
                value = (char)50;
                break;
            case TYPE_INDEX_DOUBLE:  
                value = (double)50;
                break;
            case TYPE_INDEX_FLOAT:  
                value = (float)50;
                break;
            case TYPE_INDEX_INT:  
                value = (int)50;
                break;
            case TYPE_INDEX_LONG:
                if( toType == TYPE_INDEX_DATE )
                    value = TCKUtil.getDateValue().getTime();
                else
                    value = (long)50;
                break;
            case TYPE_INDEX_SHORT: 
                value = (short)50;
                break;
            default:
            case TYPE_INDEX_STRING:
                if( toType == TYPE_INDEX_BOOLEAN )
                    value = "true";
                else if( toType == TYPE_INDEX_CHARACTER )
                    value = "P";
                else if( toType == TYPE_INDEX_DATE )
                    value = TCKUtil.DATE_TEST_DATA;
                else
                    value = "50";
                break;
            case TYPE_INDEX_BYTES:    
                byte[] bytes = {'5','0'};
                value = bytes;
                break;
            case TYPE_INDEX_DECIMAL:  
                value = new BigDecimal(50);
                break;
            case TYPE_INDEX_INTEGER:   
                value = new BigInteger("50");
                break;
            case TYPE_INDEX_DATE:  
                value = TCKUtil.getDateValue();
                break;
        }
        return value;
    }
}

