/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 *  
 *  $Rev: 537762 $  $Date: 2008/01/15 11:17:52 $
 */
package com.ibm.sdo.tck.testcasebase;

import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;


import org.junit.Test;

import com.ibm.sdo.tck.framework.TCKTestBase;
import com.ibm.sdo.tck.framework.annotations.Operation;
import com.ibm.sdo.tck.framework.annotations.SpecRef;
import com.ibm.sdo.tck.framework.annotations.TCKTest;
import com.ibm.sdo.tck.framework.annotations.Operation.ErrorOutcome;
import com.ibm.sdo.tck.util.TCKUtil;


import commonj.sdo.DataObject;
import commonj.sdo.Type;

@TCKTest(testFor={@Operation(className="HelperContext", methodName="getDataFactory", signature=""),
                  @Operation(className="DataFactory", methodName="create", signature="Type"),
                  @Operation(className="DataFactory", methodName="create", signature="Type", precondition="Parameter Type=Abstract", expectedException=ErrorOutcome.EXCEPTION),                  
                  @Operation(className="DataFactory", methodName="create", signature="Type", expectedException=ErrorOutcome.EXCEPTION),                  
                  /* @Operation(className="DataFactory", methodName="create", signature="Class"), */ // See testDataFactoryCreateClass for explanation
                  @Operation(className="DataFactory", methodName="create", signature="String,String"), 
                  @Operation(className="DataObject", methodName="getType", signature=""),
                  @Operation(className="Type", methodName="getProperties", signature=""),
                  @Operation(className="Type", methodName="getInstanceClass", signature=""),
                  @Operation(className="Property", methodName="getName", signature="")})
public abstract class  DataFactoryTestBase extends TCKTestBase {
  protected   DataObject testDO;
  
  /**
   * Verify the proper operation of DataFactory.create(Type).
   */
  @Test
  public void testDataFactoryCreateByType() {
      Type type = testDO.getType();

      DataObject typeDO = getScope().getDataFactory().create(type);

      assertNotNull("DataFactory.create(Type) did not return a DataObject as expected", typeDO);

      assertTrue("DataFactory.create(Type) did not return a DataObject of the expected Type.", TCKUtil.areEqualTypes(typeDO.getType(), type));
  }

  /**
   * Verify the proper operation of DataFactory.create(URI, Name).
   */
  @Test
  public void testDataFactoryCreateByURI() {
      Type type = testDO.getType();

      DataObject uriDO = getScope().getDataFactory().create(type.getURI(), type.getName());

      assertNotNull("DataFactory.create(URI, Name) did not return a DataObject as expected.", uriDO);

      assertTrue("DataFactory.create(URI, Name) did not return a DataObject of the expected Type.",
                 TCKUtil.areEqualTypes(uriDO.getType(), type));
  }

  /**
   * Verify the proper operation of DataFactory.create(Class).
   * Here we test that a datatype class returns a wrapper DO as resolved in JSR235 Issue 13.
   * Real DataObject create by Class are covered in the Static SDO tests.
   */
  @Test
  public void testDataFactoryCreateByClass() {
      DataObject classDO = getScope().getDataFactory().create(Boolean.class);
      classDO.set("value", true);
      assertNotNull("DataFactory.create(Class) did not return a DataObject as expected.", classDO);
      
      assertTrue("DataFactory.create(Class) did not return a wrapper DataObject as expected.",
                     classDO.getInstanceProperty("value") != null);
  }

  /**
   * Verify that DataFactory.create with DataType results in
   * IllegalArgumentException
   */
  //@Test
  public void testDataFactoryCreateDataTypeException() {
      try {
          Type dataType = getScope().getTypeHelper().getType("commonj.sdo", "Int");
          assertTrue("Could not test DataFactory.create(dataType) because dataType.isDataType = false", dataType
              .isDataType());
          getScope().getDataFactory().create(dataType);
          fail("DataFactory.create(dataType) did not throw an Exception.");
      } catch (IllegalArgumentException e) {
          // Do nothing
      } catch (Exception e) {
          fail("DataFactory.create(dataType) did not throw the expected IllegalArgumentException Exception.");
      }
  }

  /**
   * Verify that DataFactory.create with Abstract Type results in
   * IllegalArgumentException.
   */
  @Test
  @TCKTest(specRef={@SpecRef(version="2.1",section="3.7.2")})
  public void testDataFactoryCreateAbstractException() {
      try {
          Type abstractType =
              getScope().getTypeHelper().getType(TCKUtil.TEST_NAMESPACE, "Abstract");

          assertTrue("Could not test DataFactory.create(abstractType) because abstractType.isAbstract = false",
                     abstractType.isAbstract());
          getScope().getDataFactory().create(abstractType);
          fail("DataFactory.create(abstractType) did not throw an Exception.");

      } catch (IllegalArgumentException e) {
          // Do nothing
      } catch (Exception e) {
          fail("DataFactory.create(abstractType) did not throw the expected IllegalArgumentException Exception.");
      }
  }
}
