/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 *  
 *  $Rev: 538143 $  $Date: 2008/01/25 22:02:12 $
 */
package com.ibm.sdo.tck.testcasebase;

import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.Reader;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.Writer;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.junit.Test;
import org.w3c.dom.Document;

import com.ibm.sdo.tck.framework.TCKTestBase;
import com.ibm.sdo.tck.framework.annotations.Operation;
import com.ibm.sdo.tck.framework.annotations.SpecRef;
import com.ibm.sdo.tck.framework.annotations.TCKTest;
import com.ibm.sdo.tck.framework.annotations.Operation.ErrorOutcome;
import com.ibm.sdo.tck.util.EqualityTest;
import com.ibm.sdo.tck.util.TCKUtil;


import commonj.sdo.DataObject;
import commonj.sdo.Type;
import commonj.sdo.helper.HelperContext;
import commonj.sdo.helper.XMLDocument;

// The following empty meta-data tag is currently required in order for
// post-jar processing of the meta-data tags above methods below.  
@TCKTest()
public abstract class XMLHelperTestBase extends TCKTestBase {
    protected   DataObject testDO;
    protected   HelperContext helperContext;

    /**
     * Verifies the XMLHelper.load(String) method. <p/>
     * @throws Exception 
     * @see commonj.sdo.XSDHelper#load(String)
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=46">2.1 spec section 3.11.2</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.11.2")},
            testFor={@Operation(className="XMLHelper",methodName="load",signature="String")})    
    @Test
    public void testXMLHelperLoadStringMethod() throws Exception {
        TCKUtil.populateAPITestObjectProperties(testDO, helperContext);
        Type type = testDO.getType();
        
        String strTestXMLDocument = helperContext.getXMLHelper().save(testDO, TCKUtil.TEST_NAMESPACE, TCKUtil.API_ELEMENT );
        assertNotNull("Serialization returned a null value.",strTestXMLDocument);
        
        XMLDocument testXMLDocument = helperContext.getXMLHelper().load(strTestXMLDocument);
        assertNotNull("Deserialization returned a null value.",testXMLDocument);
        DataObject tempDO = testXMLDocument.getRootObject();
        EqualityTest et = new EqualityTest(testDO,tempDO,true);        
        assertTrue("Serialization and deserialization resulted in a nonequivalent DataObject: " + et.getFailure(),
                    et.getResult() );
    }
    
    /**
     * Verifies the XMLHelper.load(InputStream) method. <p/>
     * @throws Exception 
     * @see commonj.sdo.XSDHelper#load(InputStream)
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=46">2.1 spec section 3.11.2</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.11.2")},
            testFor={@Operation(className="XMLHelper",methodName="load",signature="InputStream")})    
    @Test
    public void testXMLHelperLoadStreamMethod() throws Exception {
        TCKUtil.populateAPITestObjectProperties(testDO, helperContext);
        Type type = testDO.getType();
        
        String strTestXMLDocument = helperContext.getXMLHelper().save(testDO, type.getURI(), type.getName() );
        assertNotNull("Serialization returned a null value.",strTestXMLDocument);
        
        ByteArrayInputStream bais = new ByteArrayInputStream(strTestXMLDocument.getBytes());
        XMLDocument testXMLDocument = helperContext.getXMLHelper().load(bais);
        assertNotNull("Deserialization returned a null value.",testXMLDocument);
        DataObject tempDO = testXMLDocument.getRootObject();
        EqualityTest et = new EqualityTest(testDO,tempDO,true);        
        assertTrue("Serialization and deserialization resulted in a nonequivalent DataObject: " + et.getFailure(),
                    et.getResult() );        
    }
    
    /**
     * Verifies the XMLHelper.load(InputStream,String,Object) method. <p/>
     * @throws Exception 
     * @see commonj.sdo.XSDHelper#load(InputStream, String, Object)
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=46">2.1 spec section 3.11.2</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.11.2")},
            testFor={@Operation(className="XMLHelper",methodName="load",signature="InputStream,String,Object")})    
    @Test
    public void testXMLHelperLoadStreamStringObjectMethod() throws Exception {
        TCKUtil.populateAPITestObjectProperties(testDO, helperContext);
        Type type = testDO.getType();
        
        String strTestXMLDocument = helperContext.getXMLHelper().save(testDO, type.getURI(), type.getName() );
        assertNotNull("Serialization returned a null value.",strTestXMLDocument);
        
        ByteArrayInputStream bais = new ByteArrayInputStream(strTestXMLDocument.getBytes());
        XMLDocument testXMLDocument = helperContext.getXMLHelper().load(bais,"",null);
        assertNotNull("Deserialization returned a null value.",testXMLDocument);
        DataObject tempDO = testXMLDocument.getRootObject();
        EqualityTest et = new EqualityTest(testDO,tempDO,true);        
        assertTrue("Serialization and deserialization resulted in a nonequivalent DataObject: " + et.getFailure(),
                    et.getResult() );
        
    }
    
    /**
     * Verifies the XMLHelper.load(Reader,String,Object) method. <p/>
     * @throws Exception 
     * @see commonj.sdo.XSDHelper#load(Reader,String,Object)
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=46">2.1 spec section 3.11.2</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.11.2")},
            testFor={@Operation(className="XMLHelper",methodName="load",signature="Reader,String,Object")})    
    @Test
    public void testXMLHelperLoadReaderStringObjectMethod() throws Exception {
        TCKUtil.populateAPITestObjectProperties(testDO, helperContext);
        Type type = testDO.getType();
        
        String strTestXMLDocument = helperContext.getXMLHelper().save(testDO, type.getURI(), type.getName() );
        assertNotNull("Serialization returned a null value.",strTestXMLDocument);
        
        Reader rdr = new StringReader(strTestXMLDocument);
        XMLDocument testXMLDocument = helperContext.getXMLHelper().load(rdr,"",null);
        assertNotNull("Deserialization returned a null value.",testXMLDocument);
        DataObject tempDO = testXMLDocument.getRootObject();
        EqualityTest et = new EqualityTest(testDO,tempDO,true);        
        assertTrue("Serialization and deserialization resulted in a nonequivalent DataObject: " + et.getFailure(),
                    et.getResult() );
    }
    
    /**
     * Verifies the XMLHelper.save(XMLDocument,OutputStream,Object) method. <p/>
     * @throws Exception 
     * @see commonj.sdo.XSDHelper#save(XMLDocument,OutputStream,Object)
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=46">2.1 spec section 3.11.2</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.11.2")},
        testFor={@Operation(className="XMLHelper",methodName="save",signature="XMLDocument,OutputStream,Object")})
    @Test
    public void testXMLHelperSaveXMLDocumentOutputStreamObjectMethod() throws Exception {           
        TCKUtil.populateAPITestObjectProperties(testDO, helperContext);
        Type type = testDO.getType();
            
        String strTestXMLDocument = helperContext.getXMLHelper().save(testDO, type.getURI(), type.getName() );
        assertNotNull("Serialization returned a null value.",strTestXMLDocument);
        XMLDocument testXMLDocument = helperContext.getXMLHelper().load(strTestXMLDocument);
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        helperContext.getXMLHelper().save(helperContext.getXMLHelper().createDocument(testXMLDocument.getRootObject(), 
                testXMLDocument.getRootElementURI(), testXMLDocument.getRootElementName()), baos, null);
        DataObject tempDO = helperContext.getXMLHelper().load(baos.toString()).getRootObject();
        EqualityTest et = new EqualityTest(testDO,tempDO,true);        
        assertTrue("Serialization and deserialization resulted in a nonequivalent DataObject: " + et.getFailure(),
                    et.getResult() );
                    
    }
    
    /**
     * Verifies the XMLHelper.load(Source,String,Object) method. <p/>
     * @throws Exception 
     * @see commonj.sdo.XSDHelper#load(Source,String,Object)
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=46">2.1 spec section 3.11.2</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.11.2")},
            testFor={@Operation(className="XMLHelper",methodName="load",signature="Source,String,Object")})    
    @Test
    public void testXMLHelperLoadSourceStringObjectMethod() throws Exception {
        TCKUtil.populateAPITestObjectProperties(testDO, helperContext);
        Type type = testDO.getType();

        String strTestXMLDocument = helperContext.getXMLHelper().save(testDO, type.getURI(), type.getName() );
        assertNotNull("Serialization returned a null value.",strTestXMLDocument);
        
        DocumentBuilder db = DocumentBuilderFactory.newInstance().newDocumentBuilder();
        Document doc = db.newDocument();
        DOMResult domRes = new DOMResult();
        domRes.setNode(doc);
        XMLDocument testXMLDocument = helperContext.getXMLHelper().load(strTestXMLDocument);
        helperContext.getXMLHelper().save(helperContext.getXMLHelper().createDocument(testXMLDocument.getRootObject(), 
                            testXMLDocument.getRootElementURI(), testXMLDocument.getRootElementName()), domRes, null);
        DOMSource ds = new DOMSource(domRes.getNode());
        testXMLDocument = helperContext.getXMLHelper().load(ds, "", null);
        assertNotNull("Deserialization returned a null value.",testXMLDocument);
        DataObject tempDO = testXMLDocument.getRootObject();
        EqualityTest et = new EqualityTest(testDO,tempDO,true);        
        assertTrue("Serialization and deserialization resulted in a nonequivalent DataObject: " + et.getFailure(),
                    et.getResult() );

    }
    

    /**
     * Verifies the XMLHelper.save(XMLDocument,Result,Object) method. <p/>
     * @throws Exception 
     * @see commonj.sdo.XSDHelper#save(XMLDocument,Result,Object)
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=46">2.1 spec section 3.11.2</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.11.2")},
        testFor={@Operation(className="XMLHelper",methodName="save",signature="XMLDocument,Result,Object")})
    @Test
    public void testXMLHelperSaveXMLDocumentResultObjectMethod() throws Exception {           
        TCKUtil.populateAPITestObjectProperties(testDO, helperContext);
        Type type = testDO.getType();
            
        String strTestXMLDocument = helperContext.getXMLHelper().save(testDO, type.getURI(), type.getName() );
        assertNotNull("Serialization returned a null value.",strTestXMLDocument);
        XMLDocument testXMLDocument = helperContext.getXMLHelper().load(strTestXMLDocument);
        DocumentBuilder db = DocumentBuilderFactory.newInstance().newDocumentBuilder();
        Document doc = db.newDocument();
        DOMResult domRes = new DOMResult();
        domRes.setNode(doc);
        helperContext.getXMLHelper().save(testXMLDocument, domRes, null);
        DOMSource ds = new DOMSource(domRes.getNode());
        StringWriter stringWriter = new StringWriter();
        Transformer xmlTransform = TransformerFactory.newInstance().newTransformer();            
        xmlTransform.transform(ds, new StreamResult(stringWriter));
        StringBuffer buffer =  stringWriter.getBuffer();            
        testXMLDocument = helperContext.getXMLHelper().load(buffer.toString());
        assertNotNull("Deserialization returned a null value.",testXMLDocument);
        DataObject tempDO = testXMLDocument.getRootObject();
        EqualityTest et = new EqualityTest(testDO,tempDO,true);        
        assertTrue("Serialization and deserialization resulted in a nonequivalent DataObject: " + et.getFailure(),
                     et.getResult() );
            
 
    }

    /**
     * Verifies the XMLHelper.save(DataObject,String,String,OutputStream) method. <p/>
     * @throws Exception 
     * @see commonj.sdo.XSDHelper#save(DataObject,String,String,OutputStream)
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=46">2.1 spec section 3.11.2</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.11.2")},
        testFor={@Operation(className="XMLHelper",methodName="save",signature="DataObject,String,String,OutputStream")})
    @Test
    public void testXMLHelperSaveDataObjectStringStringOtuputStreamMethod() throws Exception {           
        TCKUtil.populateAPITestObjectProperties(testDO, helperContext);
        Type type = testDO.getType();
            
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        helperContext.getXMLHelper().save(testDO, type.getURI(), type.getName(), baos);
        DataObject tempDO = helperContext.getXMLHelper().load(baos.toString()).getRootObject();
        EqualityTest et = new EqualityTest(testDO,tempDO,true);        
        assertTrue("Serialization and deserialization resulted in a nonequivalent DataObject: " + et.getFailure(),
                    et.getResult() );            
    }

    /**
     * Verifies the XMLHelper.save(DataObject,String,String) method. <p/>
     * @throws Exception 
     * @see commonj.sdo.XSDHelper#save(DataObject,String,String)
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=46">2.1 spec section 3.11.2</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.11.2")},
        testFor={@Operation(className="XMLHelper",methodName="save",signature="DataObject,String,String")})
    @Test
    public void testXMLHelperSaveMethods() throws Exception {           
        TCKUtil.populateAPITestObjectProperties(testDO, helperContext);
        Type type = testDO.getType();
            
        String strTestXMLDocument = helperContext.getXMLHelper().save(testDO, type.getURI(), type.getName() );
        assertNotNull("Serialization returned a null value.",strTestXMLDocument);
        XMLDocument testXMLDocument = helperContext.getXMLHelper().load(strTestXMLDocument);
        assertNotNull("Deserialization returned a null value.",testXMLDocument);
        DataObject tempDO = testXMLDocument.getRootObject();
        EqualityTest et = new EqualityTest(testDO,tempDO,true);        
        assertTrue("Serialization and deserialization resulted in a nonequivalent DataObject: " + et.getFailure(),
                    et.getResult() );
            
    }
    
    /**
     * Verifies the XMLHelper.save(XMLDocument,Writer,Object) method. <p/>
     * @throws Exception 
     * @see commonj.sdo.XSDHelper#save(XMLDocument,Writer,Object)
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=46">2.1 spec section 3.11.2</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.11.2")},
        testFor={@Operation(className="XMLHelper",methodName="save",signature="XMLDocument,Writer,Object")})
    @Test
    public void testXMLHelperSaveXMLDocumentWriterObjectMethod() throws Exception {           
        TCKUtil.populateAPITestObjectProperties(testDO, helperContext);
        Type type = testDO.getType();
            
        String strTestXMLDocument = helperContext.getXMLHelper().save(testDO, type.getURI(), type.getName() );
        assertNotNull("Serialization returned a null value.",strTestXMLDocument);
        XMLDocument testXMLDocument = helperContext.getXMLHelper().load(strTestXMLDocument);
        assertNotNull("Deserialization returned a null value.",testXMLDocument);
            
        StringWriter xmlStringOutput = new StringWriter();
        helperContext.getXMLHelper().save(helperContext.getXMLHelper().createDocument(testXMLDocument.getRootObject(), 
                testXMLDocument.getRootElementURI(), testXMLDocument.getRootElementName()), xmlStringOutput, null);
        DataObject tempDO = helperContext.getXMLHelper().load(xmlStringOutput.toString()).getRootObject();
        EqualityTest et = new EqualityTest(testDO,tempDO,true);        
        assertTrue("Serialization and deserialization resulted in a nonequivalent DataObject: " + et.getFailure(),
                    et.getResult() );
 
    }
    

    /**
     * Verifies that appropriate exception is thrown when an
     * invalid XML document is specified. <p/>
     * @see commonj.sdo.XSDHelper#load(InputStream)
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=46">2.1 spec section 3.11.2</a>
     */   
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.11.2")},
             testFor={@Operation(className = "XMLHelper", methodName = "load", signature = "String", precondition="Faulty XML Source", expectedException=ErrorOutcome.EXCEPTION)})
    @Test
    public void testFaultyXML() {
        try {
           helperContext.getXMLHelper().load(XMLHelperTestBase.class.getResourceAsStream("/tobeplaced/faulty.xml"));
           fail("Attempting to load a faulty XML file should result in an Exception.");
        } catch (Exception e) {
           // success
        }
    }

    /**
     * Tests the condition that an XML document is in itself valid, but does not comply with the XSD.
     * @see commonj.sdo.XSDHelper#load(InputStream)
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=46">2.1 spec section 3.11.2</a>
     */ 
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.11.2")},
             testFor={@Operation(className = "XMLHelper", methodName = "load", signature = "String", precondition="Missing Schema", expectedException=ErrorOutcome.EXCEPTION)})
    @Test
    public void testXMLWrongXSD() {
        try {
            helperContext.getXMLHelper().load(XMLHelperTestBase.class.getResourceAsStream("/tobeplaced/wrongNamespace.xml"));
            fail("Attempting to load an XML specifying the wrong namespace file should result in an Exception.");
        } catch (Exception e) {
            // success
        }
    }
    
    /**
     * Test XMLHelper.createDocument method
     * @throws Exception 
     * @see commonj.sdo.XSDHelper#createDocument(DataObject,String,String)
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=46">2.1 spec section 3.11.2</a>
     */ 
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.11.2")},
             testFor={@Operation(className = "XMLHelper", methodName = "createDocument", signature = "DataObject,String,String")})
    @Test
    public void testXMLHelperCreateDocument() throws Exception {
        TCKUtil.populateAPITestObjectProperties(testDO, helperContext);
        Type type = testDO.getType();
        XMLDocument testXMLDocument = helperContext.getXMLHelper().createDocument(testDO, type.getURI(), type.getName());
        DataObject tempDO = testXMLDocument.getRootObject();
        EqualityTest et = new EqualityTest(testDO,tempDO,true);        
        assertTrue("Serialization and deserialization resulted in a nonequivalent DataObject: " + et.getFailure(),
                    et.getResult());
    }
}
