/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 *  
 *  $Rev: 538928 $  $Date: 2008/01/04 16:43:14 $
 */
package com.ibm.sdo.tck.tests.core;


import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertSame;
import static org.junit.Assert.assertTrue;

import java.io.InputStream;
import java.math.BigDecimal;
import java.net.URL;
import java.util.List;


import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import com.ibm.sdo.tck.framework.TCKTestBase;
import com.ibm.sdo.tck.framework.annotations.Operation;
import com.ibm.sdo.tck.framework.annotations.SpecRef;
import com.ibm.sdo.tck.framework.annotations.TCKTest;
import com.ibm.sdo.tck.util.TCKUtil;

import commonj.sdo.ChangeSummary;
import commonj.sdo.DataObject;
import commonj.sdo.Property;
import commonj.sdo.Sequence;
import commonj.sdo.Type;


public class ChangeSummaryTest extends TCKTestBase
{

    protected final String NS             = "http://www.example.com/cstest";
    protected final String ROOTDGNAME     = "RootDG";
    protected final String ROOTNAME       = "Root";
    protected final String PARENT1NAME    = "Parent1";
    protected final String PARENT2NAME    = "Parent2";
    protected final String CHILDNAME      = "Child";
    
    @Before
    public void setUp () throws Exception {
      super.setUp();
      
      URL url = getClass().getResource("/sharedFiles/changeSummaryTest.xsd");
      InputStream inputStream = url.openStream();
      getScope().getXSDHelper().define(inputStream, url.toString());
      inputStream.close();      
    }
   
    @After
    public void tearDown() throws Exception {
      super.tearDown();
    }
    
    public DataObject getRootDataObject()
    {
        Type type = getScope().getTypeHelper().getType(NS,ROOTNAME);
        DataObject dobj = getScope().getDataFactory().create(type);
        return dobj;
    }
    
    
    /**
     * Verifies basic ChangeSummary behavior <p/>
     * @see commonj.sdo.DataObject#getChangeSummary()
     * @see commonj.sdo.ChangeSummary#getChangedDataObjects()
     * @see commonj.sdo.ChangeSummary#getOldValues(DataObject)
     * @see commonj.sdo.ChangeSummary#beginLogging()
     * @see commonj.sdo.ChangeSummary#endLogging()
     * @see commonj.sdo.ChangeSummary#isLogging()
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=54">2.1 spec section 3.3.7</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.3.7")},
            testFor={@Operation(className="ChangeSummary",methodName="getChangedDataObjects",signature=""),
                     @Operation(className="ChangeSummary",methodName="getOldValues",signature="DataObject"),
                     @Operation(className="ChangeSummary",methodName="beginLogging",signature=""),
                     @Operation(className="ChangeSummary",methodName="endLogging",signature=""),
                     @Operation(className="ChangeSummary",methodName="isLogging",signature=""),
                     @Operation(className="DataObject",methodName="getChangeSummary",signature="")})    
    @Test
    public void testSimple() 
    {
        DataObject root = getRootDataObject();
        ChangeSummary cs = root.getChangeSummary();
        cs.beginLogging();

        List co = cs.getChangedDataObjects();
        assertEquals(0, co.size());
        root.set("symbol", "some text");
        assertEquals(0, co.size());
        co = cs.getChangedDataObjects();
        assertEquals(1, co.size());
      
        List oldValues = cs.getOldValues((DataObject)co.get(0));
      
        ChangeSummary.Setting ov1 = (ChangeSummary.Setting)oldValues.get(0);
        Property p = ov1.getProperty();
        assertEquals("symbol", p.getName());
        Object v = ov1.getValue();
        assertEquals(null, v);

        assertTrue(cs.isLogging());
        cs.endLogging();
        assertFalse(cs.isLogging());
    }
     
    /**
     * Verifies basic ChangeSummary behavior <p/>
     * @see commonj.sdo.DataObject#getChangeSummary()
     * @see commonj.sdo.ChangeSummary#getOldValues(DataObject,Property)
     * @see commonj.sdo.ChangeSummary#undoChanges()
     * @see commonj.sdo.ChangeSummary#beginLogging()
     * @see commonj.sdo.ChangeSummary#endLogging()
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=54">2.1 spec section 3.3.7</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.3.7")},
            testFor={@Operation(className="ChangeSummary",methodName="getOldValues",signature="DataObject,Property"),
                     @Operation(className="ChangeSummary",methodName="undoChanges",signature=""),
                     @Operation(className="ChangeSummary",methodName="endLogging",signature=""),
                     @Operation(className="ChangeSummary",methodName="beginLogging",signature=""),
                     @Operation(className="DataObject",methodName="getChangeSummary",signature="")})    
    @Test
    public void testBasics() throws Exception 
    {
        DataObject quote = getRootDataObject();

        ChangeSummary changeSummary = quote.getChangeSummary();
        assertNotNull(changeSummary);
        assertFalse(changeSummary.isLogging());
        assertTrue(changeSummary.getRootObject() == quote);

        quote.setString("symbol", "NONE");
        quote.setBigDecimal("low", new BigDecimal("999.9"));

        changeSummary.beginLogging();
        assertTrue(changeSummary.isLogging());

        quote.setString("symbol", "fbnt");
        quote.setString("companyName", "FlyByNightTechnology");

        quote.setBigDecimal("open1", new BigDecimal("1000.9"));
        quote.setBigDecimal("price", new BigDecimal("1000.9"));

        quote.setBigDecimal("high", new BigDecimal("999.9"));
        quote.setBigDecimal("high", new BigDecimal("1000.9"));
        quote.setBigDecimal("low", new BigDecimal("1000.9"));
        quote.setDouble("volume", 1000);
        quote.setDouble("change1", 1000);

        DataObject child = quote.createDataObject("quotes");
        child.setBigDecimal("price", new BigDecimal("2000.9"));

        DataObject child2 = getScope().getDataFactory().create("http://www.example.com/cstest", "Quote");
        quote.getList("quotes").add(0, child2);
        child.setBigDecimal("open1", new BigDecimal("1000.9"));

        changeSummary.endLogging();
        assertFalse(changeSummary.isLogging());

        assertEquals(3, changeSummary.getChangedDataObjects().size()); 
        assertTrue(changeSummary.getChangedDataObjects().contains(quote));
        assertTrue(changeSummary.getChangedDataObjects().contains(child));
        assertFalse(changeSummary.isCreated(quote));
        assertTrue(changeSummary.isCreated(child));
        assertTrue(changeSummary.isCreated(child2));

        ChangeSummary.Setting ov = 
        changeSummary.getOldValue(quote, quote.getType().getProperty("symbol"));
        assertEquals("NONE", ov.getValue());
        assertEquals("fbnt", quote.getString("symbol"));

        ChangeSummary.Setting ov2 = 
        changeSummary.getOldValue(quote, quote.getType().getProperty("companyName"));
        assertNull(ov2.getValue());
        assertEquals(quote.getString("companyName"), "FlyByNightTechnology");
        
        ChangeSummary.Setting ov3 = 
            changeSummary.getOldValue(quote, quote.getType().getProperty("low"));
            assertEquals(new BigDecimal("999.9"), ov3.getValue());
            assertEquals(new BigDecimal("1000.9"), quote.getBigDecimal("low"));

        assertEquals(2, quote.getList("quotes").size());
        assertEquals("fbnt", quote.getString("symbol"));

        quote.getChangeSummary().undoChanges();

        assertEquals(0, quote.getList("quotes").size());
        assertEquals("NONE", quote.getString("symbol"));
        assertEquals(new BigDecimal("999.9"), quote.getBigDecimal("low"));
        assertNull(quote.getString("companyName"));
    }


    /**
     * Verifies basic ChangeSummary behavior using nested DataObjects <p/>
     * @see commonj.sdo.DataObject#getChangeSummary()
     * @see commonj.sdo.ChangeSummary#getChangedDataObjects()
     * @see commonj.sdo.ChangeSummary#endLogging()
     * @see commonj.sdo.ChangeSummary#beingLogging()
     * @see commonj.sdo.ChangeSummary#isLogging()
     * @see commonj.sdo.ChangeSummary#isCreated()
     * @see commonj.sdo.ChangeSummary#undoChanges()
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=54">2.1 spec section 3.3.7</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.3.7")},
            testFor={@Operation(className="ChangeSummary",methodName="getChangedDataObjects",signature=""),
                     @Operation(className="ChangeSummary",methodName="beginLogging",signature=""),
                     @Operation(className="ChangeSummary",methodName="endLogging",signature=""),
                     @Operation(className="ChangeSummary",methodName="isLogging",signature=""),
                     @Operation(className="ChangeSummary",methodName="isCreated",signature="DataObject"),   
                     @Operation(className="ChangeSummary",methodName="undoChanges",signature=""),
                     @Operation(className="DataObject",methodName="getChangeSummary",signature="")})    
    @Test
    public void testNested() 
    {
      
        DataObject quote = getRootDataObject();  
        
        // Begin logging changes
        //
        ChangeSummary changeSummary = quote.getChangeSummary();
        assertNotNull(changeSummary);
        assertFalse(changeSummary.isLogging());
        changeSummary.beginLogging();
      
        // Modify the data graph in various fun and interesting ways
        //
        quote.setString("symbol", "fbnt");
        quote.setString("companyName", "FlyByNightTechnology");
        quote.setBigDecimal("price", new BigDecimal("1000.9"));
        quote.setBigDecimal("open1", new BigDecimal("1000.9"));
        quote.setBigDecimal("high", new BigDecimal("1000.9"));
        quote.setBigDecimal("low", new BigDecimal("1000.9"));
        quote.setDouble("volume", 1000);
        quote.setDouble("change1", 1000);

        DataObject child = quote.createDataObject("quotes");
        child.setBigDecimal("price", new BigDecimal("2000.9"));

        changeSummary.endLogging();
        assertEquals(2, changeSummary.getChangedDataObjects().size());  // 2 DataObjects
        assertTrue(changeSummary.getChangedDataObjects().contains(quote));
        assertTrue(changeSummary.getChangedDataObjects().contains(child));
        assertFalse(changeSummary.isCreated(quote));
        assertTrue(changeSummary.isCreated(child));
      
        ChangeSummary.Setting ov = changeSummary.getOldValue(quote, quote.getType().getProperty("symbol"));
        assertNull(ov.getValue());
        assertEquals(1, quote.getList("quotes").size());
        assertEquals("fbnt", quote.getString("symbol"));
      
        quote.getChangeSummary().undoChanges();
        assertEquals(0, quote.getList("quotes").size());
        assertNull(quote.getString("symbol"));

    }
    
    /**
     * Verifies extended ChangeSummary methods <p/>
     * @see commonj.sdo.ChangeSummary#beginLogging()
     * @see commonj.sdo.ChangeSummary#endLogging()
     * @see commonj.sdo.ChangeSummary#getOldContainer(DataObject)
     * @see commonj.sdo.ChangeSummary#getOldContainmentProperty(DataObject)
     * @see commonj.sdo.ChangeSummary#getOldValue(DataObject,Property)
     * @see commonj.sdo.DataObject#getChangeSummary()
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=54">2.1 spec section 3.3.7</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.3.7")},
            testFor={@Operation(className="ChangeSummary",methodName="beginLogging",signature=""),
                     @Operation(className="ChangeSummary",methodName="endLogging",signature=""),
                     @Operation(className="ChangeSummary",methodName="getOldContainer",signature="DataObject"),                   
                     @Operation(className="ChangeSummary",methodName="getOldContainment",signature="DataObject"),                   
                     @Operation(className="ChangeSummary",methodName="getOldValue",signature="DataObject,Property"),                   
                     @Operation(className="DataObject",methodName="getChangeSummary",signature="")})    
    @Test
    public void testExtended() 
    {
        DataObject child = getScope().getDataFactory().create(getScope().getTypeHelper().getType(NS,CHILDNAME));
        DataObject parent1 = getScope().getDataFactory().create(getScope().getTypeHelper().getType(NS,PARENT1NAME));
        DataObject parent2 = getScope().getDataFactory().create(getScope().getTypeHelper().getType(NS,PARENT2NAME));
        DataObject root = getRootDataObject();
        
        Property pp2 = parent2.getType().getProperty("childProp");
        
        ChangeSummary cs = root.getChangeSummary();
        cs.beginLogging();
        root.set("parentProp1",parent1);
        root.set("parentProp2",parent2);
        assertNull(cs.getOldContainer(child));
        cs.endLogging();
        parent2.set("childProp",child);
        child.setString("strProperty", "Original Value");
        cs.beginLogging();
        assertSame(cs.getOldContainer(child),parent2);
        assertSame(cs.getOldContainmentProperty(child),pp2);
        parent1.set("childProp",child);
        assertSame(cs.getOldContainer(child),parent2);    
        assertSame(cs.getOldContainmentProperty(child),pp2);
        child.setString("strProperty", "New Value");
        ChangeSummary.Setting oldValue = cs.getOldValue(child, child.getType().getProperty("strProperty"));
        assertTrue("Original Value".equals((String)oldValue.getValue()));
        cs.endLogging();
    }

    /**
     * Verifies DO state methods in ChangeSummary <p/>
     * @see commonj.sdo.ChangeSummary#beginLogging()
     * @see commonj.sdo.ChangeSummary#endLogging()
     * @see commonj.sdo.ChangeSummary#isCreated(DataObject)
     * @see commonj.sdo.ChangeSummary#isDeleted(DataObject)
     * @see commonj.sdo.ChangeSummary#isModified(DataObject)
     * @see commonj.sdo.DataObject#getChangeSummary()
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=54">2.1 spec section 3.3.7</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.3.7")},
            testFor={@Operation(className="ChangeSummary",methodName="beginLogging",signature=""),
                     @Operation(className="ChangeSummary",methodName="endLogging",signature=""),
                     @Operation(className="ChangeSummary",methodName="isCreated",signature="DataObject"),
                     @Operation(className="ChangeSummary",methodName="isModified",signature="DataObject"),
                     @Operation(className="ChangeSummary",methodName="isDeleted",signature="DataObject"),
                     @Operation(className="DataObject",methodName="getChangeSummary",signature="")})    
    @Test
    public void testStateMethods() 
    {
        DataObject rootDO = getRootDataObject();
        DataObject parentDO = getScope().getDataFactory().create(NS,PARENT1NAME);
        
        rootDO.set("parentProp1",parentDO);
    
        DataObject child1 = parentDO.createDataObject("childProp");
        child1.setString("strProperty", "Original Value");
      
        ChangeSummary cs = parentDO.getChangeSummary();
        cs.beginLogging();
        DataObject child2 = parentDO.createDataObject("childProp2");
      
        assertFalse(cs.isCreated(child1));
        assertTrue(cs.isCreated(child2));
        cs.endLogging();
      
        child2.setString("strProperty", "Original Value");
        cs = parentDO.getChangeSummary();
        cs.beginLogging();
        child1.setString("strProperty", "New Value");
        cs.endLogging();
      
        assertTrue(cs.isModified(child1));
        assertFalse(cs.isModified(child2));
      
        cs = parentDO.getChangeSummary();
        cs.beginLogging();
        child2.delete();
        cs.endLogging();
      
        assertFalse(cs.isDeleted(child1));
        assertTrue(cs.isDeleted(child2));
    }
    
    /**
     * Verifies ChangeSummary.Setting methods using a DO-based ChangeSummary <p/>
     * @see commonj.sdo.ChangeSummary#beginLogging()
     * @see commonj.sdo.ChangeSummary#endLogging()
     * @see commonj.sdo.ChangeSummary.Setting#getValue()
     * @see commonj.sdo.ChangeSummary.Setting#getProperty()
     * @see commonj.sdo.ChangeSummary.Setting#isSet()
     * @see commonj.sdo.ChangeSummary#getOldValue(DataObject,Property)
     * @see commonj.sdo.DataObject#getChangeSummary()
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=54">2.1 spec section 3.3.7</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.3.7")},
            testFor={@Operation(className="ChangeSummary",methodName="beginLogging",signature=""),
                     @Operation(className="ChangeSummary",methodName="endLogging",signature=""),
                     @Operation(className="ChangeSummary.Setting",methodName="getValue",signature=""),
                     @Operation(className="ChangeSummary.Setting",methodName="getProperty",signature=""),
                     @Operation(className="ChangeSummary.Setting",methodName="isSet",signature=""),
                     @Operation(className="ChangeSummary",methodName="getOldValue",signature="DataObject,Property"),                   
                     @Operation(className="DataObject",methodName="getChangeSummary",signature="")})    
    @Test
    public void testSettingMethodsBody() 
    {
        String propName = "symbol";
        String oldValue = "Old Value";
        String newValue = "New Value";
      
        DataObject testObj = getRootDataObject();
        testObj.setString(propName, oldValue);
        ChangeSummary cs = testObj.getChangeSummary();
        cs.beginLogging();
        ChangeSummary.Setting ov = cs.getOldValue(testObj, testObj.getType().getProperty(propName));
        assertNull(ov);
        testObj.setString(propName, newValue);
        ov = cs.getOldValue(testObj, testObj.getType().getProperty(propName));
        assertNotNull(ov);
        assertTrue(oldValue.equals((String)ov.getValue()));
        assertEquals(testObj.getType().getProperty(propName),ov.getProperty());
        testObj.unset(propName);
        assertFalse(testObj.isSet(propName));
        assertTrue(ov.isSet());
        cs.endLogging();
    }
    
    /**
     * Verifies ChangeSummary tracking of sequences using a DO-based ChangeSummary <p/>
     * @see commonj.sdo.ChangeSummary#beginLogging()
     * @see commonj.sdo.ChangeSummary#endLogging()
     * @see commonj.sdo.DataObject#getChangeSummary()
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=54">2.1 spec section 3.3.7</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.3.7")},
            testFor={@Operation(className="ChangeSummary",methodName="beginLogging",signature=""),
                     @Operation(className="ChangeSummary",methodName="endLogging",signature=""),
                     @Operation(className="ChangeSummary",methodName="getOldSequence",signature="DataObject"),
                     @Operation(className="DataObject",methodName="getChangeSummary",signature="")})    
    @Test
    public void testSequenceBody() 
    {
        DataObject root = getRootDataObject();
        DataObject dobj = root.createDataObject("seqProp");
        
        Sequence seq = dobj.getSequence();
        seq.addText("one");
        seq.add("price", new BigDecimal("2000.9"));
        seq.addText("three");
        
        //System.out.println(getScope().getXMLHelper().save(dobj, "dummyURI", "dummy"));
        int seqSize = dobj.getSequence().size();
        assertEquals(3, seqSize);
        
        ChangeSummary cs = dobj.getChangeSummary();
        cs.beginLogging();
        seq.remove(2);
        seq.addText("four");
        seq.add("price", new BigDecimal("3000.9"));
        seq.remove(2);
        seq.add(0, "price", new BigDecimal("1000.9"));
        cs.endLogging();
        
        //System.out.println(getScope().getXMLHelper().save(root, "dummyURI", "dummy"));
        assertEquals(4, dobj.getSequence().size());
        assertEquals(new BigDecimal("1000.9"), dobj.getSequence().getValue(0));
        assertEquals("one", dobj.getSequence().getValue(1));
        assertEquals(new BigDecimal("2000.9"), dobj.getSequence().getValue(2));
        assertEquals(new BigDecimal("3000.9"), dobj.getSequence().getValue(3));
        
        assertEquals(seqSize, cs.getOldSequence(dobj).size());
        
        cs.undoChanges();
        
        //System.out.println(getScope().getXMLHelper().save(dobj, "dummyURI", "dummy"));
        assertEquals(seqSize, dobj.getSequence().size());
        assertEquals("one", dobj.getSequence().getValue(0));
        assertEquals(new BigDecimal("2000.9"), dobj.getSequence().getValue(1));
        assertEquals("three", dobj.getSequence().getValue(2));
    }
    
    
    /**
     * Verifies ChangeSummary tracking of sequences using a DO-based ChangeSummary <p/>
     * @see commonj.sdo.ChangeSummary#beginLogging()
     * @see commonj.sdo.ChangeSummary#endLogging()
     * @see commonj.sdo.ChangeSummary#undoChanges()
     * @see commonj.sdo.DataObject#getChangeSummary()
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=54">2.1 spec section 3.3.7</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.3.7")},
            testFor={@Operation(className="ChangeSummary",methodName="beginLogging",signature=""),
                     @Operation(className="ChangeSummary",methodName="endLogging",signature=""),
                     @Operation(className="ChangeSummary",methodName="undoChanges",signature=""),
                     @Operation(className="DataObject",methodName="getChangeSummary",signature="")})    
    @Test
    public void testDelete() throws Exception 
    {
        DataObject quote = getRootDataObject();

        ChangeSummary changeSummary = quote.getChangeSummary();
        assertNotNull(changeSummary);
        assertFalse(changeSummary.isLogging());
        assertTrue(changeSummary.getRootObject() == quote);

        quote.setString("symbol", "NONE");

        quote.setString("symbol", "fbnt");
        quote.setString("companyName", "FlyByNightTechnology");

        quote.setBigDecimal("open1", new BigDecimal("1000.9"));
        quote.setBigDecimal("price", new BigDecimal("1000.9"));

        quote.setBigDecimal("high", new BigDecimal("1000.9"));
        quote.setBigDecimal("low", new BigDecimal("1000.9"));
        quote.setDouble("volume", 1000);
        quote.setDouble("change1", 1000);

        DataObject child = quote.createDataObject("quotes");
        child.setBigDecimal("price", new BigDecimal("2000.9"));

        DataObject child2 = getScope().getDataFactory().create("http://www.example.com/cstest", "Quote");
        child2.setString("symbol", "c2");
        DataObject child3 = getScope().getDataFactory().create("http://www.example.com/cstest", "Quote");
        child3.setString("symbol", "c3");
        child2.getList("quotes").add(0, child3);

        quote.getList("quotes").add(0, child2);
        child.setBigDecimal("open1", new BigDecimal("1000.9"));
        
        changeSummary.beginLogging();
        assertTrue(changeSummary.isLogging());

        assertEquals("c2", quote.getString("quotes[1]/symbol"));
        assertEquals("c3", quote.getString("quotes[1]/quotes[1]/symbol"));

        child2.delete();

        quote.unset("high");
        quote.unset("price");

        changeSummary.endLogging();
        assertFalse(changeSummary.isLogging());

        List currList = quote.getList("quotes");
        assertEquals(1, currList.size());
        assertTrue(currList.contains(child));
        assertFalse(currList.contains(child2));
        assertTrue(changeSummary.isDeleted(child2));

        assertFalse(quote.isSet("high"));
        assertFalse(quote.isSet("price"));
        // The three items that have changes are 'quote', 'child2' and 'child3'.
        // Quote has two values unset, child2 is deleted and since child3 is contained by
        // child2, it too, is deleted.
        assertEquals(3, changeSummary.getChangedDataObjects().size());
        assertTrue(changeSummary.getChangedDataObjects().contains(quote));

        ChangeSummary.Setting ov = 
        changeSummary.getOldValue(quote, quote.getType().getProperty("quotes"));
        List ovList = (List)ov.getValue();
        assertEquals(2, ovList.size());
        assertEquals("c2", ((DataObject)ovList.get(0)).getString("symbol"));

        quote.getChangeSummary().undoChanges();

        currList = quote.getList("quotes");
        assertEquals(2, currList.size());
        assertEquals("c2", quote.getString("quotes[1]/symbol"));
        assertEquals("c3", quote.getString("quotes[1]/quotes[1]/symbol"));

        assertTrue(quote.isSet("high"));
        assertEquals(new BigDecimal("1000.9"), quote.get("high"));

        assertTrue(quote.isSet("price"));
        assertEquals(new BigDecimal("1000.9"), quote.get("price"));
    }
    
    
    @Test
    public void testJavaSerialization() throws Exception 
    {
        serializationBody(true);
    }
    
    @Test
    public void testXMLSerialization() throws Exception 
    {
        serializationBody(false);
    }
    
    /**
     * Verifies Serialization of ChangeSummary object using both
     * Java Serialization and XMLHelper Serialization. <p/>
     * 
     * @see commonj.sdo.ChangeSummary#beginLogging()
     * @see commonj.sdo.ChangeSummary#endLogging()
     * @see commonj.sdo.ChangeSummary#undoChanges()
     * @see commonj.sdo.DataObject#getChangeSummary()
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=54">2.1 spec section 3.3.7</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.3.7")},
            testFor={@Operation(className="ChangeSummary",methodName="beginLogging",signature=""),
                     @Operation(className="ChangeSummary",methodName="endLogging",signature=""),
                     @Operation(className="ChangeSummary",methodName="undoChanges",signature=""),
                     @Operation(className="ChangeSummary.Setting",methodName="isSet",signature=""),
                     @Operation(className="ChangeSummary",methodName="getOldValue",signature="DataObject,Property"),                   
                     @Operation(className="DataObject",methodName="getChangeSummary",signature="")})     
    private void serializationBody(boolean javaSerialization ) throws Exception 
    {
        DataObject quote = getRootDataObject();
        
        ChangeSummary changeSummary = quote.getChangeSummary();
        //ChangeSummary changeSummary = (ChangeSummary)quote.get("changes");
        assertNotNull(changeSummary);
        assertFalse(changeSummary.isLogging());
        assertTrue(changeSummary.getRootObject() == quote);

        quote.setString("symbol", "NONE");
        quote.setString("companyName", "FlyByNightTechnology");

        quote.setBigDecimal("open1", new BigDecimal("999.9"));
        quote.setBigDecimal("price", new BigDecimal("1001.9"));

        quote.setBigDecimal("high", new BigDecimal("1003.9"));
        quote.setBigDecimal("low", new BigDecimal("999.9"));
        
        // begin logging now
        changeSummary.beginLogging();
        assertTrue(changeSummary.isLogging());

        quote.setString("symbol", "fbnt");
        
        DataObject child = quote.createDataObject("quotes");        
        child.setBigDecimal("price", new BigDecimal("2000.9"));
        
        DataObject child2 = getScope().getDataFactory().create("http://www.example.com/cstest", "Quote");
        quote.getList("quotes").add(0, child2);
        child2.setBigDecimal("price", new BigDecimal("2001.9"));
        
        quote.setDouble("volume", 1000);
        quote.setBigDecimal("price", new BigDecimal("1002.9"));
        
        child.setBigDecimal("open1", new BigDecimal("1000.9"));
        
        // Serialize/deserialize graph before continuing
        if( javaSerialization )
        {
            byte[] bytes = TCKUtil.serializeDataObjectJava(getTestHelper(), quote,getScope());
            quote = TCKUtil.deserializeDataObjectJava(getTestHelper(), bytes, getScope());
        }
        else
        {
            byte[] bytes = TCKUtil.serializeDataObjectXML(quote, "http://www.example.com/cstest", "Root", getScope());
            quote = TCKUtil.deserializeDataObjectXML(bytes, getScope());
        }
        
        // get CS object, should still be in logging state
        changeSummary = quote.getChangeSummary();
        assertTrue(changeSummary.isLogging());
        
        quote.setBigDecimal("high", new BigDecimal("1010.9"));   // This value should be undone below
        
        changeSummary.endLogging();
        
        assertFalse(changeSummary.isLogging());
        assertEquals(new BigDecimal("1002.9"), quote.getBigDecimal("price"));
        assertEquals(3, changeSummary.getChangedDataObjects().size()); // 3 DataObjects
        assertTrue(changeSummary.getChangedDataObjects().contains(quote));
        assertTrue(changeSummary.getChangedDataObjects().contains(quote.getDataObject("quotes[2]")));
        assertFalse(changeSummary.isCreated(quote));
        assertTrue(changeSummary.isCreated(quote.getDataObject("quotes[2]")));
        assertTrue(changeSummary.isCreated(quote.getDataObject("quotes[1]")));

        ChangeSummary.Setting ov = 
        changeSummary.getOldValue(quote, quote.getType().getProperty("symbol"));
        assertEquals("NONE", ov.getValue());
        assertEquals("fbnt", quote.getString("symbol"));

        ChangeSummary.Setting ov2 = 
        changeSummary.getOldValue(quote, quote.getType().getProperty("companyName"));
        assertNull(ov2);
        assertEquals(quote.getString("companyName"), "FlyByNightTechnology");

        assertEquals(2, quote.getList("quotes").size());
        assertEquals("fbnt", quote.getString("symbol"));
        
        // now undo changes and check for original state
        quote.getChangeSummary().undoChanges();

        assertFalse(quote.isSet("volume"));  
        assertEquals(new BigDecimal("999.9"), quote.getBigDecimal("open1"));
        assertEquals(new BigDecimal("1001.9"), quote.getBigDecimal("price"));
        //TODO Fix SDOX bug    assertEquals(new BigDecimal("1003.9"), quote.getBigDecimal("high"));
        assertEquals(new BigDecimal("999.9"), quote.getBigDecimal("low"));
        assertEquals(0, quote.getList("quotes").size());
        assertEquals("NONE", quote.getString("symbol"));
    }
}
