/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 *  
 *  $Rev: 556085 $  $Date: 2008/01/08 02:33:49 $
 */

package com.ibm.sdo.tck.tests.core;

import java.io.InputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;


import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import com.ibm.sdo.tck.framework.TCKTestBase;
import com.ibm.sdo.tck.framework.annotations.Operation;
import com.ibm.sdo.tck.framework.annotations.TCKTest;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertNotNull;

import commonj.sdo.DataObject;
import commonj.sdo.helper.HelperContext;
import commonj.sdo.helper.XMLDocument;

/**
 * Test cases for the DataObject SDO implementation.
 */
@TCKTest(testFor={@Operation(className="DataObject", methodName="getList", signature="String"),
                  @Operation(className="DataObject", methodName="getList", signature="int"),
                  @Operation(className="DataObject", methodName="getList", signature="Property"),
                  @Operation(className="DataObject", methodName="isSet", signature="Property"),
                  @Operation(className="DataObject", methodName="set", signature="String,Object"),
                  @Operation(className="DataObject", methodName="set", signature="Property,Object"),
                  @Operation(className="DataObject", methodName="setList", signature="int,List"),
                  @Operation(className="DataObject", methodName="getString", signature="String"),
                  @Operation(className="DataObject", methodName="setString", signature="String,String"),
                  @Operation(className="DataObject", methodName="get", signature="int"),
                  @Operation(className="DataObject", methodName="unset", signature="String")})
public class DataObjectListTest extends TCKTestBase {

    /**
     * Number of iterations for data type tests (sets and retrieves random
     * values for each data type)
     */
    
    public static final int DATA_TYPE_ITERATIONS = 10;
    private static int unique = 0;
    private final String NS       = "http://www.sample.org/dotest";
    private final String TYPENAME = "DataObjectTestType";
    private final String CHILDTYPENAME = "ChildType";


    @Before
    public void setUp() throws Exception {
      super.setUp();
      final URL url = getClass().getResource("/DataObjectListTest/dataObjectListTest.xsd");
      final InputStream inputStream = url.openStream();
      getScope().getXSDHelper().define(inputStream, url.toString());
      inputStream.close();      
    }
    
    @After
    public void tearDown() throws Exception {
      super.tearDown();
    }

    
    /**
     * Test List manipulation methods
     *
     */
    @Test
    public void testGetMethods() {
        
        // create a DataObject that uses this type
        DataObject testDO = getScope().getDataFactory().create(NS,TYPENAME);

        List l = new ArrayList();
        l.add(1); l.add(2); l.add(3);
        testDO.setList("intValues", l);
                
        // test getList(String)
        assertTrue( java.util.Arrays.equals(l.toArray(),testDO.getList("intValues").toArray()));
        
        // test getList(int)
        assertTrue( java.util.Arrays.equals(l.toArray(),testDO.getList(testDO.getType().getProperties().indexOf(testDO.getType().getProperty("intValues"))).toArray()));
        
        // test getList(Property)
        assertTrue( java.util.Arrays.equals(l.toArray(),testDO.getList(testDO.getType().getProperty("intValues")).toArray()));

    }    
    

    /**
     * See SDO Spec 2.1.0 3.1.13 DataObject Accessor Exceptions General
     * get(String path) returns null if path does not exist
     */
    @Test
    public void testGetInvalidInstanceProperty() {
        DataObject testObj = getScope().getDataFactory().create(getScope().getTypeHelper().getType(NS,TYPENAME));
        assertNotNull(testObj);
        commonj.sdo.Property property = testObj.getInstanceProperty("nonexistentProperty");
        assertNull("testObj.getInstanceProperty(\"nonexistentProperty\") should return null", property);
    }

    String preXML = "<?xml version=\"1.0\" encoding=\"UTF-8\"?> " +
                    "<dot:DataObjectTest xmlns:dot=\""+NS+"\"> ";
    String postXML = " </dot:DataObjectTest>";
    private XMLDocument loadDocFromString(String strXML) {
        //createTestObjectTypes(URI,parentTypeName,childTypeName);
        return getScope().getXMLHelper().load(preXML+strXML+postXML);
    }

    private List getTestList(XMLDocument doc) {
        return doc.getRootObject().getDataObject("childContain").getList("child");
    }


    @Test
    public void testClear() {
        XMLDocument doc = loadDocFromString("<childContain><child/><child/></childContain>");
        List listTest = getTestList(doc);
        assertNotNull(listTest);

        assertEquals(2, listTest.size());
        listTest.clear();
        assertEquals(0, listTest.size());
    }

    @Test
    public void testIsEmpty() {
        XMLDocument doc = loadDocFromString("<childContain><child/><child/></childContain>");
        List listTest = getTestList(doc);
        assertNotNull(listTest);

        assertEquals(2, listTest.size());
        assertFalse(listTest.isEmpty());

        XMLDocument doc2 = loadDocFromString("<childContain></childContain>");
        List listTest2 = getTestList(doc2);
        assertNotNull(listTest2);
        assertEquals(0, listTest2.size());
        assertTrue(listTest2.isEmpty());
    }

    @Test
    public void testToArray() {
        String xmldoc =
            "<childContain><child name=\"name1\"/><child name=\"name2\"/></childContain>";
        // "<childContain><child nameTest=\"name1\"/><child
        // nameTest=\"name2\"/></childContain>"
        XMLDocument doc = loadDocFromString(xmldoc);
        List listTest = getTestList(doc);
        assertNotNull(listTest);

        Object[] objArray = listTest.toArray();

        assertNotNull(objArray);
        assertEquals(2, objArray.length);
        // assertEquals("name1",((DataObject)objArray[0]).getString("nameTest"));
        assertEquals("name2", ((DataObject)objArray[1]).getString("name"));
    }

    /**
     * Test that it is possible to add a null value to a List retrieved by calling DataObject.getList()
     *
     */
    @Test
    public void testAddNullToList() {
        HelperContext helperContext = getScope();
        XMLDocument doc = loadDocFromString("<childContain><child name=\"aname1\"/><child name=\"aname2\"/></childContain>");
        List listTest = getTestList(doc);
        assertNotNull(listTest);
        assertEquals(2, listTest.size());
        listTest.add( null );
        assertEquals(3, listTest.size());
        String xml = helperContext.getXMLHelper().save( doc.getRootObject(), doc.getRootElementURI(), doc.getRootElementName() );
        assertTrue( xml.indexOf("xsi:nil=\"true\"") > 0 );
        
    }

    @Test
    public void testAddObject() {
        XMLDocument doc = loadDocFromString("<childContain></childContain>");
        List listTest = getTestList(doc);
        assertNotNull(listTest);
        assertEquals(0, listTest.size());

        DataObject doInsert = getScope().getDataFactory().create(NS, CHILDTYPENAME);
        assertNotNull(doInsert);
        doInsert.setString("name", "aname");
        listTest.add(0, doInsert);

        assertEquals(1, listTest.size());
        DataObject doRes = (DataObject)listTest.get(0);
        assertEquals("aname", doRes.getString("name"));
    }

    @Test
    public void testAddAll() {
        XMLDocument doc = loadDocFromString("<childContain><child/><child/></childContain>");
        List listTest = getTestList(doc);
        assertNotNull(listTest);
        assertEquals(2, listTest.size());

        DataObject doInsert1 = getScope().getDataFactory().create(NS, CHILDTYPENAME);
        assertNotNull(doInsert1);
        doInsert1.setString("name", "aname1");

        DataObject doInsert2 = getScope().getDataFactory().create(NS, CHILDTYPENAME);
        assertNotNull(doInsert2);
        doInsert2.setString("name", "aname2");

        ArrayList arraylist2 = new ArrayList();
        arraylist2.add(doInsert1);
        arraylist2.add(doInsert2);

        listTest.addAll(0, arraylist2);

        assertEquals(4, listTest.size());
        DataObject doRes = (DataObject)listTest.get(0);
        assertEquals("aname1", doRes.getString("name"));
    }

    @Test
    public void testAddAllCollection() {
        XMLDocument doc = loadDocFromString("<childContain><child/><child/></childContain>");
        List listTest = getTestList(doc);
        assertNotNull(listTest);
        assertEquals(2, listTest.size());

        DataObject doInsert1 = getScope().getDataFactory().create(NS, CHILDTYPENAME);
        assertNotNull(doInsert1);
        doInsert1.setString("name", "aname1");

        DataObject doInsert2 = getScope().getDataFactory().create(NS, CHILDTYPENAME);
        assertNotNull(doInsert2);
        doInsert2.setString("name", "aname2");

        ArrayList arraylist2 = new ArrayList();
        arraylist2.add(doInsert1);
        arraylist2.add(doInsert2);

        boolean bRes = listTest.addAll(arraylist2);

        assertTrue(bRes);
        assertEquals(4, listTest.size());
        DataObject doRes = (DataObject)listTest.get(2);
        assertEquals("aname1", doRes.getString("name"));
    }

    @Test
    public void testAddAllCollectionEmpty() {
        XMLDocument doc = loadDocFromString("<childContain><child/><child/></childContain>");
        List listTest = getTestList(doc);
        assertNotNull(listTest);
        assertEquals(2, listTest.size());

        ArrayList arraylist2 = new ArrayList();

        boolean bRes = listTest.addAll(arraylist2);

        assertFalse(bRes);
        assertEquals(2, listTest.size());
    }

    @Test
    public void testIndexOf() {
        XMLDocument doc = loadDocFromString("<childContain><child/><child/></childContain>");
        List listTest = getTestList(doc);
        assertNotNull(listTest);
        assertEquals(2, listTest.size());

        DataObject do1 = (DataObject)listTest.get(1);
        assertNotNull(do1);

        int nIdx = listTest.indexOf(do1);

        assertEquals(1, nIdx);
    }

    @Test
    public void testLastIndexOf() {
        XMLDocument doc = loadDocFromString("<childContain><child/><child/></childContain>");
        List listTest = getTestList(doc);
        assertNotNull(listTest);
        assertEquals(2, listTest.size());

        DataObject do1 = (DataObject)listTest.get(1);
        assertNotNull(do1);

        int nIdx = listTest.lastIndexOf(do1);

        assertEquals(1, nIdx);
    }

    @Test
    public void testContains() {
        XMLDocument doc = loadDocFromString("<childContain><child/><child/></childContain>");
        List listTest = getTestList(doc);
        assertNotNull(listTest);
        assertEquals(2, listTest.size());

        DataObject do1 = (DataObject)listTest.get(1);
        assertNotNull(do1);

        boolean bRes = listTest.contains(do1);

        assertTrue(bRes);

        DataObject doTest = getScope().getDataFactory().create(NS, CHILDTYPENAME);
        assertNotNull(doTest);

        bRes = listTest.contains(doTest);

        assertFalse(bRes);
    }

    @Test
    public void testRemoveObject() {
        XMLDocument doc = loadDocFromString("<childContain><child/><child/></childContain>");
        List listTest = getTestList(doc);
        assertNotNull(listTest);
        assertEquals(2, listTest.size());

        DataObject do1 = (DataObject)listTest.get(1);
        assertNotNull(do1);

        boolean bRes = listTest.remove(do1);
        assertTrue(bRes);
        assertEquals(1, listTest.size());

        DataObject doTest = getScope().getDataFactory().create(NS, CHILDTYPENAME);
        assertNotNull(doTest);

        bRes = listTest.remove(doTest);
        assertFalse(bRes);
        assertEquals(1, listTest.size());
    }

    @Test
    public void testContainsAll() {
        XMLDocument doc = loadDocFromString("<childContain><child/><child/></childContain>");
        List listTest = getTestList(doc);
        assertNotNull(listTest);
        assertEquals(2, listTest.size());

        DataObject do1 = (DataObject)listTest.get(1);
        assertNotNull(do1);

        ArrayList arraylist1 = new ArrayList();
        arraylist1.add(do1);

        boolean bRes = listTest.containsAll(arraylist1);

        assertTrue(bRes);

        DataObject doTest = getScope().getDataFactory().create(NS, CHILDTYPENAME);
        assertNotNull(doTest);
        arraylist1.add(doTest);

        bRes = listTest.containsAll(arraylist1);

        assertFalse(bRes);
    }

    @Test
    public void testRemoveAll() {
        XMLDocument doc = loadDocFromString("<childContain><child/><child/></childContain>");
        List listTest = getTestList(doc);
        assertNotNull(listTest);
        assertEquals(2, listTest.size());

        DataObject do1 = (DataObject)listTest.get(1);
        assertNotNull(do1);

        ArrayList arraylist1 = new ArrayList();
        arraylist1.add(do1);

        boolean bRes = listTest.removeAll(arraylist1);

        assertTrue(bRes);
        assertEquals(1, listTest.size());
    }

    @Test
    public void testRetainAll() {
        XMLDocument doc = loadDocFromString("<childContain><child/><child/></childContain>");
        List listTest = getTestList(doc);
        assertNotNull(listTest);
        assertEquals(2, listTest.size());

        DataObject do1 = (DataObject)listTest.get(1);
        assertNotNull(do1);

        ArrayList arraylist1 = new ArrayList();
        arraylist1.add(do1);

        boolean bRes = listTest.retainAll(arraylist1);

        assertTrue(bRes);
        assertEquals(1, listTest.size());
    }

    @Test
    public void testIterator() {
        XMLDocument doc = loadDocFromString("<childContain><child/><child/></childContain>");
        List listTest = getTestList(doc);
        assertNotNull(listTest);
        assertEquals(2, listTest.size());

        Iterator iterRes = listTest.iterator();

        assertNotNull(iterRes);
        DataObject doRes = (DataObject)iterRes.next();
        assertNotNull(doRes);
    }

    @Test
    public void testSubList() {
        XMLDocument doc = loadDocFromString("<childContain><child/><child/><child/></childContain>");
        List listTest = getTestList(doc);
        assertNotNull(listTest);
        assertEquals(3, listTest.size());

        List listRes = listTest.subList(0, 1);

        assertNotNull(listRes);
        assertEquals("listTest.subList(0,1)", 1, listRes.size());
        DataObject doRes = (DataObject)listRes.get(0);
        assertNotNull(doRes);
    }

    @Test
    public void testListIterator() {
        XMLDocument doc = loadDocFromString("<childContain><child/><child/></childContain>");
        List listTest = getTestList(doc);
        assertNotNull(listTest);
        assertEquals(2, listTest.size());

        ListIterator iterRes = listTest.listIterator();

        assertNotNull(iterRes);
        DataObject doRes = (DataObject)iterRes.next();
        assertNotNull(doRes);
    }

    @Test
    public void testListIteratorIndex() {
        XMLDocument doc = loadDocFromString("<childContain><child/><child/></childContain>");
        List listTest = getTestList(doc);
        assertNotNull(listTest);
        assertEquals(2, listTest.size());

        ListIterator iterRes = listTest.listIterator(0);

        assertNotNull(iterRes);
        DataObject doRes = (DataObject)iterRes.next();
        assertNotNull(doRes);
    }

    @Test
    public void testSetIndexObject() {
        XMLDocument doc =
            loadDocFromString("<childContain><child name=\"aname1\"/><child name=\"aname2\"/></childContain>");
        List listTest = getTestList(doc);
        assertNotNull(listTest);
        assertEquals(2, listTest.size());

        DataObject doTest = getScope().getDataFactory().create(NS, CHILDTYPENAME);
        assertNotNull(doTest);
        doTest.setString("name", "aname3");

        Object objRes = listTest.set(0, doTest);

        assertNotNull(objRes);
        assertTrue(objRes instanceof DataObject);
    }

    @Test
    public void testToArrayObject() {
        XMLDocument doc =
            loadDocFromString("<childContain><child name=\"aname1\"/><child name=\"aname2\"/></childContain>");
        List listTest = getTestList(doc);
        assertNotNull(listTest);
        assertEquals(2, listTest.size());

        DataObject do1 = (DataObject)listTest.get(1);
        assertNotNull(do1);

        ArrayList arraylist2 = new ArrayList();

        Object[] objRes = listTest.toArray(arraylist2.toArray());

        assertNotNull(objRes);
        assertEquals(2, objRes.length);
    }

    @Test
    public void testToString() {
        XMLDocument doc =
            loadDocFromString("<childContain><child name=\"aname1\"/><child name=\"aname2\"/></childContain>");
        List listTest = getTestList(doc);
        assertNotNull(listTest);
        assertEquals(2, listTest.size());

        String strRes = listTest.toString();

        assertNotNull(strRes);
        assertTrue(strRes.length() > 0);
        // assertEquals("List: [DataObjectImpl[ <sss> ]]",strRes);
    }
    
}
