/**
 *
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 */
package com.ibm.sdo.tck.tests.dynamictypes;

import commonj.sdo.*;
import commonj.sdo.helper.*;
import org.junit.After;
import static org.junit.Assert.*;
import org.junit.Before;
import org.junit.Test;

import com.ibm.sdo.tck.framework.TCKTestBase;
import com.ibm.sdo.tck.framework.annotations.Operation;
import com.ibm.sdo.tck.framework.annotations.TCKTest;
import com.ibm.sdo.tck.util.XMLCompareUtil;

import java.io.*;
import java.math.BigDecimal;
import java.net.URL;


/**
 * Tests various concepts against the XMLHelper, including ...
 * <ul>
 * <li> Round tripping of serialization
 * <li> Definition of types using dynamic API
 * <li> Handling of sequenced types
 * <li> Handling of open sequenced types
 * </ul>
 * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=45">2.1 spec section 3.11</a>
 */
@TCKTest(testFor={@Operation(className = "HelperContext", methodName = "getXMLHelper", signature = ""),
                  @Operation(className = "DataFactory", methodName = "create", signature = "Type"),
                  @Operation(className = "DataFactory", methodName = "create", signature = "String,String"),
                  @Operation(className = "DataObject", methodName = "createDataObject", signature = "String"),
                  @Operation(className = "DataObject", methodName = "createDataObject", signature = "Property"),
                  @Operation(className = "DataObject", methodName = "delete", signature = ""),
                  @Operation(className = "DataObject", methodName = "getList", signature = "Property"),
                  @Operation(className = "DataObject", methodName = "getChangeSummary", signature = ""),
                  @Operation(className = "DataObject", methodName = "getSequence", signature = ""),
                  @Operation(className = "DataObject", methodName = "get", signature = "String"),
                  @Operation(className = "DataObject", methodName = "set", signature = "String,Object"),
                  @Operation(className = "DataObject", methodName = "setString", signature = "String,String"),
                  @Operation(className = "DataObject", methodName = "setString", signature = "Property,String",precondition="Property is open content property"),
                  @Operation(className = "DataObject", methodName = "setBigDecimal", signature = "String,BigDecimal"),
                  @Operation(className = "DataObject", methodName = "setBoolean", signature = "String,boolean"),
                  @Operation(className = "DataObject", methodName = "setString", signature = "String,String"),
                  @Operation(className = "DataObject", methodName = "setBoolean", signature = "Property,boolean"),
                  @Operation(className = "DataObject", methodName = "setBigDecimal", signature = "Property,BigDecimal"),
                  @Operation(className = "DataObject", methodName = "setInt", signature = "String,int"),
                  @Operation(className = "DataObject", methodName = "getType", signature = ""),
                  @Operation(className = "Property", methodName = "getType", signature = ""),
                  @Operation(className = "Sequence", methodName = "addText", signature = "String"),
                  @Operation(className = "TypeHelper", methodName = "getType", signature = "String,String"),
                  @Operation(className = "Type", methodName = "getProperty", signature = "String"),
                  @Operation(className = "Type", methodName = "isSequenced", signature = ""),
                  @Operation(className = "TypeHelper", methodName = "defineOpenContentProperty", signature = "String,DataObject"),
                  @Operation(className = "TypeHelper", methodName = "define", signature = "DataObject"),
                  @Operation(className = "TypeHelper", methodName = "getType", signature = "String,String"),
                  @Operation(className = "TypeHelper", methodName = "getOpenContentProperty", signature = "String,String"),
                  @Operation(className = "XMLDocument", methodName = "getRootObject", signature = ""),
                  @Operation(className = "XSDHelper", methodName = "getGlobalProperty", signature = "String,String,boolean"),
                  @Operation(className = "XMLHelper", methodName = "save", signature = "DataObject,String,String,OutputStream"),
                  @Operation(className = "XMLHelper", methodName = "load", signature = "InputStream")})
public class TypeHelperDefineTest extends TCKTestBase {
    private static final String CUSTOMER1_XML = "/TypeHelperDefineTest/customer1.xml";
    private static final String CUSTOMER2_XML = "/TypeHelperDefineTest/customer2.xml";
    private static final String MIXED_XML = "/TypeHelperDefineTest/mixed2.xml";
    private static final String MIXEDOPEN_XML = "/TypeHelperDefineTest/mixedopen.xml";
    private static final String OPEN_TYPE_NAMESPACE = "http://www.example.com/open";
    private static final String OPEN_XML = "/TypeHelperDefineTest/openContentProperty.xml";

    /**
     * Verify how the XMLHelper handles open content.<br/>
     * Ensures that the open content is serialized properly.
     * Uses global properties from the XSD, and open content properties
     * defined by the dynamic API.<br/>
     * 
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=38">2.1 spec section 3.6.6</a>
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=41">2.1 spec section 3.8.3</a>
     */
    @Test
    public void testDefineOpenContentProperty() throws IOException {
        TypeHelper typeHelper;
        XMLHelper xmlHelper;
        DataFactory dataFactory;

        typeHelper = getScope().getTypeHelper();
        dataFactory = getScope().getDataFactory();
        xmlHelper = getScope().getXMLHelper();

        Type quoteType = typeHelper.getType(OPEN_TYPE_NAMESPACE, "OpenQuote2");
        DataObject quote = dataFactory.create(quoteType);

        assertNotNull( quote );

        quote.setString("symbol", "s1");

        Property companyProperty = typeHelper.getOpenContentProperty(OPEN_TYPE_NAMESPACE, "company");
        DataObject company = quote.createDataObject(companyProperty);
        company.setString("name", "FlyByNightTechnology");

        Property priceProperty = typeHelper.getOpenContentProperty(OPEN_TYPE_NAMESPACE, "price");
        quote.getList(priceProperty).add(new BigDecimal("1000.9"));

        // Define a new SDO open content property with simple type
        Property xmlElementProp = getScope().getXSDHelper().getGlobalProperty("commonj.sdo/xml", "xmlElement", false);
        DataObject p = dataFactory.create("commonj.sdo", "Property");
        p.set("type", typeHelper.getType("commonj.sdo", "Decimal"));
        p.set("name", "highPrice");
        p.set(xmlElementProp, Boolean.TRUE);
        Property highPrice = typeHelper.defineOpenContentProperty(OPEN_TYPE_NAMESPACE, p);

        quote.setBigDecimal(highPrice, new BigDecimal("1100.9"));

        // Define a new SDO open content property with complex type
        DataObject mutualFundQuotePropertyDef = dataFactory.create("commonj.sdo", "Property");
        mutualFundQuotePropertyDef.set("type", quoteType);
        mutualFundQuotePropertyDef.set("name", "mutualFundQuote");
        mutualFundQuotePropertyDef.setBoolean("containment", true);
        Property mutualFundQuoteProperty =
            typeHelper.defineOpenContentProperty(OPEN_TYPE_NAMESPACE, mutualFundQuotePropertyDef);

        DataObject mutualFundQuote = quote.createDataObject(mutualFundQuoteProperty);
        mutualFundQuote.setString("symbol", "mutual-1");

        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        xmlHelper.save(quote, OPEN_TYPE_NAMESPACE, "openStockQuote2", baos);

        assertTrue(XMLCompareUtil.compareXML(new ByteArrayInputStream(baos.toByteArray()), getClass().getResource(OPEN_XML)));

        // validate existing property condition
        Property duplicateProp = typeHelper.defineOpenContentProperty(OPEN_TYPE_NAMESPACE, p);
        assertTrue("The expected Property equality was not true", highPrice.equals(duplicateProp));

        // validate error condition, where new property exists with different
        // type
        boolean errorCondition = false;
        try {
            p = dataFactory.create("commonj.sdo", "Property");
            p.set("type", typeHelper.getType("commonj.sdo", "String"));
            p.set("name", "highPrice");
            p.set(xmlElementProp, Boolean.TRUE);            
            highPrice = typeHelper.defineOpenContentProperty(OPEN_TYPE_NAMESPACE, p);
        } catch (IllegalArgumentException ex) {
            errorCondition = true;
        }
        assertTrue("Should not be able to redefine a Property with a different Type.", errorCondition);
    }

    /**
     * Verify how the XMLHelper handles Types defined using the TypeHelper. Save
     * and load the Dataobject with the XMLHelper as an intermediate step.<br/>
     * TODO this is a hybrid test that needs some refactoring.  It implicitly tests
     * dynamic type creation, creation of dynamic types, xml round tripping and the
     * dataObject API without a great deal of structure.  Some assertions are 
     * irrelevant since the tests would throw NPEs before performing the
     * tests.  Other assertions could be stronger than checking for non-null. See also
     * testDefineTypes for similar issues.<br/>
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=41">SDO spec 2..1 section 3.8.4</a>
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=46">SDO spec 2..1 section 3.11.3</a>
     * @see TypeHelper#define(DataObject)
     * @see XMLHelper#save(DataObject, String, String, java.io.OutputStream)
     * @see XMLHelper#load(InputStream, String, Object)
     * 
     */
    @Test
    public void testDefineTypeRoundTrip() throws Exception {
        TypeHelper types = getScope().getTypeHelper();
        DataFactory factory = getScope().getDataFactory();
        XMLHelper xmlHelper = getScope().getXMLHelper();

        Type intType = types.getType("commonj.sdo", "Int");
        assertNotNull( intType );
        Type stringType = types.getType("commonj.sdo", "String");

        // create a new Type for Customers
        DataObject customerType = factory.create("commonj.sdo", "Type");
        customerType.set("uri", "http://example.com/customer");
        customerType.set("name", "Customer");

        Property xmlElementProp = getScope().getXSDHelper().getGlobalProperty("commonj.sdo/xml", "xmlElement", false);

        
        // create a customer number property
        DataObject custNumProperty = customerType.createDataObject("property");
        custNumProperty.set("name", "custNum");
        custNumProperty.set("type", intType);
        custNumProperty.setBoolean(xmlElementProp, false);

        // create a first name property
        DataObject firstNameProperty = customerType.createDataObject("property");
        firstNameProperty.set("name", "firstName");
        firstNameProperty.set("type", stringType);
        firstNameProperty.setBoolean(xmlElementProp, false);

        // create a last name property
        DataObject lastNameProperty = customerType.createDataObject("property");
        lastNameProperty.set("name", "lastName");
        lastNameProperty.set("type", stringType);
        lastNameProperty.setBoolean(xmlElementProp, false);

        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        xmlHelper.save(customerType, "commonj.sdo", "type", baos);

        ByteArrayInputStream bais = new ByteArrayInputStream(baos.toByteArray());
        XMLDocument xdoc = xmlHelper.load(bais);

        customerType = xdoc.getRootObject();

        // now define the Customer type so that customers can be made
        types.define(customerType);

        DataObject customer1 = factory.create("http://example.com/customer", "Customer");

        customer1.setInt("custNum", 1);
        customer1.set("firstName", "John");
        customer1.set("lastName", "Adams");
        DataObject customer2 = factory.create("http://example.com/customer", "Customer");
        customer2.setInt("custNum", 2);
        customer2.set("firstName", "Jeremy");
        customer2.set("lastName", "Pavick");

        assertNotNull("Customer1 should not be null", customer1);
        Type type = customer1.getType();
        assertNotNull("Customer1.custNum should not be null", type.getProperty("custNum"));
        assertNotNull("Customer1.firstName should not be null", type.getProperty("firstName"));
        assertNotNull("Customer1.lastName should not be null", type.getProperty("lastName"));
        assertEquals("Customer1.custNum should be intType", type.getProperty("custNum").getType(), intType);
        assertEquals("Customer1.firstName should be stringType", type.getProperty("firstName").getType(), stringType);
        assertEquals("Customer1.lastName should be stringType", type.getProperty("lastName").getType(), stringType);

        assertNotNull("Customer2 should not be null", customer2);
        type = customer2.getType();
        assertNotNull("Customer1.custNum should not be null", type.getProperty("custNum"));
        assertNotNull("Customer1.custNum should not be null", type.getProperty("firstName"));
        assertNotNull("Customer1.custNum should not be null", type.getProperty("lastName"));
        assertEquals("Customer1.custNum should be intType", type.getProperty("custNum").getType(), intType);
        assertEquals("Customer1.firstName should be stringType", type.getProperty("firstName").getType(), stringType);
        assertEquals("Customer1.lastName should be stringType", type.getProperty("lastName").getType(), stringType);

        baos = new ByteArrayOutputStream();
        xmlHelper.save(customer1, "http://example.com/customer", "Customer", baos);

        assertTrue(XMLCompareUtil.compareXML(new ByteArrayInputStream(baos.toByteArray()), getClass().getResource(CUSTOMER1_XML)));

        baos = new ByteArrayOutputStream();
        xmlHelper.save(customer2, "http://example.com/customer", "Customer", baos);

        assertTrue(XMLCompareUtil.compareXML(new ByteArrayInputStream(baos.toByteArray()), getClass().getResource(CUSTOMER2_XML)));
    }

    /**
     * Verify how the XMLHelper handles Types defined using the TypeHelper.
     * TODO this is the same as {@link TypeHelperDefineTest#testDefineTypeRoundTrip()} without
     * the round tripping of the metametadata. See the issues described in the doc for that method.
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=41">SDO spec 2..1 section 3.8.4</a>
     * @see TypeHelper#define(DataObject)

     */
    @Test
    public void testDefineType() throws Exception {
        TypeHelper types = getScope().getTypeHelper();
        DataFactory factory = getScope().getDataFactory();
        XMLHelper xmlHelper = getScope().getXMLHelper();

        Type intType = types.getType("commonj.sdo", "Int");
        Type stringType = types.getType("commonj.sdo", "String");

        // create a new Type for Customers
        DataObject customerType = factory.create("commonj.sdo", "Type");
        customerType.set("uri", "http://example.com/customer");
        customerType.set("name", "Customer");

        Property xmlElementProp = getScope().getXSDHelper().getGlobalProperty("commonj.sdo/xml", "xmlElement", false);
        
        // create a customer number property
        DataObject custNumProperty = customerType.createDataObject("property");
        custNumProperty.set("name", "custNum");
        custNumProperty.set("type", intType);
        custNumProperty.setBoolean(xmlElementProp, false);

        // create a first name property
        DataObject firstNameProperty = customerType.createDataObject("property");
        firstNameProperty.set("name", "firstName");
        firstNameProperty.set("type", stringType);
        firstNameProperty.setBoolean(xmlElementProp, false);

        // create a last name property
        DataObject lastNameProperty = customerType.createDataObject("property");
        lastNameProperty.set("name", "lastName");
        lastNameProperty.set("type", stringType);
        lastNameProperty.setBoolean(xmlElementProp, false);

        // now define the Customer type so that customers can be made
        types.define(customerType);

        DataObject customer1 = factory.create("http://example.com/customer", "Customer");
        customer1.setInt("custNum", 1);
        customer1.set("firstName", "John");
        customer1.set("lastName", "Adams");
        DataObject customer2 = factory.create("http://example.com/customer", "Customer");
        customer2.setInt("custNum", 2);
        customer2.set("firstName", "Jeremy");
        customer2.set("lastName", "Pavick");

        assertNotNull("Customer1 should not be null", customer1);
        Type type = customer1.getType();
        assertNotNull("Customer1.custNum should not be null", type.getProperty("custNum"));
        assertNotNull("Customer1.firstName should not be null", type.getProperty("firstName"));
        assertNotNull("Customer1.lastName should not be null", type.getProperty("lastName"));
        assertEquals("Customer1.custNum should be intType", type.getProperty("custNum").getType(), intType);
        assertEquals("Customer1.firstName should be stringType", type.getProperty("firstName").getType(), stringType);
        assertEquals("Customer1.lastName should be stringType", type.getProperty("lastName").getType(), stringType);

        assertNotNull("Customer2 should not be null", customer2);
        type = customer2.getType();
        assertNotNull("Customer1.custNum should not be null", type.getProperty("custNum"));
        assertNotNull("Customer1.custNum should not be null", type.getProperty("firstName"));
        assertNotNull("Customer1.custNum should not be null", type.getProperty("lastName"));
        assertEquals("Customer1.custNum should be intType", type.getProperty("custNum").getType(), intType);
        assertEquals("Customer1.firstName should be stringType", type.getProperty("firstName").getType(), stringType);
        assertEquals("Customer1.lastName should be stringType", type.getProperty("lastName").getType(), stringType);

        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        xmlHelper.save(customer1, "http://example.com/customer", "Customer", baos);

        assertTrue(XMLCompareUtil.compareXML(new ByteArrayInputStream(baos.toByteArray()), getClass().getResource(CUSTOMER1_XML)));

        baos = new ByteArrayOutputStream();
        xmlHelper.save(customer2, "http://example.com/customer", "Customer", baos);

        assertTrue(XMLCompareUtil.compareXML(new ByteArrayInputStream(baos.toByteArray()), getClass().getResource(CUSTOMER2_XML)));
    }

    /**
     * Verify how the XMLHelper handles defined Sequence types.
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=18">SDO spec 2..1 section 3.1.8</a>
     * @see DataObject#getSequence()
     * @see Sequence#addText(String)
     * @see Sequence#add(Property, Object)
     */
    @Test
    public void testDefineSequencedType() throws Exception {

        TypeHelper types = getScope().getTypeHelper();
        DataFactory factory = getScope().getDataFactory();
        XMLHelper xmlHelper = getScope().getXMLHelper();

        Type stringType = types.getType("commonj.sdo", "String");
        Type decimalType = types.getType("commonj.sdo", "Decimal");

        // Define a new mixed type - MixedQuote
        DataObject mixedQuoteType = factory.create("commonj.sdo", "Type");
        mixedQuoteType.set("uri", "http://www.example.com/mixed");
        mixedQuoteType.set("name", "MixedQuote");
        mixedQuoteType.set("sequenced", Boolean.TRUE);


        DataObject symbolProperty = mixedQuoteType.createDataObject("property");
        symbolProperty.set("name", "symbol");
        symbolProperty.set("type", stringType);
        Property xmlElement = types.getOpenContentProperty("commonj.sdo/xml", "xmlElement");
        symbolProperty.set(xmlElement, Boolean.TRUE);

        DataObject companyNameProperty = mixedQuoteType.createDataObject("property");
        companyNameProperty.set("name", "companyName");
        companyNameProperty.set("type", stringType);
        companyNameProperty.set(xmlElement, Boolean.TRUE);

        DataObject priceProperty = mixedQuoteType.createDataObject("property");
        priceProperty.set("name", "price");
        priceProperty.set("type", decimalType);
        priceProperty.set(xmlElement, Boolean.TRUE);

        DataObject quotesProperty = mixedQuoteType.createDataObject("property");
        quotesProperty.set("name", "quotes");
        quotesProperty.set("type", mixedQuoteType);
        quotesProperty.set("many", Boolean.TRUE);
        quotesProperty.set("containment", Boolean.TRUE);

        types.define(mixedQuoteType);

        DataObject quote = factory.create("http://www.example.com/mixed", "MixedQuote");

        assertTrue("Quote type should be Sequenced", quote.getType().isSequenced());

        Sequence sequence = quote.getSequence();

        sequence.addText("\n  ");

        quote.setString("symbol", "fbnt");

        sequence.addText("\n  ");

        quote.setString("companyName", "FlyByNightTechnology");

        sequence.addText("\n  some text\n  ");

        DataObject child = quote.createDataObject("quotes");
        child.setBigDecimal("price", new BigDecimal("2000.9"));

        sequence.addText("\n  more text\n  ");

        // quote.setBigDecimal("price", new BigDecimal("1000.9"));
        sequence.add("price", new BigDecimal("1000.9"));

        sequence.addText("\n");

        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        xmlHelper.save(quote, "http://www.example.com/mixed", "mixedStockQuote", baos);

        assertTrue(XMLCompareUtil.compareXML(new ByteArrayInputStream(baos.toByteArray()), getClass().getResource(MIXED_XML)));
    }

    /**
     * Test Sequenced and Open dynamically defined types.
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=18">SDO spec 2..1 section 3.1.8</a>
     * @see DataObject#getSequence()
     * @see Sequence#addText(String)
     * @see Sequence#add(Property, Object)
     * @see TypeHelper#defineOpenContentProperty(String, DataObject)
     * @see TypeHelper#getOpenContentProperty(String, String)
     */
    @Test
    public void testDefineSequencedOpenType() throws Exception {
        TypeHelper types = getScope().getTypeHelper();
        DataFactory factory = getScope().getDataFactory();
        XMLHelper xmlHelper = getScope().getXMLHelper();

        Type stringType = types.getType("commonj.sdo", "String");
        Type decimalType = types.getType("commonj.sdo", "Decimal");

        // Define a new mixed type - MixedQuote
        DataObject mixedQuoteType = factory.create("commonj.sdo", "Type");
        mixedQuoteType.set("uri", "http://www.example.com/mixed");
        mixedQuoteType.set("name", "MixedOpenQuote");
        mixedQuoteType.set("sequenced", Boolean.TRUE);
        mixedQuoteType.set("open", Boolean.TRUE);

        // DataObject symbolProperty =
        // mixedQuoteType.createDataObject("property");
        // symbolProperty.set("name", "symbol");
        // symbolProperty.set("type", stringType);

        DataObject companyNameProperty = mixedQuoteType.createDataObject("property");
        companyNameProperty.set("name", "companyName");
        companyNameProperty.set("type", stringType);
        Property xmlElement = types.getOpenContentProperty("commonj.sdo/xml", "xmlElement");
        companyNameProperty.set(xmlElement, Boolean.TRUE);

        DataObject priceProperty = mixedQuoteType.createDataObject("property");
        priceProperty.set("name", "price");
        priceProperty.set("type", decimalType);
        priceProperty.set(xmlElement, Boolean.TRUE);

        DataObject quotesProperty = mixedQuoteType.createDataObject("property");
        quotesProperty.set("name", "quotes");
        quotesProperty.set("type", mixedQuoteType);
        quotesProperty.set("many", Boolean.TRUE);
        quotesProperty.set("containment", Boolean.TRUE);

        types.define(mixedQuoteType);

        // Define a global property
        DataObject symbolProperty = factory.create("commonj.sdo", "Property");
        symbolProperty.set("name", "symbol");
        symbolProperty.set("type", stringType);
        symbolProperty.set(xmlElement, Boolean.TRUE);

        types.defineOpenContentProperty("http://www.example.com/open", symbolProperty);

        DataObject quote = factory.create("http://www.example.com/mixed", "MixedOpenQuote");

        assertTrue("Quote type should be Sequenced.", quote.getType().isSequenced());

        Sequence sequence = quote.getSequence();

        sequence.addText("\n  ");

        Property definedSymbolProperty = types.getOpenContentProperty("http://www.example.com/open", "symbol");
        quote.setString(definedSymbolProperty, "fbnt");

        sequence.addText("\n  ");

        quote.setString("companyName", "FlyByNightTechnology");

        sequence.addText("\n  some text\n  ");

        DataObject child = quote.createDataObject("quotes");
        child.setBigDecimal("price", new BigDecimal("2000.9"));

        sequence.addText("\n  more text\n  ");

        // quote.setBigDecimal("price", new BigDecimal("1000.9"));
        sequence.add("price", new BigDecimal("1000.9"));

        sequence.addText("\n");

        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        xmlHelper.save(quote, "http://www.example.com/mixed", "mixedOpenStockQuote", baos);

        FileOutputStream fos = new FileOutputStream("XMLTCOutput");
        xmlHelper.save(quote, "http://www.example.com/mixed", "mixedOpenStockQuote", fos);

        assertTrue(XMLCompareUtil.compareXML(new ByteArrayInputStream(baos.toByteArray()), getClass().getResource(MIXEDOPEN_XML)));
    }
    
    @Before
    public void setUp() throws Exception {
      super.setUp();
      String TEST_MODEL = "/sharedFiles/open.xsd";
      
      // Populate the meta data for the test (Stock Quote) model
      URL url = getClass().getResource(TEST_MODEL);
      InputStream inputStream = url.openStream();
      getScope().getXSDHelper().define(inputStream, url.toString());
      inputStream.close();
      
    }
    
    @After
    public void tearDown() throws Exception {
      super.tearDown();
    }
    
    
}
