/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 *  
 *  $Rev: 2050 $  $Date: 2009-03-04 19:35:51 +0000 (Wed, 04 Mar 2009) $
 */
package com.ibm.sdo.tck.tests.helpers;

import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.assertEquals;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.Iterator;
import java.util.List;


import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import com.ibm.sdo.tck.framework.TCKTestBase;
import com.ibm.sdo.tck.framework.annotations.Operation;
import com.ibm.sdo.tck.framework.annotations.TCKTest;
import com.ibm.sdo.tck.util.EqualityTest;

import commonj.sdo.DataObject;
import commonj.sdo.Property;
import commonj.sdo.helper.HelperContext;
import commonj.sdo.helper.XMLDocument;
import commonj.sdo.helper.XSDHelper;

public class CopyHelper2Test extends TCKTestBase {

    @Before
    public void setUp() throws Exception {
        super.setUp();
        XSDHelper xsdHelper = getScope().getXSDHelper();

        InputStream is = null;
        try {

            URL url = getClass().getResource("/copyHelperTest/copyHelperTestTypes.xsd");
            is = url.openStream();
            xsdHelper.define(is, url.toString());

        } catch (Exception e) {
            throw new IllegalStateException("The copyHelperTestTypes schema must be a resource on the class path", e);
        } finally {
            is.close();
        }

    }

    @After
    public void tearDown() throws Exception {
        super.tearDown();
    }

    @Test
    @TCKTest(testFor = @Operation(className = "CopyHelper", methodName = "copyShallow", signature = "DataObject", precondition = {"input graph has non-dataType values"}, postcondition = {
                                                                                                                                                                                           "copied graph has no non-data type values",
                                                                                                                                                                                           "copy has same Type as input",
                                                                                                                                                                                           "dataType values of copy are the same as the inputs"}))
    public void shallowCopy() throws IOException {
        InputStream is = null;
        XMLDocument result = null;
        HelperContext scope = getScope();

        try {
            URL url = getClass().getResource("/copyHelperTest/copyHelperTest.xml");
            is = url.openStream();
            result = scope.getXMLHelper().load(is);
        } catch (Exception e) {
            throw new IllegalStateException("there's something wrong with the test data", e);
        }

        DataObject inputDO = result.getRootObject();

        assertNotNull(inputDO);

        DataObject copyDo = scope.getCopyHelper().copyShallow(inputDO);

        EqualityTest et = new EqualityTest(inputDO, copyDo, false);
        et.getResult();
        assertTrue("Shallow copy should not give equal graphs for this input data", et.getResult());
        assertTrue("Shallow copied DataObjects should share Type instances", inputDO.getType() == copyDo.getType());

        // prune the input data object to that which should equal the output

        List props = inputDO.getInstanceProperties();
        for (Iterator iterator = props.iterator(); iterator.hasNext();) {
            Property p = (Property)iterator.next();
            if (!p.getType().isDataType()) {
                inputDO.unset(p);
            }
        }

        EqualityTest et2 = new EqualityTest(inputDO, copyDo, true);
        assertTrue("Pruning the input graph of non-datatype Property values did not result in equivalence to shallow copy: " + et2
                       .getFailure(),
                   et2.getResult());

    }

    @Test
    @TCKTest(testFor = @Operation(className = "CopyHelper", methodName = "copy", signature = "DataObject", precondition = {"input graph has non-dataType values"}, postcondition = {
                                                                                                                                                                                    "copy has same Type as input",
                                                                                                                                                                                    "graph is deep equivalent in shape and values"}))
    public void deepCopy() throws IOException {
        InputStream is = null;
        XMLDocument result = null;
        HelperContext scope = getScope();

        try {
            URL url = getClass().getResource("/copyHelperTest/copyHelperTest.xml");
            is = url.openStream();
            result = scope.getXMLHelper().load(is);
        } catch (Exception e) {
            throw new IllegalStateException("there's something wrong with the test data", e);
        }

        DataObject inputDO = result.getRootObject();

        assertNotNull(inputDO);

        DataObject copyDo = scope.getCopyHelper().copy(inputDO);

        EqualityTest et = new EqualityTest(inputDO, copyDo, true);
        et.getResult();
        assertTrue("Deep copy should give equal graphs for this input data: " + et.getFailure(), et.getResult());
        assertTrue("Copied DataObjects should share Type instances", inputDO.getType() == copyDo.getType());

    }

    @Test
    @TCKTest(testFor = @Operation(className = "CopyHelper", methodName = "copyShallow", signature = "DataObject", precondition = {
                                                                                                                                  "input graph has non-dataType values",
                                                                                                                                  "root object has an attribute of Type ChangeSummaryType",
                                                                                                                                  "ChangeSummary has data on one alteration to an attribute",
                                                                                                                                  "ChangeSummary is logging changes"}, postcondition = {
                                                                                                                                                                                        "Copied ChangeSummary reports 0 altered object",
                                                                                                                                                                                        "logging state of copied ChangeSummary is on"}))
    public void shallowCopy_ChangeSummaryNotEmpty_Logging() throws IOException {
        InputStream is = null;
        XMLDocument result = null;
        HelperContext scope = getScope();

        try {
            URL url = getClass().getResource("/copyHelperTest/copyHelperTestCS.xml");
            is = url.openStream();
            result = scope.getXMLHelper().load(is);
        } catch (Exception e) {
            throw new IllegalStateException("there's something wrong with the test data", e);
        }

        DataObject inputDO = result.getRootObject();
        inputDO.getChangeSummary().beginLogging();

        inputDO.setString("se1", "changed");

        assertEquals("incorrect number of changed objects", 1, inputDO.getChangeSummary().getChangedDataObjects()
            .size());

        DataObject copyDo = scope.getCopyHelper().copyShallow(inputDO);

        EqualityTest et = new EqualityTest(inputDO, copyDo, false);
        et.getResult();

        assertTrue("Logging state of copy not equal to input", copyDo.getChangeSummary().isLogging());
        assertEquals("Change summary of copied object should be empty", 0, copyDo.getChangeSummary()
            .getChangedDataObjects().size());

    }

    @Test
    @TCKTest(testFor = @Operation(className = "CopyHelper", methodName = "copyShallow", signature = "DataObject", precondition = {
                                                                                                                                  "input graph has non-dataType values",
                                                                                                                                  "root object has an attribute of Type ChangeSummaryType",
                                                                                                                                  "ChangeSummary has data on one alteration to an attribute",
                                                                                                                                  "ChangeSummary is not logging changes"}, postcondition = {
                                                                                                                                                                                            "Copied ChangeSummary reports 0 altered object",
                                                                                                                                                                                            "logging state of copied ChangeSummary is off"}))
    public void shallowCopy_ChangeSummaryNotEmpty_NotLogging() throws IOException {
        InputStream is = null;
        XMLDocument result = null;
        HelperContext scope = getScope();

        try {
            URL url = getClass().getResource("/copyHelperTest/copyHelperTestCS.xml");
            is = url.openStream();
            result = scope.getXMLHelper().load(is);
        } catch (Exception e) {
            throw new IllegalStateException("there's something wrong with the test data", e);
        }

        DataObject inputDO = result.getRootObject();
        inputDO.getChangeSummary().beginLogging();

        inputDO.setString("se1", "changed");

        assertEquals("incorrect number of changed objects", 1, inputDO.getChangeSummary().getChangedDataObjects()
            .size());

        inputDO.getChangeSummary().endLogging();

        DataObject copyDo = scope.getCopyHelper().copyShallow(inputDO);

        EqualityTest et = new EqualityTest(inputDO, copyDo, false);
        et.getResult();

        assertTrue("Logging state of copy not equal to input", !copyDo.getChangeSummary().isLogging());
        assertEquals("Change summary of copied object should be empty", 0, copyDo.getChangeSummary()
            .getChangedDataObjects().size());

    }

    @Test
    @TCKTest(testFor = @Operation(className = "CopyHelper", methodName = "copy", signature = "DataObject", precondition = {
                                                                                                                           "input graph has non-dataType values",
                                                                                                                           "root object has an attribute of Type ChangeSummaryType",
                                                                                                                           "ChangeSummary has data on one alteration to an attribute",
                                                                                                                           "ChangeSummary is logging changes"}, postcondition = {
                                                                                                                                                                                 "Copied ChangeSummary reports same number of altered objects as input",
                                                                                                                                                                                 "logging state of copied ChangeSummary is on"}))
    public void deepCopy_ChangeSummaryNotEmpty_Logging() throws IOException {
        InputStream is = null;
        XMLDocument result = null;
        HelperContext scope = getScope();

        try {
            URL url = getClass().getResource("/copyHelperTest/copyHelperTestCS.xml");
            is = url.openStream();
            result = scope.getXMLHelper().load(is);
        } catch (Exception e) {
            throw new IllegalStateException("there's something wrong with the test data", e);
        }

        DataObject inputDO = result.getRootObject();
        inputDO.getChangeSummary().beginLogging();

        inputDO.setString("se1", "changed");

        assertEquals("incorrect number of changed objects", 1, inputDO.getChangeSummary().getChangedDataObjects()
            .size());

        DataObject copyDo = scope.getCopyHelper().copy(inputDO);

        EqualityTest et = new EqualityTest(inputDO, copyDo, false);
        et.getResult();

        assertTrue("Logging state of copy not equal to input", copyDo.getChangeSummary().isLogging());
        assertEquals("Change summary of copied object should be equivalent to input object", inputDO.getChangeSummary()
            .getChangedDataObjects().size(), copyDo.getChangeSummary().getChangedDataObjects().size());

    }

}
