/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 *  
 *  $Rev: 553220 $  $Date: 2008/01/15 11:17:51 $
 */
package com.ibm.sdo.tck.tests.helpers;

import static org.junit.Assert.assertEquals;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.TimeZone;

import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.datatype.XMLGregorianCalendar;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import com.ibm.sdo.tck.framework.annotations.Operation;
import com.ibm.sdo.tck.framework.annotations.SpecRef;
import com.ibm.sdo.tck.framework.annotations.TCKTest;
import com.ibm.sdo.tck.testcasebase.ConversionTestBase;

import commonj.sdo.helper.DataHelper;
import commonj.sdo.helper.HelperContext;

/**
 * Test the DataHelper API.
 * DataHelperTest date tests insure that the DataHelper conversions accurately retain the
 * information in the specified fields (e.g. month, day or year).
 */
@TCKTest(testFor = {
    @Operation(className = "HelperContext", methodName = "getDataHelper", signature = ""),
    @Operation(className = "DataHelper", methodName = "toDate", signature = "String"),
    @Operation(className = "DataHelper", methodName = "toDay", signature = "Date"),
    @Operation(className = "DataHelper", methodName = "toDay", signature = "Calendar"),
    @Operation(className = "DataHelper", methodName = "toCalendar", signature = "String"),
    // TODO @Operation(className = "DataHelper", methodName = "toCalendar", signature = "String,Locale"),
    @Operation(className = "DataHelper", methodName = "toDateTime", signature = "Date"),
    @Operation(className = "DataHelper", methodName = "toDateTime", signature = "Calendar"),
    @Operation(className = "DataHelper", methodName = "toTime", signature = "Date"),
    @Operation(className = "DataHelper", methodName = "toTime", signature = "Calendar"),
    @Operation(className = "DataHelper", methodName = "toMonth", signature = "Date"),
    @Operation(className = "DataHelper", methodName = "toMonth", signature = "Calendar"),
    @Operation(className = "DataHelper", methodName = "toMonthDay", signature = "Date"),
    @Operation(className = "DataHelper", methodName = "toMonthDay", signature = "Calendar"),
    @Operation(className = "DataHelper", methodName = "toYear", signature = "Date"),
    @Operation(className = "DataHelper", methodName = "toYear", signature = "Calendar"),
    @Operation(className = "DataHelper", methodName = "toYearMonth", signature = "Date"),
    @Operation(className = "DataHelper", methodName = "toYearMonth", signature = "Calendar"),
    @Operation(className = "DataHelper", methodName = "toYearMonthDay", signature = "Date"),
    @Operation(className = "DataHelper", methodName = "toYearMonthDay", signature = "Calendar")
  }

)
public class DataHelperTest extends ConversionTestBase 
{
    static final String DateTime      = "DateTime";
    static final String Day           = "Day";
    static final String Month         = "Month";
    static final String MonthDay      = "MonthDay";
    static final String YearMonthDay  = "YearMonthDay";
    static final String YearMonth     = "YearMonth";
    static final String Year          = "Year";
    static final String Time          = "Time";

    private HelperContext scope;
    private DataHelper    dh;
    
    private static ArrayList <String> CONV_TABLE_TYPES;
    private static ArrayList <int []> COMPARISON_FIELDS;
    
    // This table provides a map of 'compatible' SDO Date/Time strings.
    private static final char [] CONV_TABLE = { 
        /*DateTime*/     'X', 'x', 'x', 'x', 'x', 'x', 'x', 'x',   
        /*Day*/          ' ', 'X', ' ', ' ', ' ', ' ', ' ', ' ',  
        /*Month*/        ' ', ' ', 'X', ' ', ' ', ' ', ' ', ' ',  
        /*MonthDay*/     ' ', 'x', 'x', 'X', ' ', ' ', ' ', ' ',  
        /*YearMonthDay*/ ' ', 'x', 'x', 'x', 'X', 'x', 'x', ' ',  
        /*YearMonth*/    ' ', ' ', 'x', ' ', ' ', 'X', 'x', ' ',  
        /*Year*/         ' ', ' ', ' ', ' ', ' ', ' ', 'X', ' ',  
        /*Time*/         ' ', ' ', ' ', ' ', ' ', ' ', ' ', 'X',  
   };

    // This table contains the core set of date/time strings to use for 
    // the tests contained in this class.  The source string is in the first
    // column.  The type of that string is identified int the second column.
    // The third column is used by the roundtrip test to validate Date converted
    // roundtrips and the fourth column for Calendar converted roundtrips.
    private static final String[][] testData = 
    { 
                                                                                                                    // This column is here due to a limitation
                                                                                                                    // in the Java libraries for the XMLGregorianCalendar
    //  * Test Value *              * Test Type *  * Expected Date Roundtrip *   * Expected Calendar Roundtrip *     * Skip gMonth Conv. *
        {"2006-03-31T03:30:45.123",   DateTime,      "2006-03-31T03:30:45.123Z",   "2006-03-31T03:30:45.123",        "false" },
        {"2006-03-31T03:30:45.123Z",  DateTime,      "2006-03-31T03:30:45.123Z",   "2006-03-31T03:30:45.123Z",       "true"  }, 
        {"-2006-03-31T03:30:45.001Z", DateTime,      "-2006-03-31T03:30:45.001Z",  "-2006-03-31T03:30:45.001Z",      "true"  }, 
        {"2006-03-31T03:30:45Z",      DateTime,      "2006-03-31T03:30:45Z",       "2006-03-31T03:30:45Z",           "true"  },
        {"2006-03-31T03:30:45.123",   DateTime,      "2006-03-31T03:30:45.123Z",   "2006-03-31T03:30:45.123",        "false" },
        {"2006-03-31T03:30:45.001",   DateTime,      "2006-03-31T03:30:45.001Z",   "2006-03-31T03:30:45.001",        "false" },
        {"-2006-03-31T03:30:45",      DateTime,      "-2006-03-31T03:30:45Z",      "-2006-03-31T03:30:45",           "false" },
        {"2006-03-31T03:30:45",       DateTime,      "2006-03-31T03:30:45Z",       "2006-03-31T03:30:45",            "false" },
        {"1999-12-31T23:59:00",       DateTime,      "1999-12-31T23:59:00Z",       "1999-12-31T23:59:00",            "false" },
        {"1999-12-31T23:59:00-05:00", DateTime,      "2000-01-01T04:59:00Z",       "1999-12-31T23:59:00-05:00",      "false" },
        {"---05+01:00",               Day,            "---04",                     "---05+01:00",                    "false" },
        {"---04",                     Day,            "---04",                     "---04",                          "false" },
        {"--12+02:00",                Month,          "--11",                      "--12+02:00",                     "false" },
        {"--12",                      Month,          "--12",                      "--12",                           "false" },
        {"--08-08+01:00",             MonthDay,       "--08-07",                   "--08-08+01:00",                  "false" },
        {"--08-08",                   MonthDay,       "--08-08",                   "--08-08",                        "false" },
        {"13:20:00",                  Time,           "13:20:00Z",                 "13:20:00",                       "false" },
        {"13:20:00.5555",             Time,           "13:20:00.555Z",             "13:20:00.555",                   "false" },
        {"13:20:00-05:00",            Time,           "18:20:00Z",                 "13:20:00-05:00",                 "false" },
        {"1976-08-08+04:00",          YearMonthDay,   "1976-08-07",                "1976-08-08+04:00",               "false" },
        {"1976-08-08",                YearMonthDay,   "1976-08-08",                "1976-08-08",                     "false" },
        {"1988-12-01:00",             YearMonth,      "1988-12",                   "1988-12-01:00",                  "false" },
        {"1988-12",                   YearMonth,       "1988-12",                  "1988-12",                        "false" },
        {"2005+02:00",                Year,           "2004",                      "2005+02:00",                     "false" },
        {"1999",                      Year,           "1999",                      "1999",                           "false" }
   };    

   static
   {
       CONV_TABLE_TYPES = new ArrayList<String>();
       CONV_TABLE_TYPES.add(DateTime);
       CONV_TABLE_TYPES.add(Day);
       CONV_TABLE_TYPES.add(Month);
       CONV_TABLE_TYPES.add(MonthDay);
       CONV_TABLE_TYPES.add(YearMonthDay);
       CONV_TABLE_TYPES.add(YearMonth);
       CONV_TABLE_TYPES.add(Year);
       CONV_TABLE_TYPES.add(Time);
       
       COMPARISON_FIELDS = new ArrayList<int []>();
       COMPARISON_FIELDS.add(new int[] {Calendar.YEAR, Calendar.MONTH, Calendar.DAY_OF_MONTH,
                                        Calendar.HOUR_OF_DAY, Calendar.MINUTE, Calendar.SECOND,
                                        Calendar.MILLISECOND});
       COMPARISON_FIELDS.add(new int[] {Calendar.DAY_OF_MONTH});
       COMPARISON_FIELDS.add(new int[] {Calendar.MONTH});
       COMPARISON_FIELDS.add(new int[] {Calendar.MONTH, Calendar.DAY_OF_MONTH});
       COMPARISON_FIELDS.add(new int[] {Calendar.YEAR, Calendar.MONTH, Calendar.DAY_OF_MONTH});
       COMPARISON_FIELDS.add(new int[] {Calendar.YEAR, Calendar.MONTH});        
       COMPARISON_FIELDS.add(new int[] {Calendar.YEAR});               
       COMPARISON_FIELDS.add(new int[] {Calendar.HOUR_OF_DAY, Calendar.MINUTE, Calendar.SECOND, Calendar.MILLISECOND});                     
   }
    
    @Before 
    public void setUp() throws Exception
    {
      super.setUp();
      
      scope = getScope();
      dh    = scope.getDataHelper();
      
    }
    
    @After
    public void tearDown() throws Exception {
      super.tearDown();
    }   
    
    /*
     * Tests the SDO Impls ability to convert string based Date/Time data 
     * to a Date object, first, then to a compatible SDO Date/Time string. 
     */
    @Test
    public void testDateBasedConversions() throws DatatypeConfigurationException
    {        
        for (int test = 0; test < testData.length; test++) 
        {
            Date date = dh.toDate(testData[test][0]);
            int testTypeIndex = CONV_TABLE_TYPES.indexOf(testData[test][1]);
            for (int convType=testTypeIndex*CONV_TABLE_TYPES.size(); convType<(testTypeIndex+1)*CONV_TABLE_TYPES.size(); convType++ )
            {
                int convTypeIndex = convType-testTypeIndex*CONV_TABLE_TYPES.size();
                boolean skip = Boolean.parseBoolean(testData[test][4]) && Month.equals(CONV_TABLE_TYPES.get(convTypeIndex));
                if( (CONV_TABLE[convType] == 'x' || CONV_TABLE[convType] == 'X') && !skip )
                {
                    String value = getStringFromDateValue(date,CONV_TABLE_TYPES.get(convTypeIndex));
                    String convInfo = "Converting from " + testData[test][1] + " to " + CONV_TABLE_TYPES.get(convTypeIndex) + ".  Comparing control='" + testData[test][2] + "' to test='" + value + "'.";
                    // for Date Conversions we compare the to the date rountrip column.
                    compareFields(testData[test][2], value, COMPARISON_FIELDS.get(convTypeIndex),convInfo);
                }
            }
        }
    }
    
    /*
     * Tests the SDO Impls ability to convert string based Date/Time data 
     * to a Calendar object, first, then to a compatible SDO Date/Time string. 
     */
    @Test
    public void testCalendarBasedConversions() throws DatatypeConfigurationException
    {        
        for (int test = 0; test < testData.length; test++) 
        {
            Calendar calendar = dh.toCalendar(testData[test][0]);
            GregorianCalendar control = (GregorianCalendar)calendar;
            int testTypeIndex = CONV_TABLE_TYPES.indexOf(testData[test][1]);
            for (int convType=testTypeIndex*CONV_TABLE_TYPES.size(); convType<(testTypeIndex+1)*CONV_TABLE_TYPES.size(); convType++ )
            {
                int convTypeIndex = convType-testTypeIndex*CONV_TABLE_TYPES.size();
                boolean skip = Boolean.parseBoolean(testData[test][4]) && Month.equals(CONV_TABLE_TYPES.get(convTypeIndex));
                if( (CONV_TABLE[convType] == 'x' || CONV_TABLE[convType] == 'X') && !skip  )
                {
                    String value = getStringFromCalendarValue(calendar,CONV_TABLE_TYPES.get(convTypeIndex));
                    String convInfo = "Converting from " + testData[test][1] + " to " + CONV_TABLE_TYPES.get(convTypeIndex) + ".  Comparing control='" + testData[test][3] + "' to test='" + value + "'.";
                    compareFields(testData[test][3], value, COMPARISON_FIELDS.get(convTypeIndex), convInfo  );
                }
            }
        }
    }
    
    /*
     * Compares to GregorianCalendar values will assert failure, if needed.  
     */
    private void compareFields(String controlString, String testString, int[] compare_fields, String convInfo ) throws DatatypeConfigurationException 
    {
        XMLGregorianCalendar controlXMLCal = DatatypeFactory.newInstance().newXMLGregorianCalendar(controlString);
        GregorianCalendar control = controlXMLCal.toGregorianCalendar(TimeZone.getTimeZone("GMT+00:00"), null, null);
        XMLGregorianCalendar testXMLCal = DatatypeFactory.newInstance().newXMLGregorianCalendar(testString);
        GregorianCalendar test = testXMLCal.toGregorianCalendar(TimeZone.getTimeZone("GMT+00:00"), null, null);
        //System.out.println(controlString + " ___ " + testString + " >>>> " + convInfo);        
        for (int i = 0; i < compare_fields.length; i++) 
        {
            int expectedValue = control.get(compare_fields[i]);
            int actualValue = test.get(compare_fields[i]);
            assertEquals(convInfo, expectedValue,actualValue);
        }
    }
    
    /*
     * Shared utility method that makes DataHelper calls to retrieve
     * SDO Date/Time strings from the source Date object.
     */
    private String getStringFromDateValue(Date value, String type)
    {
        if(DateTime.equals(type)) 
            return(dh.toDateTime(value));
        if(Day.equals(type)) 
            return(dh.toDay(value));
        if(Month.equals(type)) 
            return(dh.toMonth(value));
        if(MonthDay.equals(type)) 
            return(dh.toMonthDay(value));
        if(YearMonthDay.equals(type)) 
            return(dh.toYearMonthDay(value));
        if(YearMonth.equals(type)) 
            return(dh.toYearMonth(value));
        if(Year.equals(type)) 
            return(dh.toYear(value));
        if(Time.equals(type)) 
            return(dh.toTime(value));
        
        return(null);
    }
    
    /*
     * Shared utility method that makes DataHelper calls to retrieve
     * SDO Date/Time strings from the source Calendar object.
     */
    private String getStringFromCalendarValue(Calendar value, String type)
    {
        if(DateTime.equals(type)) 
            return(dh.toDateTime(value));
        if(Day.equals(type)) 
            return(dh.toDay(value));
        if(Month.equals(type)) 
            return(dh.toMonth(value));
        if(MonthDay.equals(type)) 
            return(dh.toMonthDay(value));
        if(YearMonthDay.equals(type)) 
            return(dh.toYearMonthDay(value));
        if(YearMonth.equals(type)) 
            return(dh.toYearMonth(value));
        if(Year.equals(type)) 
            return(dh.toYear(value));
        if(Time.equals(type)) 
            return(dh.toTime(value));
        
        return(null);
    }
    
    /**
     * testToDateFormats verifies that strings that should be recognized by
     * toDate do not result in a null Date value when passed to
     * DataHelper.toDate(String).
     */
    @org.junit.Test
    public void testToDateRoundtrip() {
        Roundtrips(true);
    }
      
    /**
     * testToCalendarFormats verifies that strings that should be recognized by
     * toCalendar do not result in a null Calendar value when passed to
     * DataHelper.toCalendar(String).
     */
    @org.junit.Test
    public void testToCalendarRoundtrip() 
    {
        Roundtrips(false);
    }
    
    /*
     * Utility method used by the Date and Calendar roundtrip tests.
     */
    public void Roundtrips(boolean toDate) 
    {
        for (int i = 0; i < testData.length; i++) 
        {
            String test = "";
            Object value; 
            String control = toDate?testData[i][2]:testData[i][3];
            
            if( control != null )
            {    
                if( toDate )
                    test = getStringFromDateValue(dh.toDate(testData[i][0]),testData[i][1]);
                else
                    test = getStringFromCalendarValue(dh.toCalendar(testData[i][0]),testData[i][1]);
            
                if( control.indexOf('.')<0 && test.indexOf('.')>=0 )
                {
                    // In some cases the implementation adds '.000' milliseconds.
                    // This value can be discarded for this comparison.
                    int idx = test.indexOf('.');
                    int len = 1;
                    while((idx+len)<test.length() && Character.isDigit(test.charAt(idx+len)))
                        len++;
                    test = test.substring(0,idx) + test.substring(idx+len);
                }
                
                String msg = "";
                if( toDate )
                    msg = "DataHelper.toDate() failed to round-trip back to original format for control value '" + testData[i][0] + "'.  The conversion type is '" + testData[i][1] +"'.";
                else
                    msg = "DataHelper.toCalendar() failed to round-trip back to original format for control value '" + testData[i][0] + "'.  The conversion type is '" + testData[i][1] +"'.";
                assertEquals(msg,control,test);        
            }
        }
    }   
    
    /**
     * Verifies the DataHelper.convert Property method. <p/>
     * @throws Exception 
     * @see commonj.sdo.DataHelper#convert(Property,Object)
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=54">2.1 spec section 3.14</a>
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=138">2.1 spec section 16</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.14")},
            testFor={@Operation(className="DataHelper",methodName="convert",signature="Property,Object")})    
    @Test
    public void testPropertyConvert() throws Exception 
    {
        for( int i=0; i<TYPE_CONV_TABLE.length; i++ )
        {
            if( TYPE_CONV_TABLE[i] == 'x' )
            {
                int toType   = i/TYPE_NAMES.length;
                int fromType = i%TYPE_NAMES.length;
                Object value = getTestValue(fromType,toType);
                Object newVal = getScope().getDataHelper().convert(
                                       propContainerType.getProperty(TYPE_NAMES[toType]), 
                                       value);
                assertEquals("Conversion did not yield expected value for property type '" + 
                             TYPE_NAMES[toType] + "' converting from type '" + TYPE_NAMES[fromType] + "'",
                             0,COMPARE_ANY.compare(value,newVal));
            }
        }
    
    }
    
    /**
     * Verifies the DataHelper.convert Type method. <p/>
     * @throws Exception 
     * @see commonj.sdo.DataHelper#convert(Type,Object)
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=54">2.1 spec section 3.14</a>
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=138">2.1 spec section 16</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.14")},
            testFor={@Operation(className="DataHelper",methodName="convert",signature="Type,Object")})    
    @org.junit.Test
    public void testTypeConvert() throws Exception {
        for( int i=0; i<TYPE_CONV_TABLE.length; i++ )
        {
            if( TYPE_CONV_TABLE[i] == 'x' )
            {
                int toType   = i/TYPE_NAMES.length;
                int fromType = i%TYPE_NAMES.length;
                Object value = getTestValue(fromType,toType);
                Object newVal = getScope().getDataHelper().convert(
                                       propContainerType.getProperty(TYPE_NAMES[toType]).getType(), 
                                       value);
                assertEquals("Conversion did not yield expected value for type '" + 
                             TYPE_NAMES[toType] + "' converting from type '" + TYPE_NAMES[fromType] + "'",
                             0,COMPARE_ANY.compare(value,newVal));
            }
        }
    }
}
