/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 *  
 */
package com.ibm.sdo.tck.tests.helpers;

import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import java.io.IOException;
import java.io.InputStream;
import java.math.BigDecimal;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;


import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import com.ibm.sdo.tck.framework.TCKTestBase;
import com.ibm.sdo.tck.util.TCKUtil;

import commonj.sdo.ChangeSummary;
import commonj.sdo.DataObject;
import commonj.sdo.Sequence;
import commonj.sdo.Type;
import commonj.sdo.helper.HelperContext;

public class EqualityHelperTest extends TCKTestBase 
{
  protected DataObject testDO;
  protected HelperContext scope;
  
  @Test
  public void testShallow()
  {
      // Validate simple duplicate DO
      DataObject copyDO = scope.getCopyHelper().copy(testDO);
      assertTrue("EqualityHelper.equalShallow() failed to find equivalent DOs equal",scope.getEqualityHelper().equalShallow(testDO, copyDO));
      
      // Now make a change in the contained DO, validate the equalShallow is still true
      List <DataObject> manyList = copyDO.getList("containMany");
      DataObject containedDO = manyList.get(0);
      containedDO.setString("stringVal", "a different string");
      
      assertTrue("EqualityHelper.equalShallow() failed to find equivalent DOs equal.  Contained objects are different.", scope.getEqualityHelper().equalShallow(testDO, copyDO));
      
  }
  
  @Test
  public void testDeep()
  {
      // Validate duplicate DO
      DataObject copyDO = scope.getCopyHelper().copy(testDO);
      assertTrue("EqualityHelper.equal() failed to find equivalent DOs equal", scope.getEqualityHelper().equal(testDO, copyDO));      
  }
  
  @Test
  public void testShallowUnequal()
  {
      // Validate simple differences are found
      DataObject copyDO = scope.getCopyHelper().copy(testDO);
      copyDO.setInt("intVal", 8888);
      assertFalse("EqualityHelper.equalShallow() failed to find unequivalent DOs unequal", scope.getEqualityHelper().equalShallow(testDO, copyDO));
      
  }
  
  @Test
  public void testDeepUnequal()
  {
      // Make a change in the contained DO, validate the equal catches the difference
      DataObject copyDO = scope.getCopyHelper().copy(testDO);
      List <DataObject> manyList = copyDO.getList("containMany");
      DataObject containedDO = manyList.get(0);
      containedDO.setString("stringVal", "a different string");      
      assertFalse("EqualityHelper.equal() failed to find unequivalent DOs unequal", scope.getEqualityHelper().equal(testDO, copyDO));
}
  
  @Test
  public void testSequenced() throws IOException
  {
      URL url = getClass().getResource("/sharedFiles/sequenceTest.xsd");
      InputStream inputStream = url.openStream();
      getScope().getXSDHelper().define(inputStream, url.toString());
      inputStream.close();      
      
      Type openType = scope.getTypeHelper().getType("http://www.example.com/cts/SequenceTest", "OpenDataObject");
      DataObject seq1DO = scope.getDataFactory().create( openType );
      DataObject seq2DO = scope.getDataFactory().create( openType );
      
      assertTrue(seq1DO.getType().isSequenced());
      
      Sequence seq1 = seq1DO.getSequence();
      seq1.add("one","a");
      seq1.add("two","b");
      seq1.add("three","c");

      Sequence seq2 = seq2DO.getSequence();
      seq2.add("one","a");
      seq2.add("two","b");
      seq2.add("three","c");
      
      assertTrue("EqualityHelper.equal() failed to find equivalent sequenced DOs equal", scope.getEqualityHelper().equal(seq1DO, seq2DO));
      assertTrue("EqualityHelper.equalShallow() failed to find equivalent sequenced DOs equal", scope.getEqualityHelper().equalShallow(seq1DO, seq2DO));
      
      seq2.move(2, 0);
      
      assertFalse("EqualityHelper.equal() failed to find unequivalent sequenced DOs unequal", scope.getEqualityHelper().equal(seq1DO, seq2DO));
      assertFalse("EqualityHelper.equalShallow() failed to find unequivalent sequenced DOs unequal", scope.getEqualityHelper().equalShallow(seq1DO, seq2DO));
  }
  
  @Test
  public void testMany() throws IOException
  {      
      URL url = getClass().getResource("/EqualityHelperTest/equalityHelperMany.xsd");
      InputStream inputStream = url.openStream();
      getScope().getXSDHelper().define(inputStream, url.toString());
      inputStream.close();      
      
      Type manyType = scope.getTypeHelper().getType("http://www.example.com/EqualityHelperTest", "ManyType");
      DataObject many1DO = scope.getDataFactory().create( manyType );
      DataObject many2DO = scope.getDataFactory().create( manyType );
      
      List <String> list1 = new ArrayList();
      list1.add("red");
      list1.add("blue");
      list1.add("green");
      
      List <String> list2 = new ArrayList();
      list2.add("red");
      list2.add("blue");
      list2.add("green");
      
      many1DO.setList("colors", list1);
      many2DO.setList("colors", list2);
      
      assertTrue("EqualityHelper.equal() failed to find equivalent many-valued DOs equal", scope.getEqualityHelper().equal(many1DO, many2DO));
      assertTrue("EqualityHelper.equalShallow() failed to find equivalent many-valued DOs equal", scope.getEqualityHelper().equalShallow(many1DO, many2DO));
      
      DataObject many3DO = scope.getDataFactory().create( manyType );
      
      List <String> list3 = new ArrayList();
      list3.add("green");
      list3.add("blue");
      list3.add("red");
      
      many3DO.setList("colors", list3);
      
      assertFalse("EqualityHelper.equal() failed to find unequivalent many-valued DOs unequal", scope.getEqualityHelper().equal(many1DO, many3DO));
      assertFalse("EqualityHelper.equalShallow() failed to find unequivalent many-valued DOs unequal", scope.getEqualityHelper().equalShallow(many1DO, many3DO));
  }
  
  @Test
  public void testChangeSummary() throws IOException
  {
      URL url = getClass().getResource("/sharedFiles/changeSummaryTest.xsd");
      InputStream inputStream = url.openStream();
      getScope().getXSDHelper().define(inputStream, url.toString());
      inputStream.close();      

      Type type = scope.getTypeHelper().getType("http://www.example.com/cstest","Root");
      
      DataObject cs1DO = scope.getDataFactory().create( type );
      DataObject cs2DO = scope.getDataFactory().create( type );           
      ChangeSummary cs1 = cs1DO.getChangeSummary();
      ChangeSummary cs2 = cs2DO.getChangeSummary();
      cs1.beginLogging();
      cs2.beginLogging();
      
      cs1DO.setString("symbol", "fbnt");
      cs1DO.setString("companyName", "FlyByNightTechnology");
      cs1DO.setBigDecimal("open1", new BigDecimal("1000.9"));
      cs1DO.setBigDecimal("price", new BigDecimal("1000.9"));

      cs2DO.setString("symbol", "fbnt");
      cs2DO.setString("companyName", "FlyByNightTechnology");
      cs2DO.setBigDecimal("open1", new BigDecimal("1000.9"));
      cs2DO.setBigDecimal("price", new BigDecimal("1000.9"));

      cs1.endLogging();
      cs2.endLogging();
      
      assertTrue("EqualityHelper.equal() failed to find equivalent change summary DOs equal", scope.getEqualityHelper().equal(cs1DO, cs2DO));
      
      // After some discussion, the corner case tested below is
      // being ignored for now.  EG issue #67.
      //cs1DO.setString("symbol", "fbnt2");
      //cs2.beginLogging();
      //cs2DO.setString("symbol", "fbnt2");
      //cs2.endLogging();
      
      //assertFalse("EqualityHelper.equal() failed to find unequivalent change summary DOs unequal", scope.getEqualityHelper().equal(cs1DO, cs2DO));

  }
  
  @Before
  public void setUp () throws Exception 
  {
    super.setUp();
    scope = getScope();
    TCKUtil.defineXSDAPITestTypes(scope);
    testDO = TCKUtil.createTestDataObject(scope, TCKUtil.API_TYPE);
    TCKUtil.populateAPITestObjectProperties(testDO, scope);
  }
    
  @After
  public void tearDown() throws Exception 
  {
    super.tearDown();
  }
}