/**
 *
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 */
package com.ibm.sdo.tck.tests.helpers;

import commonj.sdo.*;
import commonj.sdo.helper.*;
import org.junit.After;
import static org.junit.Assert.*;

import org.junit.Before;
import org.junit.Test;

import com.ibm.sdo.tck.framework.TCKTestBase;
import com.ibm.sdo.tck.framework.annotations.Operation;
import com.ibm.sdo.tck.framework.annotations.SpecRef;
import com.ibm.sdo.tck.framework.annotations.TCKTest;
import com.ibm.sdo.tck.util.EqualityTest;
import com.ibm.sdo.tck.util.TCKUtil;

import java.io.*;
import java.net.URL;
import java.util.List;


/**
 * Tests various concepts against XMLDocument, including ...
 * <ul>
 * <li> Retrieval of the root document
 * <li> Manage document encoding
 * <li> Manage XML version information
 * <li> Manage Schema location information
 * <li> Manage no namespace location information
 * </ul>
 * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=45">2.1 spec section 3.12</a>
 */

//The following empty meta-data tag is currently required in order for
//post-jar processing of the meta-data tags above methods below.  
@TCKTest()
public class XMLDocumentTest extends TCKTestBase {
    private static final String TEST_NS = "http://www.example.com/simple";
    private static final String TEST_ROOT_TYPE = "Quote";
    private static final String TEST_ROOT_ELEMENT = "stockQuote";
    
    /**
     * Verify getRootObject method. <p/>
     * @throws Exception 
     * @see commonj.sdo.XMLDocument#getRootObject()
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=53">2.1 spec section 3.13.3</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.13.3")},
        testFor={@Operation(className="XMLDocument",methodName="getRootObject",signature="")})
    @Test
    public void testRootObject() throws Exception {
        DataObject obj = getScope().getDataFactory().create(TEST_NS, TEST_ROOT_TYPE);
        obj.set("symbol", "TST");     
        obj.set("companyName", "Test Incorporated");     
        XMLDocument xmlDoc = createXMLDocument();
        DataObject newObj = xmlDoc.getRootObject();
        EqualityTest et = new EqualityTest(obj,newObj,true);        
        assertTrue(et.getFailure(),et.getResult());
    }   
    
    /**
     * Verify getRootElementURI method. <p/>
     * @throws Exception 
     * @see commonj.sdo.XMLDocument#getRootElementURI()
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=53">2.1 spec section 3.13.3</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.13.3")},
        testFor={@Operation(className="XMLDocument",methodName="getRootElementURI",signature="")})
    @Test
    public void testRootElementURI() throws Exception {
        XMLDocument xmlDoc = createXMLDocument();
        assertTrue(TEST_NS.equals(xmlDoc.getRootElementURI()));
    }   
    
    //String getRootElementName();  
    /**
     * Verify getRootElementName method. <p/>
     * @throws Exception 
     * @see commonj.sdo.XMLDocument#getRootElementName()
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=53">2.1 spec section 3.13.3</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.13.3")},
        testFor={@Operation(className="XMLDocument",methodName="getRootElementName",signature="")})
    @Test
    public void testRootElementName() throws Exception {
        XMLDocument xmlDoc = createXMLDocument();
        assertTrue(TEST_ROOT_ELEMENT.equals(xmlDoc.getRootElementName()));
    }   
    
    /**
     * Verifies the ability to manage encoding. <p/>
     * @throws Exception 
     * @see commonj.sdo.XMLDocument#getEncoding()
     * @see commonj.sdo.XMLDocument#setEncoding(String)
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=53">2.1 spec section 3.13.3</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.13.3")},
        testFor={@Operation(className="XMLDocument",methodName="getEncoding",signature=""),
                 @Operation(className="XMLDocument",methodName="setEncoding",signature="String")})
    @Test
    public void testEncoding() throws Exception {
        XMLDocument xmlDoc = createXMLDocument();
        assertTrue("UTF-8".equals(xmlDoc.getEncoding()));
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        XMLDocument xmlDoc2 = getScope().getXMLHelper().createDocument(xmlDoc.getRootObject(), xmlDoc.getRootElementURI(), xmlDoc.getRootElementName());
        xmlDoc2.setEncoding("ASCII");
        getScope().getXMLHelper().save(xmlDoc2, baos, null);        
        ByteArrayInputStream bais = new ByteArrayInputStream(baos.toString().getBytes());
        xmlDoc = getScope().getXMLHelper().load(bais);
        assertTrue("ASCII".equals(xmlDoc.getEncoding()) || "US-ASCII".equals(xmlDoc.getEncoding()) );
    }   
    
    /**
     * Verifies the XML declaration management methods. <p/>
     * @throws Exception 
     * @see commonj.sdo.XMLDocument#isXMLDeclaration()
     * @see commonj.sdo.XMLDocument#setXMLDeclaration(boolean)
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=53">2.1 spec section 3.13.3</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.13.3")},
        testFor={@Operation(className="XMLDocument",methodName="isXMLDeclaration",signature=""),
                 @Operation(className="XMLDocument",methodName="setXMLDeclaration",signature="boolean")})
    @Test
    public void testXMLDeclaration() throws Exception {
        XMLDocument xmlDoc = createXMLDocument();
        assertTrue(xmlDoc.isXMLDeclaration());
        xmlDoc.setXMLDeclaration(false);
        ByteArrayOutputStream baos = new ByteArrayOutputStream();;
        getScope().getXMLHelper().save(xmlDoc, baos, null);
        String xml = baos.toString();
        assertFalse( (xml.indexOf("<?xml") >= 0) );
        xmlDoc.setXMLDeclaration(true);
        baos = new ByteArrayOutputStream();;
        getScope().getXMLHelper().save(xmlDoc, baos, null);
        xml = baos.toString();
        assertTrue( (xml.indexOf("<?xml") >= 0) );
    }   

    /**
     * Verifies the ability to manage XML version. <p/>
     * @throws Exception 
     * @see commonj.sdo.XMLDocument#getXMLVersion()
     * @see commonj.sdo.XMLDocument#setXMLVersion(String)
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=53">2.1 spec section 3.13.3</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.13.3")},
        testFor={@Operation(className="XMLDocument",methodName="getXMLVersion",signature=""),
                 @Operation(className="XMLDocument",methodName="setXMLVersion",signature="String")})
    @Test
    public void testXMLVersion() throws Exception {
        XMLDocument xmlDoc = createXMLDocument();
        xmlDoc.setXMLVersion("1.0");
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        getScope().getXMLHelper().save(xmlDoc, baos, null);        
        ByteArrayInputStream bais = new ByteArrayInputStream(baos.toString().getBytes());
        xmlDoc = getScope().getXMLHelper().load(bais);
        assertTrue("1.0".equals(xmlDoc.getXMLVersion()));
    }       

    /**
     * Verifies the ability to manage schema location. <p/>
     * @throws Exception 
     * @see commonj.sdo.XMLDocument#getSchemaLocation()
     * @see commonj.sdo.XMLDocument#setSchemaLocation(String)
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=53">2.1 spec section 3.13.3</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.13.3")},
        testFor={@Operation(className="XMLDocument",methodName="getSchemaLocation",signature=""),
                 @Operation(className="XMLDocument",methodName="setSchemaLocation",signature="String")})
    @Test
    public void testSchemaLocation() throws Exception {
        XMLDocument xmlDoc = createXMLDocument();
        assertTrue(xmlDoc.getSchemaLocation() == null);
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        XMLDocument xmlDoc2 = getScope().getXMLHelper().createDocument(xmlDoc.getRootObject(), xmlDoc.getRootElementURI(), xmlDoc.getRootElementName());
        xmlDoc2.setSchemaLocation(TEST_NS + " simple.xsd");
        getScope().getXMLHelper().save(xmlDoc2, baos, null); 
        ByteArrayInputStream bais = new ByteArrayInputStream(baos.toString().getBytes());
        xmlDoc = getScope().getXMLHelper().load(bais);
        assertTrue((TEST_NS + " simple.xsd").equals(xmlDoc.getSchemaLocation())); 
    }       

    /**
     * Verifies the ability to manage no namespace schema location. <p/>
     * @throws Exception 
     * @see commonj.sdo.XMLDocument#getNoNamespaceSchemaLocation()
     * @see commonj.sdo.XMLDocument#setNoNamespaceSchemaLocation(String)
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=53">2.1 spec section 3.13.3</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.13.3")},
        testFor={@Operation(className="XMLDocument",methodName="getNoNamespaceSchemaLocation",signature=""),
                 @Operation(className="XMLDocument",methodName="setNoNamespaceSchemaLocation",signature="String")})
    @Test
    public void testNoNamespaceSchemaLocation() throws Exception {
        XMLDocument xmlDoc = createXMLDocument();
        assertTrue(xmlDoc.getNoNamespaceSchemaLocation() == null);        
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        XMLDocument xmlDoc2 = getScope().getXMLHelper().createDocument(xmlDoc.getRootObject(), xmlDoc.getRootElementURI(), xmlDoc.getRootElementName());
        xmlDoc2.setNoNamespaceSchemaLocation("schema.xsd");
        getScope().getXMLHelper().save(xmlDoc2, baos, null);
        ByteArrayInputStream bais = new ByteArrayInputStream(baos.toString().getBytes());
        xmlDoc = getScope().getXMLHelper().load(bais);
        assertTrue("schema.xsd".equals(xmlDoc.getNoNamespaceSchemaLocation())); 
    }       

    
    @Override
    @Before
    public void setUp() throws Exception {
        super.setUp();
        URL url = TCKUtil.class.getResource("/sharedFiles/simple.xsd");
        InputStream inputStream = url.openStream();
        List types = getScope().getXSDHelper().define(inputStream, url.toString());
        inputStream.close();
        
    }
    
    @Override
	@After
    public void tearDown() throws Exception {
      super.tearDown();
    }
    
    private XMLDocument createXMLDocument() throws Exception {
        URL url = TCKUtil.class.getResource("/XMLDocumentTest/XMLDocumentTest.xml");
        InputStream inputStream = url.openStream();
        XMLDocument doc = getScope().getXMLHelper().load(inputStream);
        inputStream.close();
        return doc;
    }
}
