/**
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 *  
 *  $Rev: 537209 $  $Date: 2008/01/31 17:01:46 $
 */
package com.ibm.sdo.tck.tests.helpers;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.Iterator;
import java.util.List;

import javax.xml.namespace.QName;
import javax.xml.transform.stream.StreamSource;

import org.apache.ws.commons.schema.XmlSchema;
import org.apache.ws.commons.schema.XmlSchemaCollection;
import org.apache.ws.commons.schema.XmlSchemaObjectTable;
import org.apache.ws.commons.schema.XmlSchemaType;

import org.junit.After;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import com.ibm.sdo.tck.framework.TCKTestBase;
import com.ibm.sdo.tck.framework.annotations.Operation;
import com.ibm.sdo.tck.framework.annotations.SpecRef;
import com.ibm.sdo.tck.framework.annotations.TCKTest;
import com.ibm.sdo.tck.util.TCKUtil;
import com.ibm.sdo.tck.util.XMLCompareUtil;

import commonj.sdo.Property;
import commonj.sdo.Type;
import commonj.sdo.helper.HelperContext;
import commonj.sdo.helper.TypeHelper;
import commonj.sdo.helper.XSDHelper;

/**
 * Tests XSD serialization/deserialization.<p/>
 * This tests requires extension by import or creation of further tests.
 * It currently only tests one flavour of {@link XSDHelper#define(java.io.InputStream, String)} and {@link XSDHelper#generate(List)})
 * 
 * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=52">2.1 spec section 3.13</a>
 * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=56">2.1 spec section 4</a>
 * 
 */
@TCKTest(specRef={@SpecRef(version="2.1",section = "3.13",note={""})},
    testFor={@Operation(className="XSDHelper",methodName="define",signature="InputStream,String",note="class tests for behaviour with and without schemalocation values"),
             @Operation(className="HelperContext",methodName="getXSDHelper",signature="")})
public class XSDHelperTest extends TCKTestBase {
    HelperContext scope;
    private static final String TEST_MODEL = "/sharedFiles/simple.xsd";
    private static final String TEST_MODEL_APPINFO = "/XSDHelperTest/simpleWithAppinfo.xsd";
    private static final String TEST_MODEL_ATTRIBUTE = "/XSDHelperTest/simpleWithAttribute.xsd";
    private static final String TEST_MODEL_MIXED = TCKUtil.TEST_MODEL;
    private XmlSchemaCollection col = new XmlSchemaCollection();
    private static URL modelURL;
    private static URL appInfoModelURL;
    private static URL attributeModelURL;
    private static URL mixedModelURL;

    /**
     * Obtains test model resource. Runs once before any of the test methods.
     */
    @BeforeClass
    public static void obtainResource() {
        modelURL         = XSDHelperTest.class.getResource(TEST_MODEL);
        appInfoModelURL  = XSDHelperTest.class.getResource(TEST_MODEL_APPINFO);
        attributeModelURL= XSDHelperTest.class.getResource(TEST_MODEL_ATTRIBUTE);
        mixedModelURL    = XSDHelperTest.class.getResource(TEST_MODEL_MIXED);
    }
    
    @Before 
    public void setUp() throws Exception
    {
      super.setUp();
      scope = getScope();
    }
    
    @After
    public void tearDown() throws Exception {
      super.tearDown();
    }

    /**
     * Verifies the performance of XSDHelper.define() when a SchemaLocation is provided.
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=52">2.1 spec section 3.13</a>
     * @see commonj.sdo.XSDHelper#define(InputStream, String)
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.13.3")},
        testFor={@Operation(className="XSDHelper",methodName="define",signature="InputStream,String")}) 
    @Test
    public void testDefineWithLocation() {
        try {

            XSDHelper xsdHelper = scope.getXSDHelper();
            List types = xsdHelper.define(modelURL.openStream(), modelURL.toString());
            checkTypes(modelURL, scope.getTypeHelper(), types);
         } catch (Exception e) {
            fail("Exception calling xsdHelper.define" + e.toString());
        }
    }

    /**
     * Verifies the performance of XSDHelper.define() when a SchemaLocation is provided.
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=52">2.1 spec section 3.13</a>
     * @see commonj.sdo.XSDHelper#define(Reader, String)
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.13.3")},
        testFor={@Operation(className="XSDHelper",methodName="define",signature="Reader,String")})   
    @Test
    public void testDefineWithReaderAndLocation() {
        try {

            XSDHelper xsdHelper = scope.getXSDHelper();
            List types = xsdHelper.define(new InputStreamReader(modelURL.openStream()), modelURL.toString());
            checkTypes(modelURL, scope.getTypeHelper(), types);
         } catch (Exception e) {
            fail("Exception calling xsdHelper.define" + e.toString());
        }
    }
    
    /**
     * Utility method to ensure that the set of types in the list includes all those explicity
     * defined in the schema<p>
     * Checking xsd type definition post condition ---<br/>
     * <i>for all t in schemaTypes there exists a t' in types such that QName(t) == QName(t')</i>
     * @param modelUrl location of the schema related to the set of types
     * @param typeHelper associated with the scope for the types in the list
     * @param types
     */
    private void checkTypes(URL modelUrl, TypeHelper typeHelper, List types) {
      try {
          XmlSchema schema = col.read(new StreamSource(modelUrl.openStream()), null);
          XmlSchemaObjectTable schemaTypes = schema.getSchemaTypes();
          
          Iterator<XmlSchemaType> it = schemaTypes.getValues();
          XmlSchemaType schemaType;
          while (it.hasNext()){
              schemaType = it.next();
              QName qname = schemaType.getQName();
              Type sdoType = typeHelper.getType(qname.getNamespaceURI(), qname.getLocalPart());
              assertNotNull("Type not known to SDO environment: "+ qname, sdoType);
              assertTrue("Sdo type not created from this invocation of type definition", types.contains(sdoType));
          }
      } catch (IOException e) {

          fail("Exception parsing schema" + e);
      }
  }

    /**
     * Verifies the performance of XSDHelper.define() when a SchemaLocation is not provided.
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=52">2.1 spec section 3.13</a>
     * @see commonj.sdo.XSDHelper#define(InputStream, String)
     */
    @Test
    public void testDefineWithNoLocation() {
        try {
            XSDHelper xsdHelper = scope.getXSDHelper();
            List types = xsdHelper.define(XSDHelperTest.class.getResourceAsStream(TEST_MODEL), null);
            checkTypes(modelURL, scope.getTypeHelper(), types);
            //assertEquals("XSDHelper.define() did not create the expected number of Types", 2, types.size());

        } catch (Exception e) {
            e.printStackTrace();
            fail("Exception caught" + e.toString());
        }
    }

    /**
     * Verifies that duplicate Types are not redefined.
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=52">2.1 spec section 3.13</a>
     * @see commonj.sdo.XSDHelper#define(InputStream, String)
     */
    @Test
    public void testDuplicateDefineWithLocation() {
        try {
            XSDHelper xsdHelper = scope.getXSDHelper();
            List types = xsdHelper.define(modelURL.openStream(), modelURL.toString());
            assertTrue("XSDHelper.define() did not create the expected number of Types", types.size() > 0);
            // redefine type
            List duplicateTypes = xsdHelper.define(modelURL.openStream(), modelURL.toString());
            assertEquals("XSDHelper.define() did not create the expected number of Types", 0, duplicateTypes.size());

        } catch (Exception e) {
            e.printStackTrace();
            fail("Exception caught" + e.toString());
        }
    }

    
    /**
     * Verifies the ability to retrieve schema AppInfo. <p/>
     * @see commonj.sdo.XSDHelper#getAppinfo(Property,String)
     * @see commonj.sdo.XSDHelper#getAppinfo(Type,String)
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=53">2.1 spec section 3.13.3</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.13.3")},
        testFor={@Operation(className="XSDHelper",methodName="getAppinfo",signature="Property,String"),
                 @Operation(className="XSDHelper",methodName="getAppinfo",signature="Type,String")})
    @Test
    public void testAppInfo() {
        try {
            scope.getXSDHelper().define(appInfoModelURL.openStream(), appInfoModelURL.toString());
            Type t = scope.getTypeHelper().getType("http://www.example.com/simple","Quote");
            String appInfo = scope.getXSDHelper().getAppinfo(t, "name");
                        
            URL url = TCKUtil.class.getResource("/XSDHelperTest/appInfoResult.xml");
            InputStream inputStream = url.openStream();
            assertTrue(XMLCompareUtil.compareXML(inputStream, appInfo));
            inputStream.close();
        } 
        catch (IOException e) {
            fail("Exception parsing schema" + e);
        }
    }
    
    /**
     * Verifies the ability to get schmea's local name for property or type. <p/>
     * @see commonj.sdo.XSDHelper#getLocalName(Property)
     * @see commonj.sdo.XSDHelper#getLocalName(Type)
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=53">2.1 spec section 3.13.3</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.13.3")},
        testFor={@Operation(className="XSDHelper",methodName="getLocalName",signature="Property"),
                 @Operation(className="XSDHelper",methodName="getLocalName",signature="Type")})
    @Test
    public void testLocalName() {
        try {
            scope.getXSDHelper().define(modelURL.openStream(), modelURL.toString());
            Type t = scope.getTypeHelper().getType("http://www.example.com/simple","Quote");
            assertTrue("Quote".equals(scope.getXSDHelper().getLocalName(t)));
            
            Property p = t.getProperty("symbol");
            assertTrue("symbol".equals(scope.getXSDHelper().getLocalName(p)));
        } 
        catch (IOException e) {
            fail("Exception parsing schema" + e);
        }
    }
    
    /**
     * Verifies the ability to retrieve a property's namespace uri. <p/>
     * @see commonj.sdo.XSDHelper#getNamespaceURI(Property)
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=53">2.1 spec section 3.13.3</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.13.3")},
        testFor={@Operation(className="XSDHelper",methodName="getNamespaceURI",signature="Property")})
    @Test
    public void testNamespaceURI() {
        try {
            String NS = "http://www.example.com/simple";
            scope.getXSDHelper().define(modelURL.openStream(), modelURL.toString());
            Type t = scope.getTypeHelper().getType("http://www.example.com/simple","Quote");
            
            // Test type
            assertNotNull(t);
            assertEquals(NS,scope.getXSDHelper().getNamespaceURI(t));
            
            // Test properties
            Property p = t.getProperty("symbol");
            assertNotNull(p);
            assertEquals("",scope.getXSDHelper().getNamespaceURI(p)); // JSR235 Issue 4
            p = scope.getTypeHelper().getOpenContentProperty(NS, "stockQuote");
            assertNotNull(p);
            assertEquals(NS,scope.getXSDHelper().getNamespaceURI(p));
            
        } 
        catch (IOException e) {
            fail("Exception parsing schema" + e);
        }
    }
    
    /**
     * Verifies the ability to test for Attributes and Elements. <p/>
     * @see commonj.sdo.XSDHelper#isAttribute(Property)
     * @see commonj.sdo.XSDHelper#isElement(Property)
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=53">2.1 spec section 3.13.3</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.13.3")},
        testFor={@Operation(className="XSDHelper",methodName="isAttribute",signature="Property"),
                 @Operation(className="XSDHelper",methodName="isElement",signature="Property")})
    @Test
    public void testAttributeOrElement() {
        try {
            scope.getXSDHelper().define(attributeModelURL.openStream(), attributeModelURL.toString());
            Type t = scope.getTypeHelper().getType("http://www.example.com/simple","Quote");
            Property p1 = t.getProperty("attr1");
            Property p2 = t.getProperty("symbol");
            assertTrue(scope.getXSDHelper().isAttribute(p1));
            assertFalse(scope.getXSDHelper().isAttribute(p2));
            assertFalse(scope.getXSDHelper().isElement(p1));
            assertTrue(scope.getXSDHelper().isElement(p2));
        } 
        catch (IOException e) {
            fail("Exception parsing schema" + e);
        }
    }   
    
    /**
     * Test for mixed type <p/>
     * @see commonj.sdo.XSDHelper#isMixed(Type)
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=53">2.1 spec section 3.13.3</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.13.3")},
        testFor={@Operation(className="XSDHelper",methodName="isMixed",signature="Type")})
    @Test
    public void testMixed() {
        try {
            scope.getXSDHelper().define(mixedModelURL.openStream(), mixedModelURL.toString());
            Type t = scope.getTypeHelper().getType("http://www.example.com/tck_api_test","Sequenced");
            assertTrue(scope.getXSDHelper().isMixed(t));
            
            scope.getXSDHelper().define(modelURL.openStream(), modelURL.toString());
            t = scope.getTypeHelper().getType("http://www.example.com/simple","Quote");
            assertFalse(scope.getXSDHelper().isMixed(t));
        } 
        catch (IOException e) {
            fail("Exception parsing schema" + e);
        }
    }
               
    /**
     * Verify the expected behavior of TypeHelper.define(DataObject) when a Type
     * of the same name has already been defined differently.
     * 
     * @throws Exception
     */
    @Test
    public void testRedefinition() throws Exception {
        String schemaStart = "<xsd:schema targetNamespace=\"redefinition.test\" xmlns:xsd=\"http://www.w3.org/2001/XMLSchema\" xmlns:test=\"redefinition.test\">"; 
        String schemaEnd   = "</xsd:schema>";
        String strTypeA1   = "<xsd:complexType name=\"TypeA\">" +
                             "<xsd:sequence>" +
                             "<xsd:element name=\"strA1\" type=\"xsd:string\"/>" +
                             "<xsd:element name=\"strZ\" type=\"xsd:string\"/>" +
                             "</xsd:sequence>" +
                             "</xsd:complexType>";
        String strTypeA2   = "<xsd:complexType name=\"TypeA\">" +
                             "<xsd:sequence>" +
                             "<xsd:element name=\"strA2\" type=\"xsd:string\"/>" +
                             "</xsd:sequence>" +
                             "</xsd:complexType>";
        //String strTypeB    = "<xsd:complexType name=\"TypeB\">" +
        //                     "<xsd:sequence>" +
        //                     "<xsd:element name=\"strB\" type=\"xsd:string\"/>" +
        //                     "</xsd:sequence>" +
        //                     "</xsd:complexType>";
        
      scope.getXSDHelper().define(schemaStart+strTypeA1+schemaEnd);
      scope.getXSDHelper().define(schemaStart+strTypeA2+schemaEnd);
      //scope.getXSDHelper().define(schemaStart+strTypeA2+strTypeB+schemaEnd);
      Type typeA = scope.getTypeHelper().getType("redefinition.test","TypeA");
      //Type typeB = scope.getTypeHelper().getType("redefinition.test","TypeB");
      
      assertNotNull(typeA);
      assertNotNull(typeA.getProperty("strA1"));
      assertNotNull(typeA.getProperty("strZ"));
      //assertNotNull(typeB);
      //assertNotNull(typeB.getProperty("strB"));
    }
}
