/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 *
 */
package com.ibm.sdo.tck.tests.metadata;

import java.util.List;
import java.math.BigInteger;


import commonj.sdo.Type;
import commonj.sdo.Property;
import commonj.sdo.DataObject;

import org.junit.Test;

import com.ibm.sdo.tck.framework.annotations.Operation;
import com.ibm.sdo.tck.framework.annotations.SpecRef;
import com.ibm.sdo.tck.framework.annotations.TCKTest;
import com.ibm.sdo.tck.testcasebase.XSDTypeTestBase;

import static org.junit.Assert.*;


@TCKTest(
    specRef={
        @SpecRef(
            version="2.1",section = "3.13"
        ),
        @SpecRef(
            version="2.1",section = "9"
        ),
        @SpecRef(
            version="2.1",section = "9.7"
        )
    },
    testFor={
      @Operation(
          className="XSDHelper",
          methodName="define",
          signature="InputStream,String",
          note="class tests for variations on simple type definitions, extensions, restrictions, derived types, unions, annotations, enumerations ..."
      )
    }
    
  )
public class XSDSimpleTypeTest extends XSDTypeTestBase {

    /**
     * <simpleType> with a restriction from type xs:string
     *
     * @throws Exception
     */
    @TCKTest(
      specRef={
          @SpecRef(
              version="2.1",section = "3.13"
          ),
          @SpecRef(
              version="2.1",section = "9",
              note="haven't found specific spec text yet"
          )
      },
      testFor={
        @Operation(
            className="XSDHelper",
            methodName="define",
            signature="InputStream,String",
            precondition={"schema contains simple type as a restriction on xsd:string and with final=\"restriction\""}
        )
      }
      
    )
    @Test
    public void testRestriction() throws Exception {
        List types = define("/XSDSimpleTypeTest/restriction.xsd");

        // at least one type should have been returned
        assertTrue(types.size() > 0);

        Type type = getType(types, "simpleTypeRestriction");
        assertEquals(false, type.isAbstract());
        assertEquals(true, type.isDataType());
        assertEquals(false, type.isOpen());
        assertEquals(false, type.isSequenced());
        assertEquals(0, type.getProperties().size());
        assertEquals(1, type.getBaseTypes().size());
        assertEquals(String.class, type.getInstanceClass());
        
        Type baseType = (Type) type.getBaseTypes().get(0);
        assertEquals("commonj.sdo", baseType.getURI());
        assertEquals("String", baseType.getName());
    }

    @Test
    public void testRestrictionOptimization() throws Exception {
        List types = define("/XSDSimpleTypeTest/restrictionOptimization.xsd");
        
        // at least one type should have been returned
        assertTrue(types.size() > 0);
        
        Type type1 = getType(types, "fontbynumber");

        assertEquals("fontbynumber", type1.getName());

        assertTrue(type1.isDataType());
        assertEquals(int.class,type1.getInstanceClass());
    }
    
    /**
     * <element> using anonymous <simpleType>. We expect the SDO Type to have the same name as the element
     * as per page 80 of the Java spec.
     */
    @Test
    @TCKTest(
        specRef={
            @SpecRef(
                version="2.1",section = "3.13"
            ),
            @SpecRef(
                version="2.1",section = "9.2.2",
                page=80,
                note="row 2"
            )
        },
        testFor={
          @Operation(
              className="XSDHelper",
              methodName="define",
              signature="InputStream,String",
              precondition={"schema contains <element> using anonymous <simpleType>"},
              postcondition={"The SDO Type has the same name as the containing element"},
              // TODO resolve
              note="doubt expressed in comment in code that the spec is correct here"
          )
        }
        
      )
    public void testAnonOpenContentProperty() throws Exception {
        List types = define("/XSDSimpleTypeTest/anonymous.xsd");

        // at least one type should have been returned
        assertTrue(types.size() > 0);

        // get open content property defined by global element in XSD
        String uri = "http://www.example.com/simpleType/anonymous";
        Property property = typeHelper.getOpenContentProperty(uri, "simpleTypeAnon");

        assertNotNull(property);

        // FIXME: this comment (& assertions?)has no place in an implementation neutral test suite
        // The spec maps the name of anonymous types to the name
        // of their enclosing element. This is a broken method however,
        // as it can conflict with same named types and other same
        // named elements with anonymous types. So currently we are
        // intentionally non-compliant. using unique names for
        // anonymous types

        Type type = property.getType();
        assertEquals(false, type.isAbstract());
        assertEquals(true, type.isDataType());
        assertEquals(false, type.isOpen());
        assertEquals(false, type.isSequenced());
        assertEquals(0, type.getProperties().size());
        assertEquals(1, type.getBaseTypes().size());
        assertEquals(String.class, type.getInstanceClass());
        
        Type baseType = (Type) type.getBaseTypes().get(0);
        assertEquals("commonj.sdo", baseType.getURI());
        assertEquals("String", baseType.getName());
    }

    /**
     * <simpleType> with final="list"
     *
     * @throws Exception
     */
    @Test
    @TCKTest(
        specRef={
            @SpecRef(
                version="2.1",section = "9",
                note="seek more specific reference"
            )
        },
        testFor={
          @Operation(
              className="XSDHelper",
              methodName="define",
              signature="InputStream,String",
              precondition={"schema contains <element> with final=\"list\""}
          )
        } 
      )
    public void testFinalList() throws Exception {
        List types = define("/XSDSimpleTypeTest/finalList.xsd");

        // at least one type should have been returned
        assertTrue(types.size() > 0);

        Type type = getType(types, "simpleTypeFinalList");
        assertEquals(false, type.isAbstract());
        assertEquals(true, type.isDataType());
        assertEquals(false, type.isOpen());
        assertEquals(false, type.isSequenced());
        assertEquals(0, type.getProperties().size());
        assertEquals(1, type.getBaseTypes().size());
        assertEquals(String.class, type.getInstanceClass());

        Type baseType = (Type) type.getBaseTypes().get(0);
        assertEquals("commonj.sdo", baseType.getURI());
        assertEquals("String", baseType.getName());
    }

    /**
     * <simpleType> with final="union"
     *
     * @throws Exception
     */
    @Test
    public void testFinalUnion() throws Exception {
        List types = define("/XSDSimpleTypeTest/finalUnion.xsd");

        Type type = getType(types, "simpleTypeFinalUnion");
        assertEquals(false, type.isAbstract());
        assertEquals(true, type.isDataType());
        assertEquals(false, type.isOpen());
        assertEquals(false, type.isSequenced());
        assertEquals(0, type.getProperties().size());
        assertEquals(1, type.getBaseTypes().size());
        assertEquals(String.class, type.getInstanceClass());

        Type baseType = (Type) type.getBaseTypes().get(0);
        assertEquals("commonj.sdo", baseType.getURI());
        assertEquals("String", baseType.getName());
    }

    /**
     * <simpleType> with final="restriction"
     *
     * @throws Exception
     */
    @Test
    public void testFinalRestriction() throws Exception {
        List types = define("/XSDSimpleTypeTest/finalRestriction.xsd");

        // at least one type should have been returned
        assertTrue(types.size() > 0);

        Type type = getType(types, "simpleTypeFinalRestriction");
        assertEquals(false, type.isAbstract());
        assertEquals(true, type.isDataType());
        assertEquals(false, type.isOpen());
        assertEquals(false, type.isSequenced());
        assertEquals(0, type.getProperties().size());
        assertEquals(1, type.getBaseTypes().size());
        assertEquals(String.class, type.getInstanceClass());

        Type baseType = (Type) type.getBaseTypes().get(0);
        assertEquals("commonj.sdo", baseType.getURI());
        assertEquals("String", baseType.getName());
    }

    /**
     * @throws Exception
     */
    @Test
    //@Ignore
    public void testDerivedType() throws Exception {
        List types = define("/XSDSimpleTypeTest/derived.xsd");

        // at least one type should have been returned
        assertTrue(types.size() > 0);

        Type type0 = getType(types, "derivedType");
        Type type1 = getType(types, "baseType");

        assertEquals("baseType", type1.getName());
        assertEquals("derivedType", type0.getName());

        assertTrue(type0.isDataType());
        assertTrue(type1.isDataType());
        assertFalse(type0.isOpen());
        assertFalse(type0.isSequenced());
        assertEquals(0, type0.getProperties().size());
        assertEquals(1, type0.getBaseTypes().size());
        assertEquals(BigInteger.class, type0.getInstanceClass());

        List baseTypes = type1.getBaseTypes();
        assertEquals(1, baseTypes.size());
        Type baseType = (Type) baseTypes.get(0);
        assertEquals("commonj.sdo", baseType.getURI());
        assertEquals("Integer", baseType.getName());

        List derivedTypes = type0.getBaseTypes();
        assertEquals(1, derivedTypes.size());
        Type derivedType = (Type) derivedTypes.get(0);
        assertEquals("http://www.example.com/simpleType/derived", derivedType.getURI());
        assertEquals("baseType", derivedType.getName());

        
    }

    /**
     * @throws Exception
     */
    @Test
    public void testDerivedType2() throws Exception {
        List types = define("/XSDSimpleTypeTest/derived2.xsd");

        // at least one type should have been returned
        assertTrue(types.size() > 0);
        
        Type type0 = getType(types, "derivedType");
        Type type1 = getType(types, "baseType");

        assertEquals("derivedType", type0.getName());
        assertEquals("baseType", type1.getName());

        assertTrue(type0.isDataType());
        assertTrue(type1.isDataType());
        assertFalse(type0.isOpen());
        assertFalse(type0.isSequenced());
        assertEquals(0, type0.getProperties().size());
        assertEquals(1, type0.getBaseTypes().size());

        List baseTypes = type0.getBaseTypes();
        assertEquals(1, baseTypes.size());
        Type baseType = (Type) baseTypes.get(0);
        assertEquals("http://www.example.com/simpleType/derived2", baseType.getURI());
        assertEquals("baseType", baseType.getName());
        
        List baseTypes2 = type1.getBaseTypes();
        assertEquals(1, baseTypes2.size());
        
        Type baseType2 = (Type) baseTypes2.get(0);
        assertEquals("commonj.sdo", baseType2.getURI());
        assertEquals("Integer", baseType2.getName());
    }

    @Test
    public void testDerivedUnionType() throws Exception {
        List types = define("/XSDSimpleTypeTest/derivedUnion.xsd");

        // at least three types should have been returned
        assertTrue(types.size() > 2);
        
        Type type0 = getType(types, "fontbystringname");
        Type type1 = getType(types, "fontbynumber");

        assertEquals("fontbystringname", type0.getName());
        assertEquals("fontbynumber", type1.getName());

        assertTrue(type0.isDataType());
        assertTrue(type1.isDataType());
        assertEquals(String.class,type0.getInstanceClass());
        assertEquals(String.class,type1.getInstanceClass());

        Type fontsize = getType(types, "fontsize");
        assertEquals("fontsize", fontsize.getName());
        assertTrue(fontsize.getInstanceClass() == String.class);

    }

    /**
     * <simpleType> containing <annotation> (no effect on SDO type but want to make
     * sure we can parse it ok)
     *
     * @throws Exception
     */
    @Test
    public void testAnnotation() throws Exception {
        List types = define("/XSDSimpleTypeTest/annotation.xsd");

        // at least one type should have been returned
        assertTrue(types.size() > 0);

        Type type = getType(types, "simpleTypeAnnotated");
        assertEquals(false, type.isAbstract());
        assertEquals(true, type.isDataType());
        assertEquals(false, type.isOpen());
        assertEquals(false, type.isSequenced());
        assertEquals(0, type.getProperties().size());
        assertEquals(1, type.getBaseTypes().size());
        assertEquals(String.class, type.getInstanceClass());

        Type baseType = (Type) type.getBaseTypes().get(0);
        assertEquals("commonj.sdo", baseType.getURI());
        assertEquals("String", baseType.getName());
    }

    /**
     * <simpleType> containing <list>
     *
     * @throws Exception
     */
    @Test
    public void testTC200SimpleType() throws Exception {
        List types = define("/XSDSimpleTypeTest/list.xsd");

        Type type = getType(types, "simpleTypeList");
        assertEquals(false, type.isAbstract());
        assertEquals(true, type.isDataType());
        assertEquals(false, type.isOpen());
        assertEquals(false, type.isSequenced());
        assertEquals(0, type.getProperties().size());
        //Removing check below according to issue #57
        //assertEquals(0, type.getBaseTypes().size());
        assertEquals(List.class, type.getInstanceClass());
    }

    @Test
    public void testEnumeration() throws Exception {
        List types = define("/XSDSimpleTypeTest/enumeration.xsd");

        // at least three type should have been returned
        assertTrue(types.size() > 2);
        
        Type color = getType(types, "color");
        assertEquals(false, color.isAbstract());
        assertEquals(true, color.isDataType());
        assertEquals(false, color.isOpen());
        assertEquals(false, color.isSequenced());
        assertEquals(0, color.getProperties().size());
        assertEquals(1, color.getBaseTypes().size());
        assertEquals(String.class, color.getInstanceClass());

        Type thing = getType(types, "thing");
        Property colorofthing = thing.getProperty("colorofthing");
        Property sizeofthing = thing.getProperty("sizeofthing");

        assertEquals("thing", thing.getName());
        assertEquals("colorofthing", colorofthing.getName());
        assertEquals("sizeofthing", sizeofthing.getName());

        DataObject dobj = dataFactory.create(thing.getURI(), thing.getName());
        dobj.setInt("sizeofthing", 2);
        dobj.setString("colorofthing", "red");

        assertEquals("red", dobj.getString("colorofthing"));
        assertEquals(2, dobj.getInt("sizeofthing"));
    }

    /**
     * <simpleType> containing <union>
     *
     * @throws Exception
     */
    @Test
    public void testUnion() throws Exception {
        List types = define("/XSDSimpleTypeTest/union.xsd");

        // at least one type should have been returned
        assertTrue(types.size() > 0);

        Type type = getType(types, "simpleTypeUnion");
        assertEquals(false, type.isAbstract());
        assertEquals(true, type.isDataType());
        assertEquals(false, type.isOpen());
        assertEquals(false, type.isSequenced());
        assertEquals(0, type.getProperties().size());
        assertEquals(Object.class, type.getInstanceClass());
    }

}
