/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 *  
 */
package com.ibm.sdo.tck.tests.staticsdo;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertSame;
import static org.junit.Assert.assertTrue;

import java.io.IOException;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.List;


import org.junit.Before;
import org.junit.Test;

import com.ibm.sdo.tck.framework.TCKTestBase;
import com.ibm.sdo.tck.framework.annotations.SpecRef;
import com.ibm.sdo.tck.framework.annotations.TCKTest;
import com.ibm.sdo.tck.staticTest.CT2;
import com.ibm.sdo.tck.staticTest.ComplexType1;
import com.ibm.sdo.tck.util.EqualityTest;
import com.ibm.sdo.tck.util.TCKUtil;

import commonj.sdo.DataObject;
import commonj.sdo.Type;
import commonj.sdo.helper.DataFactory;
import commonj.sdo.helper.HelperContext;

/**
 * Simple static SDO tests
 * 
 */
@TCKTest(
        precondition={
            "Generated Java types ComplexType1 and CT2 are available relating to the types as defined in staticTest/simple.xsd",
            "The generator has heeded the sdoJava:package annotation on the schema"
        },
        specRef={@SpecRef(version="2.1.1 Early Draft", section="5")},
        note={"Individual methods are annotated",
              "generation of methods isSet[Property] and unset[Property] are not compliance points"
             }
        )
public class StaticSimpleTest extends TCKTestBase {
    
    private class MethodAttrs {
        public MethodAttrs(Class ret, String name, Class[] args) {
            returnType = ret;
            methodname = name;
            arguments = args;
        }
        public Class returnType;
        String methodname;
        public Class[] arguments;
    }
    

    
    /** IBM binary license notice. */
    final static String IBM_COPYRIGHT = "Licensed Materials - Property of IBM\n"
            + "\n"
            + "YOR7-2004-0083\n"
            + "\n"
            + "(c) Copyright IBM Corp. 2008. All Rights Reserved.\n"
            + "\n"
            + "US Government Users Restricted Rights - Use, duplication or disclosure\n"
            + "restricted by GSA ADP Schedule Contract with IBM Corp.";

    private HelperContext helperContext;
    private DataFactory dataFactory;

    private String RES_PATH_SIMPLE_STATIC_XSD = "/staticTest/simple.xsd";

    private String NS_SIMPLE_STATIC_XSD = "http://www.example.com/staticTest";

    // Tests methods

    
    @Test
    @TCKTest(
            precondition={
                    "The interface being tested is not the one provided by the TCK",
                    "all required methods are available on the generated type"
                  },
            specRef={@SpecRef(version="2.1.1 Early Draft",section="5.1")}
        )
    public void introspectInterfaceCT1Test() throws IllegalArgumentException, IllegalAccessException  {        

        MethodAttrs ct1Methods[] = {        
          new MethodAttrs(String.class, "getElem1", new Class[] {}),
          new MethodAttrs(void.class,   "setElem1", new Class[] {String.class}),
          new MethodAttrs(String.class, "getAttr1", new Class[] {}),
          new MethodAttrs(void.class,   "setAttr1", new Class[] {String.class}),
        };
        
        Class c = null;
        
        try {
            c = Class.forName("com.ibm.sdo.tck.staticTest.ComplexType1");
        } catch (ClassNotFoundException e) {
            assertTrue("Generated interface not found", false);
        }
        

        try {
            Field f = c.getDeclaredField("SHOULD_NOT_APPEAR_IN_GENERATED_INTERFACES");
            String implem_failed_to_substitute_msg = (String)f.get(null);
            assertTrue(implem_failed_to_substitute_msg, false);
        } catch (NoSuchFieldException e) {
            // arrival here implies success - i.e. something other than the TCK provided interfaces were found
            // on the classpath, ahead of the supplied versions
        }
        
        try {
            checkMethods(c, ct1Methods);
        }catch (NoSuchMethodException e) {
            assertTrue(false);
        }
        
        MethodAttrs primitiveAccessor1[] = {
            new MethodAttrs(boolean.class,"isBool",   new Class[] {}),
            new MethodAttrs(void.class,   "setBool",  new Class[] {boolean.class}),
        };

        // implementations have the choice of generating interfaces with primitives
        // or wrapped primitives
        try {
            checkMethods(c, primitiveAccessor1);
        } catch (NoSuchMethodException e1) {
            try {
                MethodAttrs primitiveAccessor2[] = {
                    new MethodAttrs(Boolean.class,"isBool",   new Class[] {}),
                    new MethodAttrs(void.class,   "setBool",  new Class[] {Boolean.class}),
                };
                checkMethods(c, primitiveAccessor2);
            } catch (NoSuchMethodException e) {
                assertTrue(false);
            }
        }

        

    }

    @Test
    @TCKTest(
            precondition={
                    "The interface being tested is not the one provided by the TCK",
                    "all required methods are available on the generated type"
                  },
            specRef={@SpecRef(version="2.1.1 Early Draft",section="5.1")}
        )
     public void introspectInterfaceCT2Test() throws IllegalArgumentException, IllegalAccessException  {        

        MethodAttrs ct2Methods[] = {        
          new MethodAttrs(ComplexType1.class, "getCt1", new Class[] {}),
          new MethodAttrs(void.class,   "setCt1", new Class[] {ComplexType1.class}),
          new MethodAttrs(List.class, "getCt1List", new Class[] {})
        };
        
        Class c = null;
        
        try {
            c = Class.forName("com.ibm.sdo.tck.staticTest.CT2");
        } catch (ClassNotFoundException e) {
            assertTrue("Generated interface not found", false);
        }
        
        try {
            Field f = c.getDeclaredField("SHOULD_NOT_APPEAR_IN_GENERATED_INTERFACES");
            String implem_failed_to_substitute_msg = (String)f.get(null);
            assertTrue(implem_failed_to_substitute_msg, false);
        } catch (NoSuchFieldException e) {
            // arrival here implies success - i.e. something other than the TCK provided interfaces were found
            // on the classpath, ahead of the supplied versions
        }
        
        try {
            checkMethods(c, ct2Methods);
        }catch (NoSuchMethodException e) {
            assertTrue(false);
        }

    }

 

    private void checkMethods(Class clazz, MethodAttrs[] methods) throws NoSuchMethodException {
        int i = 0;
        Method m;

        for (i = 0; i < methods.length; i++) {
            m = clazz.getMethod(methods[i].methodname, methods[i].arguments);
            assertEquals(methods[i].returnType, m.getReturnType());
        }


    }
    

    @Test
    @TCKTest(
            precondition = {
                    "all required methods are available on the generated type ComplexType1",
                    "for all Properties, getProperty == default for Property",
                    "for all Properties, isSet(property) == false"
                },
            postcondition = {
                    "isSet Property is true after setProperty",
                    "getProperty is as expected after setProperty",
                    "isSet Property is false after unset Property",
                    "getInstanceClass on the Type of an instance is consistent with"
                    + " the instance"
                },
            specRef = { @SpecRef(version = "2.1.1 Early Draft", section = "5.1") }
        )
    public void ct1BasicInterfaceTest() {
        ComplexType1 ct1 = (ComplexType1)dataFactory.create(ComplexType1.class);
    
        assertTrue(ct1 instanceof DataObject);
        assertTrue(((DataObject)ct1).getType().getInstanceClass().isInstance(ct1));
        
        assertFalse(isSet("attr1", (DataObject)ct1));
        assertNull(ct1.getAttr1());
        ct1.setAttr1("attr1Value");
        String attr1Val = ct1.getAttr1();
        assertEquals("attr1Value", attr1Val);
        assertTrue(isSet("attr1", (DataObject)ct1));
        unset("attr1", (DataObject)ct1);
        assertFalse(isSet("attr1", (DataObject)ct1));

        assertFalse(isSet("Elem1", (DataObject)ct1));
        assertNull(ct1.getElem1());
        ct1.setElem1("elem1Value");
        String elem1Val = ct1.getElem1();
        assertEquals("elem1Value", elem1Val);
        unset("attr1", (DataObject)ct1);
        assertFalse(isSet("attr1", (DataObject)ct1));
    }


    @Test
    @TCKTest(
            precondition = {
                    "all required methods are available on the generated type ComplexType1",
                    "for all Properties, getProperty == default for Property",
                    "for all Properties, isSet(property) == false"
                },
            postcondition = {
                    "isSet Property is true after setProperty",
                    "getProperty is as expected after setProperty",
                    "isSet Property is false after unset Property"
                },
            specRef = { @SpecRef(version = "2.1.1 Early Draft", section = "5.1") }
        )
    public void ct2BasicInterfaceTest()  {
        CT2 ct2 = (CT2)dataFactory.create(CT2.class);

        assertTrue(ct2 instanceof DataObject);

        assertFalse(isSet("ct1",(DataObject)ct2));
        assertNull(ct2.getCt1());
        ct2.setCt1((ComplexType1)dataFactory.create(ComplexType1.class));
        assertNotNull("attr1Value", ct2.getCt1());
        assertTrue(isSet("ct1",(DataObject)ct2));
        unset("ct1", (DataObject)ct2);
        assertFalse(isSet("ct1",(DataObject)ct2));

        assertEquals(0, ct2.getCt1List().size());;      
    }
    
    @Test
    @TCKTest(
            precondition = {
                    "The Type found by looking up using the generated class is "
                    + "the same as that found by looking up by NS URI and Type "
                    + "name, and is the same as that associated with an instance of the Type"
                },
            specRef = { @SpecRef(version = "2.1.1 Early Draft", section = "5") }
        )
    public void ct1TypeTest() {
        Type t1 = helperContext.getTypeHelper().getType(ComplexType1.class);
        assertNotNull(t1);
        Type t2 = helperContext.getTypeHelper().getType(NS_SIMPLE_STATIC_XSD, "ComplexType1");
        assertSame(t1, t2);
        ComplexType1 ct1 = (ComplexType1)dataFactory.create(ComplexType1.class);
        assertSame(t1, ((DataObject)ct1).getType());
    }

    @Test
    @TCKTest(
            precondition = {
                    "The Type found by looking up using the generated class is "
                    + "the same as that found by looking up by NS URI and Type "
                    + "name, and is the same as that associated with an instance of the Type",
                    "The generator has heeded the sdo:name annotation on the input schema"
                },
            specRef = { @SpecRef(version = "2.1.1 Early Draft", section = "5") }
        )
    public void ct2TypeTest() {
        Type t1 = helperContext.getTypeHelper().getType(CT2.class);
        assertNotNull(t1);
        Type t2 = helperContext.getTypeHelper().getType(NS_SIMPLE_STATIC_XSD, "CT2");
        assertSame(t1, t2);
        assertEquals("CT2", t1.getName());

    }
    
    

    @Test
    @TCKTest(
            precondition = {
                    "The result of serializing and deserializing an instance of a "
                    + "generated class results in another instance of that generated class "
                    + "and the instances are equal"
                },
            specRef = { @SpecRef(version = "2.1.1 Early Draft", section = "5.1") }
        )
    public void serializeDeserializeTest() throws IOException, ClassNotFoundException {
        
        CT2 ct2 = (CT2)dataFactory.create(CT2.class);
        ComplexType1 ct1 = (ComplexType1)dataFactory.create(ComplexType1.class);
        ct2.setCt1(ct1);
        ct1.setAttr1("attr1");
        ct1.setElem1("elem1");
        
         byte[] bytes = TCKUtil.serializeDataObjectXML(
                 (DataObject)ct2, "http://www.example.com/cttest", "Root", helperContext);
         DataObject dataObject2 = TCKUtil.deserializeDataObjectXML(bytes, helperContext);

         assertTrue(dataObject2 instanceof CT2);
         assertEquals("CT2", ((DataObject)ct2).getType().getName());

         CT2 ct2_2 = (CT2)dataObject2;
         assertEquals("attr1", ct2_2.getCt1().getAttr1());
         EqualityTest et = new EqualityTest((DataObject)ct2,(DataObject)ct2_2,true);        
         assertTrue(et.getFailure(),et.getResult());
    }
    
    private boolean isSet(String propertyName, DataObject dob) {
        Boolean result = null;
        try {
            String methodName = "isSet" + (propertyName.substring(0,1).toUpperCase()) + propertyName.substring(1);
            Method m = dob.getClass().getMethod(methodName);
            result = (Boolean)m.invoke(dob);
        }
        catch (NoSuchMethodException e) {
            result = new Boolean(dob.isSet(propertyName));
        } catch (IllegalArgumentException e) {
            assertFalse(true);
        } catch (IllegalAccessException e) {
            assertFalse(true);
        } catch (InvocationTargetException e) {
            assertFalse(true);
        }

        return result.booleanValue();
    }
    
    
    private void unset(String propertyName, DataObject dob) {
        try {
            String methodName = "unset" + (propertyName.substring(0,1).toUpperCase()) + propertyName.substring(1);
            Method m = dob.getClass().getMethod(methodName);
            m.invoke(dob);
        }
        catch (NoSuchMethodException e) {
            dob.unset(propertyName);
        } catch (IllegalArgumentException e) {
            assertFalse(true);
        } catch (IllegalAccessException e) {
            assertFalse(true);
        } catch (InvocationTargetException e) {
            assertFalse(true);
        }
    }

    @Before
    public void setUp() throws Exception {
        helperContext = getTestHelper().createHelperContext();
        dataFactory = helperContext.getDataFactory();
        
        
        getTestHelper().defineStaticTypes(helperContext, NS_SIMPLE_STATIC_XSD,
                RES_PATH_SIMPLE_STATIC_XSD, this.getClass().getClassLoader());
        
    }
}
