/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 *  
 */

package com.ibm.sdo.tck.tests.xml;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import java.io.IOException;
import java.io.InputStream;
import java.math.BigDecimal;
import java.net.URL;
import java.util.List;


import org.junit.Before;
import org.junit.Test;

import com.ibm.sdo.tck.framework.TCKTestBase;
import com.ibm.sdo.tck.framework.annotations.Operation;
import com.ibm.sdo.tck.framework.annotations.TCKTest;

import commonj.sdo.ChangeSummary;
import commonj.sdo.DataObject;
import commonj.sdo.Type;
import commonj.sdo.helper.HelperContext;
import commonj.sdo.helper.XMLDocument;

/**
 * Tests ChangeSummary serialization ...
 * <ul>
 * <li> Verification of change summary serialization before and after undoChanges()
 * </ul>
 * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=115">2.1 spec section 13</a>
 */
@TCKTest(testFor={@Operation(className = "ChangeSummary", methodName = "beginLogging", signature = ""),
                  @Operation(className = "ChangeSummary", methodName = "endLogging", signature = ""),
                  @Operation(className = "ChangeSummary", methodName = "undoChanges", signature = "")})
public class ChangeSummarySerializationTest extends TCKTestBase
{
    static final String NS              = "http://www.example.com/cstest";
    static final String ROOTNAME        = "Root";
    
    private HelperContext scope;
    
    @Before
    public void setUp() throws Exception 
    {
        super.setUp();
        
        scope = getScope();
        
        // Populate the meta data for the test (Stock Quote) model
        URL url = getClass().getResource("/sharedFiles/changeSummaryTest.xsd");
        InputStream inputStream = url.openStream();
        scope.getXSDHelper().define(inputStream, url.toString());
        inputStream.close();
    }
    
    @Test
    public void testLoadChangeSummary() throws IOException
    {
        URL url = getClass().getResource("/changeSummarySerializationTest/ChangeSummarySerializationTest.xml");
        InputStream inputStream = url.openStream();
        XMLDocument xmlDoc = scope.getXMLHelper().load(inputStream);
        inputStream.close();
        
        assertNotNull(xmlDoc);
        DataObject dobj = xmlDoc.getRootObject();
        assertNotNull(dobj);
        
        ChangeSummary cs = dobj.getChangeSummary();
        assertNotNull(cs);
        
        assertTrue(cs.isLogging());        
        cs.endLogging();
        assertEquals("1009.2",dobj.getString("price"));
        List <DataObject> children = dobj.getList("quotes");
        assertNotNull(children);
        DataObject child = children.get(0);
        assertNotNull(child);
        assertEquals("1985.5",child.getString("price"));           
        cs.undoChanges();
        assertEquals("1000.9",dobj.getString("price"));
        assertEquals("2000.9",child.getString("price"));   
    }
    
    //The following method is the source used
    //to generate the xml instance used in the testLoadChangeSummary()
    private void generateCSInstance()
    {
        Type type = getScope().getTypeHelper().getType(NS,ROOTNAME);
        DataObject quote = getScope().getDataFactory().create(type);

        ChangeSummary changeSummary = quote.getChangeSummary();

        quote.setString("symbol", "fbnt");
        quote.setString("companyName", "FlyByNightTechnology");
        quote.setBigDecimal("open1", new BigDecimal("1000.9"));
        quote.setBigDecimal("price", new BigDecimal("1000.9"));
        quote.setBigDecimal("high", new BigDecimal("1000.9"));
        quote.setBigDecimal("low", new BigDecimal("1000.9"));
        quote.setDouble("volume", 1000);
        quote.setDouble("change1", 1000);

        DataObject child = quote.createDataObject("quotes");
        child.setString("symbol", "fbnt-child");
        child.setBigDecimal("price", new BigDecimal("2000.9"));

        changeSummary.beginLogging();
        
        quote.setBigDecimal("price", new BigDecimal("1009.2"));
        child.setBigDecimal("price", new BigDecimal("1985.5")); 
        
        String xml = scope.getXMLHelper().save(quote, NS, ROOTNAME);
        System.out.println(xml);
        
    }
}
