/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 *  
 */
package com.ibm.sdo.tck.tests.xml;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;


import org.junit.Before;
import org.junit.Test;

import com.ibm.sdo.tck.framework.TCKTestBase;
import com.ibm.sdo.tck.framework.annotations.Operation;
import com.ibm.sdo.tck.framework.annotations.SpecRef;
import com.ibm.sdo.tck.framework.annotations.TCKTest;

import commonj.sdo.helper.XMLDocument;
import commonj.sdo.DataObject;
import commonj.sdo.Property;

/**
 * This tests compliance with section 9.10 of the SDO 2.1 Java specification.
 */
@TCKTest(testFor={@Operation(className = "DataObject", methodName = "getDataObject", signature = "String"),                    
                  @Operation(className = "DataObject", methodName = "getInstanceProperty", signature = "String"),                    
                  @Operation(className = "DataObject", methodName = "getString", signature = "Property"),                    
                  @Operation(className = "Property", methodName = "isOpenContent", signature = ""),                    
                  @Operation(className = "Property", methodName = "getType", signature = ""),                    
                  @Operation(className = "Property", methodName = "isContainment", signature = ""),
                  @Operation(className = "Type", methodName = "isDataType", signature = ""),                    
                  @Operation(className = "Type", methodName = "getName", signature = ""),                    
                  @Operation(className = "XMLHelper", methodName = "load", signature = "String"),                    
                  @Operation(className = "XMLDocument", methodName = "getRootObject", signature = "")})
public class XMLWithoutSchemaTest extends TCKTestBase {


  protected String xml = "<XMLWithoutSchemaTest xmlns:xs=\"http://www.w3.org/2001/XMLSchema\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\">" +
        "<person id=\"1\"><name>Joe</name><age xsi:type=\"xs:int\">21</age></person>" +
        "<person id=\"2\"><name>Sam</name><age xsi:type=\"xs:int\">40</age></person>" +
        "</XMLWithoutSchemaTest>";
  

  @Before
  public void setUp() throws Exception {
      super.setUp();
  }
  
  @TCKTest(specRef={@SpecRef(version="2.1",section = "9.10",note="bullet 1")})
  @Test
  public void testRootObject() {
    XMLDocument doc = getScope().getXMLHelper().load( xml );
    DataObject root = doc.getRootObject();
    assertTrue( root.getType().isOpen() );
    assertTrue( root.getType().isSequenced() );
  }

  @TCKTest(specRef={@SpecRef(version="2.1",section = "9.10",note="bullet 3")})
  @Test
  public void testAttributeProperties() {
    XMLDocument doc = getScope().getXMLHelper().load( xml );
    DataObject root = doc.getRootObject();
    DataObject person = root.getDataObject( "person.0" );

    // test that the property is correctly defined
    Property idProperty = person.getInstanceProperty( "id" );
    assertTrue( idProperty.isOpenContent() );
    if( idProperty.getType().isDataType() )
        assertEquals( "1", person.getString(idProperty) );                
    else 
        assertEquals( "1", person.getDataObject(idProperty).getString("value") );                
  }

  @TCKTest(specRef={@SpecRef(version="2.1",section = "9.10",note="bullet 4,5,6")})
  @Test
  public void testElementProperties() {
    XMLDocument doc = getScope().getXMLHelper().load( xml );
    DataObject root = doc.getRootObject();

    Property personProperty = root.getInstanceProperty( "person" );
    assertTrue( personProperty.isContainment() );
    assertTrue( personProperty.isOpenContent() );
    
    // test that we can access the instance property
    DataObject person = root.getDataObject( "person.0" );
    assertNotNull( person );
  }

  @TCKTest(specRef={@SpecRef(version="2.1",section = "9.10",note="bullet 6"),
                    @SpecRef(version="2.1",section = "3.11.4")})
  @Test
  public void testComplexElementWithXsiType() {
    XMLDocument doc = getScope().getXMLHelper().load( xml );
    DataObject root = doc.getRootObject();
    DataObject person = root.getDataObject( "person.0" );
    Property ageProperty = person.getInstanceProperty( "age" );
    String typeName = ageProperty.getType().getName();
    String typeURI  = ageProperty.getType().getURI();
    if( "commonj.sdo".equals(typeURI) )
    {
        assertTrue( "DataObject".equals(typeName) |
                "Int".equals(typeName) |
                "Object".equals(typeName) );         
    }
    else if( "commonj.sdo/java".equals(typeURI) )
    {
        assertTrue( "IntObject".equals(typeName) );                 
    }
    else
        fail("Property type URI not an expected value.");
  }

  @TCKTest(specRef={@SpecRef(version="2.1",section = "9.10",note="bullet 6")})
  @Test
  public void testElementsWithSimpleContent() {
    XMLDocument doc = getScope().getXMLHelper().load( xml );
    DataObject root = doc.getRootObject();
    DataObject person = root.getDataObject( "person.0" );

    // test that the property is correctly defined
    Property nameProperty = person.getInstanceProperty( "name" );
    assertTrue( nameProperty.isOpenContent() );
    String typeName = nameProperty.getType().getName();
    if( nameProperty.getType().isDataType() )
        assertEquals( "Joe", person.getString(nameProperty) );                
    else
        assertEquals( "Joe", person.getDataObject(nameProperty).getString("value") );                        
  }
  
  @Test
  public void testComplexElementWithSimpleContent() {
    String xml = "<root><name lang=\"en_US\">Adam</name></root>";
    XMLDocument doc = getScope().getXMLHelper().load( xml );
    DataObject root = doc.getRootObject();
    Property nameProperty = root.getInstanceProperty( "name" );
    
    assertEquals( "commonj.sdo", nameProperty.getType().getURI() );
    assertEquals( "DataObject", nameProperty.getType().getName() );
    
    DataObject dobj = root.getDataObject( "name.0" );
    assertEquals( "en_US", dobj.getString( "lang" ) );
    assertEquals( "Adam", dobj.getSequence().getValue(0) );
  }
}
