/**
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 *  
 *  $Rev: 537209 $  $Date: 2008/01/31 17:01:46 $
 */
package com.ibm.sdo.tck.tests.xml;

import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.util.List;
import java.util.Vector;


import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import com.ibm.sdo.tck.framework.TCKTestBase;
import com.ibm.sdo.tck.framework.annotations.Operation;
import com.ibm.sdo.tck.framework.annotations.SpecRef;
import com.ibm.sdo.tck.framework.annotations.TCKTest;

import commonj.sdo.DataObject;
import commonj.sdo.Type;
import commonj.sdo.helper.TypeHelper;
import commonj.sdo.helper.XSDHelper;

/**
 * Tests XSD serialization/deserialization.<p/>
 * This tests requires extension by import or creation of further tests.
 * It currently only tests one flavour of {@link XSDHelper#define(java.io.InputStream, String)} and {@link XSDHelper#generate(List)})
 * 
 * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=52">2.1 spec section 3.13</a>
 * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=56">2.1 spec section 4</a>
 * 
 */

public class XSDHelperGenerateTest extends TCKTestBase {
    @Before 
    public void setUp() throws Exception
    {
      super.setUp();
    }
    
    @After
    public void tearDown() throws Exception 
    {
      super.tearDown();
    }

    /**
     * Verifies the performance of XSDHelper.generate for dynamic SDOs with no XSD model.<p/>
     * Could improve this by postconditions on generated schema. 
     * @see commonj.sdo.XSDHelper#generate(List)
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=53">2.1 spec section 3.13.2</a>
     */
    @Test
    @TCKTest(
        specRef={
            @SpecRef(
                version="2.1",section = "3.13.2"
            )
        },
        testFor={
            @Operation(
                className="XSDHelper",
                methodName="generate",
                signature="List",
                note={
                    "Generation of XSD for types defined by dynamic API."
                }
            )
        }
      )

    public void testXSDGeneration_DynamicSDOType() {
        try {
            boolean exceptionCaught = false;

            // test for dynamic SDOs that have no XSD model. Here the testcase
            // succeeds only if the
            // xsd is generated by XSDHelper in which case xsd must not be null
            XSDHelper xsdHelper = getScope().getXSDHelper();
            DataObject quoteType = getScope().getDataFactory().create("commonj.sdo", "Type");
            quoteType.set("uri", "http://www.example.com/dynamic");
            quoteType.set("name", "DynamicQuote");

            TypeHelper th = getScope().getTypeHelper();
            DataObject aProperty = quoteType.createDataObject("property");
            aProperty.set("name", "symbol");
            aProperty.set("type", th.getType("commonj.sdo", "String"));

            aProperty = quoteType.createDataObject("property");
            aProperty.set("name", "price");
            aProperty.set("type", th.getType("commonj.sdo", "Decimal"));

            aProperty = quoteType.createDataObject("property");
            aProperty.set("name", "volume");
            aProperty.set("type", th.getType("commonj.sdo", "Double"));

            th.define(quoteType);

            Type dynamicQuoteType = th.getType("http://www.example.com/dynamic", "DynamicQuote");

            Vector types = new Vector();
            types.add(dynamicQuoteType);
            String xsd = null;

            try {
                xsd = xsdHelper.generate(types);
                assertNotNull("XSDHelper.generate() did not complete as expected for dynamic SDOs with no XSD model.  Exception was thrown",
                    xsd);
            } catch (IllegalArgumentException e) {
                fail("XSDHelper.generate() did not complete as expected for dynamic SDOs with no XSD model.  Exception was thrown : " + e
                    .toString());
            } catch (Exception e) {
                fail("Exception caught when generating a schema");
            }

        } catch (Exception e) {
            e.printStackTrace();
            fail("Exception caught" + e.toString());
        }
    }
        
    /**
     * Test for XSD content <p/>
     * @see commonj.sdo.XSDHelper#isXSD(Type)
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=53">2.1 spec section 3.13.3</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.13.3")},
        testFor={@Operation(className="XSDHelper",methodName="isXSD",signature="Type")})
    @Test
    public void testXSD() {
        String namespace1  = "http://www.example.com/dynamic1";
        String namespace2  = "http://www.example.com/dynamic2";
        String name = "DynamicQuote";
        
        defineType(namespace1, name);
        Type t1 = getScope().getTypeHelper().getType(namespace1, name);
        Vector types = new Vector();
        types.add(t1);
        String xsd = getScope().getXSDHelper().generate(types);
        assertFalse(getScope().getXSDHelper().isXSD(t1));
        xsd = xsd.replaceAll(namespace1,namespace2);
        getScope().getXSDHelper().define(xsd);
        Type t2 = getScope().getTypeHelper().getType(namespace2,name);
        assertTrue(getScope().getXSDHelper().isXSD(t2));
    } 
    
    /**
     * Test for XSD content <p/>
     * @see commonj.sdo.XSDHelper#generate(List,Map)
     * @see <a href="http://osoa.org/download/attachments/36/Java-SDO-Spec-v2.1.0-FINAL.pdf?version=1#page=53">2.1 spec section 3.13.3</a>
     */
    @TCKTest(specRef={@SpecRef(version="2.1",section = "3.13.3")},
        testFor={@Operation(className="XSDHelper",methodName="generate",signature="List,Map")})
    @Test
    public void testGenerate() {
        String namespace1  = "http://www.example.com/dynamic";
        String namespace2  = "http://www.example.com/new";
        String name = "DynamicQuote";
        
        try {
            defineType(namespace1, name);
            Type dynamicQuoteType = getScope().getTypeHelper().getType(namespace1, name);
            
            Vector types = new Vector();
            types.add(dynamicQuoteType);
            String xsd = getScope().getXSDHelper().generate(types);
            
            xsd = xsd.replaceAll(namespace1,namespace2);
            getScope().getXSDHelper().define(xsd);
            Type dynamicQuoteType2 = getScope().getTypeHelper().getType(namespace2,name);
            
            assertNotNull(dynamicQuoteType2);
            
        } 
        catch (Exception e) {
            fail(e.toString());
        }
    } 
    
    private void defineType(String namespace, String name )
    {
        XSDHelper xsdHelper = getScope().getXSDHelper();
        DataObject quoteType = getScope().getDataFactory().create("commonj.sdo", "Type");
        quoteType.set("uri", namespace);
        quoteType.set("name", name);
        
        DataObject aProperty = quoteType.createDataObject("property");
        aProperty.set("name", "symbol");
        aProperty.set("type", getScope().getTypeHelper().getType("commonj.sdo", "String"));
        
        aProperty = quoteType.createDataObject("property");
        aProperty.set("name", "price");
        aProperty.set("type", getScope().getTypeHelper().getType("commonj.sdo", "Decimal"));
        
        aProperty = quoteType.createDataObject("property");
        aProperty.set("name", "volume");
        aProperty.set("type", getScope().getTypeHelper().getType("commonj.sdo", "Double"));
        
        getScope().getTypeHelper().define(quoteType);
    }    
}
