/**
 *
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 */
package com.ibm.sdo.tck.util;


import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.net.URL;
import java.util.ArrayList;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.List;
import java.util.ListIterator;
import java.util.TimeZone;

import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.datatype.XMLGregorianCalendar;


import commonj.sdo.helper.TypeHelper;
import commonj.sdo.DataObject;
import commonj.sdo.Property;
import commonj.sdo.Type;
import commonj.sdo.helper.HelperContext;

import commonj.sdo.impl.RuntimeHelper;

public class TCKUtil {
    
    
    public static final String TEST_MODEL = "/sharedFiles/APITest.xsd";
    public static final String TEST_NAMESPACE = "http://www.example.com/tck_api_test";
    public static final String API_TYPE = "APITest";
    public static final String SEQ_TYPE = "Sequenced";
    public static final String EXT_TYPE = "Extended";
    public static final String ABSTRACT_TYPE = "Abstract"; 
    public static final String API_ELEMENT = "apiTestElem";  
    public static final String DATE_TEST_DATA = "2008-04-06T12:30:45.005Z";
    
    public static void defineXSDAPITestTypes(HelperContext hc) throws IOException 
    {

        Type t = hc.getTypeHelper().getType(TEST_NAMESPACE, "Abstract");
        if (t != null) {
            // the test types have already been defined in this HelperContext (this will happen when the
            // implementation being tested does not support multiple contexts, which is not a 2.1 requirement).
            return;
        }

        // Populate the meta data for the test model (APITest)
        URL url = TCKUtil.class.getResource(TEST_MODEL);
        InputStream inputStream = url.openStream();
        hc.getXSDHelper().define(inputStream, url.toString());
        inputStream.close();  
    }
  
    /**
     * this factory currently simple makes the assumption that the variant string is the
     * name of a type for which an empty instance is required
     */
    public static DataObject createTestDataObject(HelperContext scope, String variantString) throws Exception {
      return scope.getDataFactory().create(TCKUtil.TEST_NAMESPACE, variantString);
    }
    
    /**
     * populateFields uses set<Type> to set each of the fields in the
     * DataObject. It is used to ensure a known set of expected values that are
     * not other than the default values for the various fields.
     * 
     * @param testDO
     * @param helperContext
     * @throws ExpectedConditionError
     */
    public static void populateAPITestObjectProperties(DataObject testDO, HelperContext scope) throws Exception {
        populateAPITestSimpleProperties(testDO);
        testDO.setDataObject("sequencedElem", createTestDataObject(scope, TCKUtil.SEQ_TYPE));
        testDO.setDataObject("extendedElem", createTestDataObject(scope, TCKUtil.EXT_TYPE));
        testDO.setDataObject("contain", createTestDataObject(scope, TCKUtil.API_TYPE));
        List containMany = new ArrayList();
        DataObject containedDO = createTestDataObject(scope, TCKUtil.API_TYPE);
        populateAPITestSimpleProperties(containedDO);
        containMany.add(containedDO);
        testDO.setList("containMany", containMany);
    }
    
    private static void populateAPITestSimpleProperties(DataObject testDO)
    {
        testDO.setString("stringVal", "String 1");
        testDO.setBoolean("booleanVal", true);
        testDO.setBoolean("booleanVal2", false);
        testDO.setByte("byteVal", (byte)-127);
        testDO.setString("stringVal2", "Second string!");
        testDO.setBigDecimal("decimalVal", new BigDecimal(-3.00003));
        testDO.setBigDecimal("decimalVal2", new BigDecimal(18883.999999));
        testDO.setInt("intVal", (int)33333);
        testDO.setFloat("floatVal", (float)0.88881);
        testDO.setDouble("doubleVal", (double)119.13813);
        testDO.setDate("dateVal", getDateValue());
        testDO.setShort("shortVal", (short)-800);
        testDO.setLong("longVal", (long)88881113);
        testDO.setBytes("bytesVal", new byte[] {120, 80, -40});
        testDO.setBigInteger("integerVal", new BigInteger("88819313"));
        testDO.setChar("charVal", '*');        
    }
    
    
  public static String createUniqueName() {
    return "name-" + System.currentTimeMillis() + "-"
        + ((int) (1000 * Math.random()));
  }
  
  /**
   * areEqualTypes is used to determine of two Types are equivalent even if
   * not identically equal. The names of the Types are compared, as well as
   * the Types of each of the Properties making up the Type.
   * 
   * @param type1
   * @param type2
   * @return
   */
  public static boolean areEqualTypes(Type type1, Type type2) {
      List properties1, properties2;
      Property property1, property2;
      int size = 0, j = 0, k;
      boolean found;

      // Equivalent Types have the same name

      if (!(type1.getName().equals(type2.getName())))
          return false;

      // Equivalent Types have the same number of Properties

      properties1 = type1.getProperties();
      properties2 = type2.getProperties();
      size = properties1.size();

      if (size != properties2.size())
          return false;

      // Equivalent Types have Properties of the same name and Type

      for (int i = 0; i < size; i++) {
          property1 = (Property)properties1.get(i);
          k = 0;
          found = false;

          while (k < size && !found) {
              // j is used to prevent the initial Properties in properties2
              // from being checked every time
              // j is particularly useful when the Types have Properties in
              // the order

              property2 = (Property)properties2.get((k + j) % size);

              if (property1.getName().equals(property2.getName())) {
                  j++;
                  found = true;

                  // Should not use recursion here to compare the Types of the
                  // Properties, because
                  // it is possible that a Type may recursively contain
                  // itself.

                  if (!(property1.getType().getName().equals(property2.getType().getName())))
                      return false;
              }
              k++;
          }
          if (!found)
              return false;
      }
      return true;
  }
  
  /**
   * Uses the XMLHelper to serialize the input DataObject
   * 
   * @param dataObject
   * @param baos
   * @param helperContexgt
   * @throws IOException
   */
  public static byte [] serializeDataObjectXML(DataObject dataObject, String uri, String name, HelperContext helperContext) throws IOException {
      ByteArrayOutputStream byteArrayOutput = new ByteArrayOutputStream();
      Type type = dataObject.getType();
      helperContext.getXMLHelper().save(dataObject, uri, name, byteArrayOutput);
      return byteArrayOutput.toByteArray();
  }
  
  /**
   * Uses the XMLHelper to deserialize the input XML file
   * 
   * @param bais
   * @param helperContext
   * @return
   * @throws IOException
   * @throws ClassNotFoundException
   */
  public static DataObject deserializeDataObjectXML(byte [] obj, HelperContext helperContext) throws IOException, ClassNotFoundException {
      return helperContext.getXMLHelper().load(new ByteArrayInputStream(obj)).getRootObject();
  }
  
  /**
   * Uses Java serialization to serialize the input DataObject
   * 
   * @throws IOException
   */
  public static byte [] serializeDataObjectJava(RuntimeHelper testHelper, DataObject dataObject, HelperContext helperContext) throws IOException 
  {
      ByteArrayOutputStream byteArrayOutput = new ByteArrayOutputStream();
      ObjectOutputStream out = testHelper.createObjectOutputStream(byteArrayOutput, helperContext);
      out.writeObject(dataObject);
      out.close();
      return byteArrayOutput.toByteArray();
  }

  /**
   * Uses Java deserialization to deserialize the input XML file
   * 
   * @throws IOException
   * @throws ClassNotFoundException
   */
  public static DataObject deserializeDataObjectJava(RuntimeHelper testHelper, byte [] obj, HelperContext helperContext) throws IOException, ClassNotFoundException 
  {
      ObjectInputStream input = testHelper.createObjectInputStream(new ByteArrayInputStream(obj), 
                                                                   helperContext);
      Object object = input.readObject();
      input.close();
      return (DataObject)object;
  }

  
  public static void defineDynamicAPITestTypes(HelperContext hc) {
      TypeHelper types = hc.getTypeHelper();
      Property xmlElementProp = hc.getXSDHelper().getGlobalProperty("commonj.sdo/xml", "xmlElement", false);
       
      Type t = types.getType(TCKUtil.TEST_NAMESPACE, "Abstract");
      if (t != null) {
          // the test types have already been defined in this HelperContext (this will happen when the
          // implementation being tested does not support multiple contexts, which is not a 2.1 requirement).
          return;
      }

      Type stringType = types.getType("commonj.sdo", "String");
      Type intType = types.getType("commonj.sdo", "Int");
      Type booleanType = types.getType("commonj.sdo", "Boolean");
      Type byteType = types.getType("commonj.sdo", "Byte");
      Type decimalType = types.getType("commonj.sdo", "Decimal");
      Type floatType = types.getType("commonj.sdo", "Float");
      Type doubleType = types.getType("commonj.sdo", "Double");
      Type dateType = types.getType("commonj.sdo", "Date");
      Type shortType = types.getType("commonj.sdo", "Short");
      Type longType = types.getType("commonj.sdo", "Long");
      Type bytesType = types.getType("commonj.sdo", "Bytes");
      Type integerType = types.getType("commonj.sdo", "Integer");
      Type charType = types.getType("commonj.sdo", "Character");

      DataObject abstractTypeDO = hc.getDataFactory().create("commonj.sdo", "Type");
      abstractTypeDO.set("uri", TCKUtil.TEST_NAMESPACE);
      abstractTypeDO.set("name", "Abstract");
      abstractTypeDO.setBoolean("abstract", true);

      DataObject firstProperty = abstractTypeDO.createDataObject("property");
      firstProperty.set("name", "firstName");
      firstProperty.set("type", stringType);
      firstProperty.set(xmlElementProp, Boolean.TRUE);

      DataObject lastProperty = abstractTypeDO.createDataObject("property");
      lastProperty.set("name", "lastName");
      lastProperty.set("type", stringType);
      lastProperty.set(xmlElementProp, Boolean.TRUE);
      
      Type abstractType = types.define(abstractTypeDO);

      DataObject extendedTypeDO = hc.getDataFactory().create("commonj.sdo", "Type");
      extendedTypeDO.set("uri", TCKUtil.TEST_NAMESPACE);
      extendedTypeDO.set("name", TCKUtil.EXT_TYPE);
      
      List baseTypes = new ArrayList();
      baseTypes.add(abstractType);
      extendedTypeDO.setList("baseType", baseTypes);
      
      DataObject middleName = extendedTypeDO.createDataObject("property");
      middleName.set("name", "middleName");
      middleName.set("type", stringType);
      middleName.set(xmlElementProp, Boolean.TRUE);
      
      DataObject nickName = extendedTypeDO.createDataObject("property");
      nickName.set("name", "nickName");
      nickName.set("type", stringType);
      nickName.set(xmlElementProp, Boolean.TRUE);
      
      Type extendedType = types.define(extendedTypeDO);
      
      DataObject sequenceTypeDO = hc.getDataFactory().create("commonj.sdo", "Type");
      sequenceTypeDO.set("uri", TCKUtil.TEST_NAMESPACE);
      sequenceTypeDO.set("name", TCKUtil.SEQ_TYPE);
      sequenceTypeDO.setBoolean("sequenced", true);
      
      // TCKTODO: remove the try block - eventually, when test implemenation is fixed
      //List aliases = new ArrayList();
      //aliases.add("Seq2"); 
      //sequenceTypeDO.set("aliasName", aliases);

      DataObject Letters = sequenceTypeDO.createDataObject("property");
      Letters.set("name", "Letters");
      Letters.set("type", stringType);
      Letters.setBoolean("many", true);
      Letters.set(xmlElementProp, Boolean.TRUE);
      
      DataObject Numbers = sequenceTypeDO.createDataObject("property");
      Numbers.set("name", "Numbers");
      Numbers.set("type", intType);
      Numbers.setBoolean("many", true);
      Numbers.set(xmlElementProp, Boolean.TRUE);
      
      DataObject containManySeq = sequenceTypeDO.createDataObject("property");
      containManySeq.set("name", "containMany");
      containManySeq.set("type", sequenceTypeDO);
      containManySeq.setBoolean("many", true);
      containManySeq.setBoolean("containment", true);
      
      DataObject containSeq = sequenceTypeDO.createDataObject("property");
      containSeq.set("name", "contain");
      containSeq.set("type", sequenceTypeDO);
      containSeq.setBoolean("containment", true);

      Type sequenceType = types.define(sequenceTypeDO);

      DataObject openTypeDO = hc.getDataFactory().create("commonj.sdo", "Type");
      openTypeDO.set("uri", TCKUtil.TEST_NAMESPACE);
      openTypeDO.set("name", "Open");
      openTypeDO.setBoolean("open", true);
      
      DataObject definedElem = openTypeDO.createDataObject("property");
      definedElem.set("name", "defined");
      definedElem.set("type", stringType);
      definedElem.set(xmlElementProp, Boolean.TRUE);
      
      Type openType = types.define(openTypeDO);
      
      DataObject testTypeDO = hc.getDataFactory().create("commonj.sdo", "Type");
      testTypeDO.set("uri", TCKUtil.TEST_NAMESPACE);
      testTypeDO.set("name", TCKUtil.API_TYPE);
      
      DataObject stringProperty = testTypeDO.createDataObject("property");
      stringProperty.set("name", "stringVal");
      stringProperty.set("type", stringType);
      stringProperty.set(xmlElementProp, Boolean.TRUE);
      stringProperty.setBoolean("nullable", true);
      stringProperty.setString("default", "");
      
      DataObject booleanProperty = testTypeDO.createDataObject("property");
      booleanProperty.set("name", "booleanVal");
      booleanProperty.set("type", booleanType);
      booleanProperty.set(xmlElementProp, Boolean.TRUE);
      booleanProperty.setString("default", Boolean.toString(false));
      
      DataObject boolean2Property = testTypeDO.createDataObject("property");
      boolean2Property.set("name", "booleanVal2");
      boolean2Property.set("type", booleanType);
      boolean2Property.set(xmlElementProp, Boolean.TRUE);
      boolean2Property.setString("default", Boolean.toString(false));
      
      DataObject byteProperty = testTypeDO.createDataObject("property");
      byteProperty.set("name", "byteVal");
      byteProperty.set("type", byteType);
      byteProperty.set(xmlElementProp, Boolean.TRUE);
      byteProperty.setString("default", "0");
      
      DataObject string2Property = testTypeDO.createDataObject("property");
      string2Property.set("name", "stringVal2");
      string2Property.set("type", stringType);
      string2Property.set(xmlElementProp, Boolean.TRUE);
      string2Property.setBoolean("nullable", true);
      string2Property.setString("default", "");
      
      DataObject decimalProperty = testTypeDO.createDataObject("property");
      decimalProperty.set("name", "decimalVal");
      decimalProperty.set("type", decimalType);
      decimalProperty.set(xmlElementProp, Boolean.TRUE);
      decimalProperty.setBoolean("nullable", true);
      decimalProperty.setString("default", "0");
      
      DataObject decimal2Property = testTypeDO.createDataObject("property");
      decimal2Property.set("name", "decimalVal2");
      decimal2Property.set("type", decimalType);
      decimal2Property.set(xmlElementProp, Boolean.TRUE);
      decimal2Property.setBoolean("nullable", true);
      decimal2Property.setString("default", "0");
      
      // TCKTODO: remove the try block - eventually, when test implemenation is fixed
      //List aliases = new ArrayList(); 
      //aliases.add("Dec2");
      //decimal2Property.set("aliasName", aliases);
      
      DataObject intProperty = testTypeDO.createDataObject("property");
      intProperty.set("name", "intVal");
      intProperty.set("type", intType);
      intProperty.set(xmlElementProp, Boolean.TRUE);
      intProperty.setString("default", "0");
      
      DataObject floatProperty = testTypeDO.createDataObject("property");
      floatProperty.set("name", "floatVal");
      floatProperty.set("type", floatType);
      floatProperty.set(xmlElementProp, Boolean.TRUE);
      floatProperty.setString("default", "0");

      DataObject doubleProperty = testTypeDO.createDataObject("property");
      doubleProperty.set("name", "doubleVal");
      doubleProperty.set("type", doubleType);
      doubleProperty.set(xmlElementProp, Boolean.TRUE);
      doubleProperty.setString("default", "0");
      
      DataObject dateProperty = testTypeDO.createDataObject("property");
      dateProperty.set("name", "dateVal");
      dateProperty.set("type", dateType);
      dateProperty.set(xmlElementProp, Boolean.TRUE);
      dateProperty.setBoolean("nullable", true);
      dateProperty.setString("default", "2000-03-02T10:00:00");
      
      DataObject shortProperty = testTypeDO.createDataObject("property");
      shortProperty.set("name", "shortVal");
      shortProperty.set("type", shortType);
      shortProperty.set(xmlElementProp, Boolean.TRUE);
      shortProperty.setString("default", "0");
      
      DataObject longProperty = testTypeDO.createDataObject("property");
      longProperty.set("name", "longVal");
      longProperty.set("type", longType);
      longProperty.set(xmlElementProp, Boolean.TRUE);
      longProperty.setString("default", "0");
      
      DataObject containManyProperty = testTypeDO.createDataObject("property");
      containManyProperty.set("name", "containMany");
      containManyProperty.setBoolean("many", true);
      containManyProperty.setBoolean("containment", true);
      containManyProperty.set("type", testTypeDO);

      DataObject bytesProperty = testTypeDO.createDataObject("property");
      bytesProperty.set("name", "bytesVal");
      bytesProperty.set("type", bytesType);
      bytesProperty.set(xmlElementProp, Boolean.TRUE);
      bytesProperty.setBoolean("nullable", true);
      bytesProperty.setString("default", "00");
      
      DataObject integerProperty = testTypeDO.createDataObject("property");
      integerProperty.set("name", "integerVal");
      integerProperty.set("type", integerType);
      integerProperty.set(xmlElementProp, Boolean.TRUE);
      integerProperty.setBoolean("nullable", true);
      integerProperty.setString("default", "0");
      
      DataObject charProperty = testTypeDO.createDataObject("property");
      charProperty.set("name", "charVal");
      charProperty.set("type", charType);
      charProperty.set(xmlElementProp, Boolean.TRUE);
      charProperty.setBoolean("nullable", true);
      
      DataObject readOnlyProperty = testTypeDO.createDataObject("property");
      readOnlyProperty.set("name", "readOnlyVal");
      readOnlyProperty.set("type", stringType);
      readOnlyProperty.setBoolean("readOnly", true);
      readOnlyProperty.set(xmlElementProp, Boolean.TRUE);
      readOnlyProperty.setBoolean("nullable", true);
      
      DataObject sequenceProperty = testTypeDO.createDataObject("property");
      sequenceProperty.set("name", "sequencedElem");
      sequenceProperty.set("type", sequenceType);
      sequenceProperty.setBoolean("containment", true);
      
      DataObject extendProperty = testTypeDO.createDataObject("property");
      extendProperty.set("name", "extendedElem");
      extendProperty.set("type", extendedType);
      extendProperty.setBoolean("containment", true);
      
      DataObject openProperty = testTypeDO.createDataObject("property");
      openProperty.set("name", "openElem");
      openProperty.set("type", openType);
      openProperty.setBoolean("containment", true);
      
      DataObject containProperty = testTypeDO.createDataObject("property");
      containProperty.set("name", "contain");
      containProperty.setBoolean("containment", true);
      containProperty.set("type", testTypeDO); 
      
      Type testType = types.define(testTypeDO);
      
      DataObject subTestTypeDO = hc.getDataFactory().create("commonj.sdo", "Type");
      subTestTypeDO.set("uri", TCKUtil.TEST_NAMESPACE);
      subTestTypeDO.set("name", "SubAPITest");
      
      baseTypes = new ArrayList();
      baseTypes.add(testType);
      subTestTypeDO.setList("baseType", baseTypes);
      
      DataObject subString1 = subTestTypeDO.createDataObject("property");
      subString1.set("name", "subString1");
      subString1.set("type", stringType);
      subString1.set(xmlElementProp, Boolean.TRUE);
            
      DataObject subString2 = subTestTypeDO.createDataObject("property");
      subString2.set("name", "subString2");
      subString2.set("type", stringType);
      subString2.set(xmlElementProp, Boolean.TRUE);

      DataObject subString3 = subTestTypeDO.createDataObject("property");
      subString3.set("name", "subString3");
      subString3.set("type", stringType);
      subString3.set(xmlElementProp, Boolean.TRUE);

      DataObject subString4 = subTestTypeDO.createDataObject("property");
      subString4.set("name", "subString4");
      subString4.set("type", stringType);
      subString4.set(xmlElementProp, Boolean.FALSE);

      DataObject subString5 = subTestTypeDO.createDataObject("property");
      subString5.set("name", "subString5");
      subString5.set("type", stringType);
      subString5.set(xmlElementProp, Boolean.FALSE);

      DataObject subString6 = subTestTypeDO.createDataObject("property");
      subString6.set("name", "subString6");
      subString6.set("type", stringType);
      subString6.set(xmlElementProp, Boolean.FALSE);

      Type subTestType = types.define(subTestTypeDO);

    }
 
  public static Property getInstanceProperty(Object p, String name) {
      Property prop = null;
      List l = null;
      if( p instanceof commonj.sdo.Type )
          l = ((Type)p).getInstanceProperties();
      else
          l = ((Property)p).getInstanceProperties();
          
      //List l = p.getInstanceProperties();
      for( ListIterator li=l.listIterator(); li.hasNext(); )
      {
          Property p1 = (Property)li.next();
          if( name.equals(p1.getName()) )
          {
              prop = p1;
              break;
          }
      }
      return prop;
  }
  
  public static Date getDateValue()
  {
      XMLGregorianCalendar controlXMLCal;
      try 
      {
        controlXMLCal = DatatypeFactory.newInstance().newXMLGregorianCalendar(DATE_TEST_DATA);
        GregorianCalendar control = controlXMLCal.toGregorianCalendar(TimeZone.getTimeZone("GMT+00:00"), null, null);
        return control.getTime();
      } 
      catch (DatatypeConfigurationException e) 
      {
        // This should never happen, we have control of the input.
        e.printStackTrace();
        return null;
      }
  }

}
