/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 *  
 *  $Rev: 538928 $  $Date: 2008/02/06 21:04:48 $
 */
package com.ibm.sdo.tck.util;

import java.io.IOException;
import java.io.InputStream;
import java.io.StringReader;
import java.net.URL;

import org.custommonkey.xmlunit.Diff;
import org.custommonkey.xmlunit.Difference;
import org.custommonkey.xmlunit.DifferenceConstants;
import org.custommonkey.xmlunit.NodeDetail;
import org.custommonkey.xmlunit.XMLUnit;
import org.w3c.dom.Document;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

public class XMLCompareUtil extends Diff
{
    public static final int DISABLE_ALL_OPTIONS = 0;
    public static final int IGNORE_NAMESPACE_PREFIX = 1;
    public static final int IGNORE_NAMESPACE_URI = 2;
    public static final int IGNORE_QNAME_NAMESPACE_PREFIX = 4;
    
    private static int Options = getDefaultOptions();
    
    public XMLCompareUtil(Document control, Document test)
    {
        super(control, test);
    }
	
    
    private static int getDefaultOptions() {return IGNORE_NAMESPACE_PREFIX | IGNORE_NAMESPACE_URI | IGNORE_QNAME_NAMESPACE_PREFIX;}
    public static void setDefaultOptions() {Options = getDefaultOptions();}
    public static void setStrictOptions()  {Options = DISABLE_ALL_OPTIONS; }
    public static void setOptions(int opt) {Options = opt;}
    private static boolean optionSet(int value) { return( (Options&value)>=1 ); }
    
    /** 
     * Compare XMLs, ignoring lexical differences such as attribute order, text and CData and whitespace
     * Namespace declarations are also ignored.
     */
    public static boolean compareXML(InputStream expected, URL resultingXML)
    {
        try
    	{
    		return compareXML(new InputSource(expected), new InputSource(resultingXML.openStream()),true);
    	}
    	catch (IOException e)
    	{
    		return false;
    	}
    }

    /** 
     * Compare XMLs, ignoring lexical differences such as attribute order, text and CData and whitespace
     * Namespace declarations are also ignored.
     */
    public static boolean compareXML(String expected, String resultingXML)
    {
        return compareXML(new InputSource(new StringReader(expected)), new InputSource(new StringReader(resultingXML)),true);
    }

    /** 
     * Compare XMLs, ignoring lexical differences such as attribute order, text and CData and whitespace
     * Namespace declarations are also ignored.
     */
    public static boolean compareXML(InputStream expected, String resultingXML)
    {
        return compareXML(new InputSource(expected), new InputSource(new StringReader(resultingXML)),true);
    }

    /** 
     * Compare XMLs, ignoring lexical differences such as attribute order, text and CData and whitespace
     * Namespace declarations are also ignored.
     */
    public static boolean compareXMLSimilar(InputStream expected, String resultingXML)
    {
        return compareXML(new InputSource(expected), new InputSource(new StringReader(resultingXML)),false);
    }

    /** 
     * Compare XMLs, ignoring lexical differences such as attribute order, text and CData and whitespace
     * Namespace declarations are also ignored.
     */
    public static boolean compareXML(InputSource expectedXML, InputSource resultingXML, boolean identical)
    {
        try
        {
            XMLUnit.setIgnoreAttributeOrder(true);
            XMLUnit.setIgnoreDiffBetweenTextAndCDATA(true);
            XMLUnit.setNormalize(true);
            XMLUnit.setNormalizeWhitespace(true);
            XMLUnit.setIgnoreWhitespace(true);
            XMLUnit.setTransformerFactory("org.apache.xalan.processor.TransformerFactoryImpl");
            
            Document control = XMLUnit.buildControlDocument(expectedXML);
            Document test = XMLUnit.buildTestDocument(resultingXML);
            Diff diff = new XMLCompareUtil(control, test);
            if( identical )
            {
                if (!diff.identical())
                {
                    // Display differences
                    System.err.println(diff.toString());
                    return false;
                }
            }
            else
            {
                if (!diff.similar())
                {
                    // Display differences
                    System.err.println(diff.toString());
                    return false;
                }                
            }
            return true;
        }
        catch (SAXException e)
        {
            return false;
        }
        catch (IOException e)
        {
            return false;
        }
    }

    /** 
     * Compare XMLs, ignoring lexical differences such as attribute order, text and CData and whitespace
     * Check if namespace declarations are the same.
     */
    public static boolean compareXMLWithNSDecl(InputSource expectedXML, InputSource resultingXML)
    {
        try
        {
            XMLUnit.setIgnoreAttributeOrder(true);
            XMLUnit.setIgnoreDiffBetweenTextAndCDATA(true);
            XMLUnit.setNormalize(true);
            XMLUnit.setNormalizeWhitespace(true);
            XMLUnit.setIgnoreWhitespace(true);
            XMLUnit.setTransformerFactory("org.apache.xalan.processor.TransformerFactoryImpl");

            Document control = XMLUnit.buildControlDocument(expectedXML);
            Document test = XMLUnit.buildTestDocument(resultingXML);
            Diff diff = new Diff(control, test);
            if (!diff.identical())
            {
                // Display differences
                System.err.println(diff.toString());
                return false;
            }
            return true;
        }
        catch (SAXException e)
        {
            return false;
        }
        catch (IOException e)
        {
            return false;
        }
    }
    
    // Implements DifferenceListener

    @Override
    public int differenceFound(Difference diff)
    {
        switch (diff.getId())
        {
            case DifferenceConstants.NAMESPACE_PREFIX_ID :
                if(optionSet(IGNORE_NAMESPACE_PREFIX))
                    return RETURN_IGNORE_DIFFERENCE_NODES_IDENTICAL;
            case DifferenceConstants.NAMESPACE_URI_ID:
                // The namespace uri check is being discarded due to issues
                // where namespaces are declared in different locations within
                // the XML document.  In most cases, these differences are simply
                // implemenation dependent and don't affect the quality of the
                // instance.
                if(optionSet(IGNORE_NAMESPACE_URI))
                    return RETURN_IGNORE_DIFFERENCE_NODES_IDENTICAL;
            case DifferenceConstants.ATTR_VALUE_ID:
                // This condition will check and see if a QName value is being
                // checked.  If so, we need to ignore the NS prefix portion.
                if(optionSet(IGNORE_QNAME_NAMESPACE_PREFIX))
                {
                    NodeDetail cnd = diff.getControlNodeDetail();
                    String controlValue = cnd.getValue();
                    NodeDetail tnd = diff.getTestNodeDetail();
                    String testValue = tnd.getValue();
                    int controlIdx = controlValue.indexOf(':');
                    int testIdx = testValue.indexOf(':');
                    if(  (controlIdx >=0) || (testIdx >= 0) )
                    {
                        if( controlIdx >=0 && controlValue.length() > controlIdx+1 )
                            controlValue = controlValue.substring(controlIdx+1);
                    
                        if( testIdx >= 0 && testValue.length() > testIdx+1 )
                            testValue = testValue.substring(testIdx+1);
                    
                        if( controlValue.equals(testValue) )
                            return RETURN_IGNORE_DIFFERENCE_NODES_IDENTICAL;
                    }
                }
            default :
                return super.differenceFound(diff);
        }
    }
}
